#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include "VGFunction.h"

/**
 * VGF for the demand model described in the MCDB paper.
 * Used by MCDB Q4.
 *
 * Input schema: (price0shape, price0scale, demand0shape, demand0scale, currentDemand, currentPrice, newPrice)
 * Dist: X[i] ~ new demand {see the paper for details}
 * Output schema: (X)
 */



/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *p0shape;
  double *p0scale;
  double *d0shape;
  double *d0scale;
  double *curDemand;
  double *curPrice;
  double *newPrice;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  double *newDemand;
};

// ----------------------------------------------------------- // 

// helper functions
class GammaPDF {
  double k;  // shape
  double theta; // scale
  double firstTermMult;
  double secTermMult;

public:
  void Init (double k, double theta) {
    firstTermMult = (k - 1);
    secTermMult = 1.0 / theta;
  }

  double lGammaPDF (double x) {
    return firstTermMult * log(x) - x * secTermMult;
  }
};

class DoubleGammaPDF {
  double lowDemand, highDemand;
  double probOfMostLikelyDemand, in_demand;
  double in_price;

  GammaPDF price, demand;

public:
  void Init (double p, double d, double k_p, double k_m, double theta_p, double theta_m);
  double GetGmAtGivenPrice (double price);        
  void GetNewPAndD (double &p_0, double &d_0, gsl_rng *rng);
};

#define MAX_ITERS 30
#define EPSILON (1e-12)
#define MAX_P 1000

void DoubleGammaPDF :: GetNewPAndD (double &p_0, double &d_0, gsl_rng *rng) {

  // this implements rejection sampling for the price
  int counter = 0;
  while (1) {

    // first, randomly select a demand intercept
    d_0 = gsl_rng_uniform(rng) * (highDemand - lowDemand) + lowDemand;

    // now compute the corresponding price intercept
    double distanceRatio = in_demand / (d_0 - in_demand);
    p_0 = in_price + in_price * distanceRatio;
    double density = price.lGammaPDF (p_0);
    density += demand.lGammaPDF (d_0);
    double target = exp (-(probOfMostLikelyDemand - density));

    if (gsl_rng_uniform (rng) < target)
      return;

    counter++;
  }
}


void DoubleGammaPDF :: Init (double obs_price, double obs_demand, double k_p, double k_d, double theta_p, double theta_d) {

  // remember the inputs
  in_demand = obs_demand;
  in_price = obs_price;

  // initialize the two gamma PDFs
  price.Init (k_p, theta_p);
  demand.Init (k_d, theta_d);

  // find the maximum spot on the PDF
  lowDemand = in_demand + EPSILON;
  highDemand = theta_d * k_d * 100;

  for (int i = 0; i < MAX_ITERS; i++) {

    // get the density at the middle point
    double midDemand = (lowDemand + highDemand) / 2;
    double distanceRatio = in_demand / (midDemand - in_demand);
    double correspondingPrice = in_price + in_price * distanceRatio;
    double midDensity = price.lGammaPDF (correspondingPrice);
    midDensity += demand.lGammaPDF (midDemand);

    // get the density a bit higher up
    double demandPlusEps = midDemand + EPSILON;
    distanceRatio = in_demand / (demandPlusEps - in_demand);        
    correspondingPrice = in_price + in_price * distanceRatio;
    double midPlusEpsDensity = price.lGammaPDF (correspondingPrice);
    midPlusEpsDensity += demand.lGammaPDF (demandPlusEps);

    // see which direction is greater
    if (midDensity > midPlusEpsDensity) {
      highDemand = midDemand;
    } else {
      lowDemand = midDemand;
    }
  }

  // now we have the most likely setup
  double mostLikelyDemand = (lowDemand + highDemand) / 2;
  double distanceRatio = in_demand / (mostLikelyDemand - in_demand);
  double correspondingPrice = in_price + in_price * distanceRatio;

  // and the log of the PDF at that spot
  probOfMostLikelyDemand = price.lGammaPDF (correspondingPrice) +
    demand.lGammaPDF (mostLikelyDemand);

  // reset the lower bound on the demand
  lowDemand = in_demand + EPSILON;

  // this gets the range for rejection sampling
  double totSoFar = 1.0;
  double upperLimit = mostLikelyDemand;
  int counter = 0;
  while (1) {

    upperLimit += (mostLikelyDemand - in_demand);

    // get the probability density at this spot
    double distanceRatio = in_demand / (upperLimit - in_demand);
    double correspondingPrice = in_price + in_price * distanceRatio;
    double probHere = price.lGammaPDF (correspondingPrice) +
      demand.lGammaPDF (upperLimit);        

    // sanity check
    if (probOfMostLikelyDemand - probHere < 0.01) {
      printf("Bad... the density seems to be going DOWN...\n");
      return;
    }

    // see if this is really small, relatively speaking
    double additionalMass = exp(-(probOfMostLikelyDemand - probHere));
    if (additionalMass / totSoFar < 0.0001) {
      highDemand = upperLimit;
      break;
    }

    // it is not, so add it in
    totSoFar += additionalMass;

    counter++;
  }
}


// ----------------------------------------------------------- // 

/** VG Function class */
class BayesianDemand : public VGFunction {

private:

  /** RNG and structures go here. */
  gsl_rng *rng;
  DoubleGammaPDF dgp;
  bool active;
  bool sampled;
  bool haveP0shape, haveP0scale, haveD0shape, haveD0scale, haveCurDemand, haveCurPrice, haveNewPrice;
  double p0shape, p0scale, d0shape, d0scale, curPrice, newPrice;
  double curDemand;
  double outputValue;

  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  BayesianDemand() {

#ifdef DEBUG
    printf("[BayesianDemand]: constructor.\n");
    fflush(stdout);
#endif

    // get a RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~BayesianDemand() {

#ifdef DEBUG
    printf("[BayesianDemand]: destructor.\n");
    fflush(stdout);
#endif

    // destroy the RNG.
    gsl_rng_free(rng);	
  }

  /** Initializes the RNG seed for a given call. */
  void initializeSeed(long seedValue) {
#ifdef DEBUG
    printf("[BayesianDemand]: initializeSeed(%ld).\n", seedValue);
    fflush(stdout);
#endif

    // set the seed.
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

#ifdef DEBUG
    printf("[BayesianDemand]: finalizeTrial.\n");
    fflush(stdout);
#endif

    // set active
    active = (haveP0shape && haveP0scale && haveD0shape && haveD0scale && haveCurDemand && haveCurPrice && haveNewPrice);
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {


#ifdef DEBUG
    printf("[BayesianDemand]: clearParams.\n");
#endif

    active = false;
    sampled = false;
    haveP0shape = haveP0scale = haveD0shape = haveD0scale = haveCurDemand = haveCurPrice = haveNewPrice = false;
  }

	
  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {

#ifdef DEBUG
    printf("[BayesianDemand]: takeParams(...)\n");
    fflush(stdout);
#endif

    if (input.p0shape != NULL) {
      p0shape = *input.p0shape;
      haveP0shape = true;
    }

    if (input.p0scale != NULL) {
      p0scale = *input.p0scale;
      haveP0scale = true;
    }

    if (input.d0shape != NULL) {
      d0shape = *input.d0shape;
      haveD0shape = true;
    }

    if (input.d0scale != NULL) {
      d0scale = *input.d0scale;
      haveD0scale = true;
    }

    if (input.curDemand != NULL) {
      curDemand = *input.curDemand;
      haveCurDemand = true;
    }

    if (input.curPrice != NULL) {
      curPrice = *input.curPrice;
      haveCurPrice = true;
    }

    if (input.newPrice != NULL) {
      newPrice = *input.newPrice;
      haveNewPrice = true;
    }

    // set active
    active = (haveP0shape && haveP0scale && haveD0shape && haveD0scale && haveCurDemand && haveCurPrice && haveNewPrice);
    sampled = false;
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 

    // not active?
    if (!active) {
      return 0;
    }

    // already initialized?
    if (!sampled) {

      // do it
      dgp.Init(curPrice, curDemand, p0shape, d0shape, p0scale, d0scale);

      // set the flag for subsequent iterations
      sampled = true;
    }

    // sample a new random function
    double p0, d0;
    dgp.GetNewPAndD(p0, d0, rng);

    // see if the new price is too high
    double newDemand;
    if (newPrice > p0) {
      newDemand = 0;
    }
    else {
      newDemand = (1.0 - (newPrice / p0)) * d0;
    }

    outputValue = newDemand;
    output.newDemand = &outputValue;
    active = false;

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){7, {DOUBLE, DOUBLE, DOUBLE, DOUBLE, DOUBLE, DOUBLE, DOUBLE}, {"p0shape", "p0scale", "d0shape", "d0scale", "curDemand", "curPrice", "newPrice"}, {0, 0, 0, 0, 0, 0, 0}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {DOUBLE}, {"newDemand"}, {1}};
  }

  const char *getName() {
    return "BayesianDemand";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new BayesianDemand());
}

void destroy(VGFunction *vgFunction) {
  delete (BayesianDemand *)vgFunction;
}

