#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "VGFunction.h"


// uncomment for debug messages
//#define DEBUG

/**
 * VGF for a Bernoulli distribution.
 * Used by MCDB Q5 and Q6.
 *
 * Input schema: (p)
 * Dist: X ~ Bernoulli(p)
 * Output schema: (X in {0,1})
 */


/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *prob;

};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *value;

};

// ----------------------------------------------------------- // 

/** VG Function class */
class Bernoulli : public VGFunction {

private:
  /** RNG and structures go here. */
  gsl_rng *rng;
  double prob;
  bool active;
  bool haveProb;
  long outValue;


  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Bernoulli() {

#ifdef DEBUG
    printf("[Bernoulli]: constructor.\n");
    fflush(stdout);
#endif

    // get a RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Bernoulli() {
	
#ifdef DEBUG
    printf("[Bernoulli]: destructor.\n");
    fflush(stdout);
#endif

    // destroy the RNG.
    gsl_rng_free(rng);
  }

  /** Initializes the RNG seed for a given call. */
  void initializeSeed(long seedValue) {

#ifdef DEBUG
    printf("[Bernoulli]: initializeSeed(%ld).\n", seedValue);
    fflush(stdout);
#endif

    // set the seed.
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

#ifdef DEBUG
    printf("[Bernoulli]: finalizeTrial.\n");
    fflush(stdout);
#endif

    // set active
    active = haveProb;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("[Bernoulli]: clearParams.\n");
    fflush(stdout);
#endif

    // reset.
    active = false;
    haveProb = false;
  }

	
  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {

#ifdef DEBUG
    printf("[Bernoulli]: takeParams(");
    if (input.prob == NULL) {
      printf("NULL).\n");
    }
    else {
      printf("%f).\n", *input.prob);
    }
    fflush(stdout);
#endif

    if (input.prob != NULL) {
      prob = *input.prob;
      haveProb = true;
    }

    // set active
    active = haveProb;
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 

#ifdef DEBUG
    printf("[Bernoulli]: outputvals => ");
#endif

    // active?
    if (!active) {

#ifdef DEBUG
      printf("NULL.\n");
      fflush(stdout);
#endif

      return 0;
    }

    // get a sample
    outValue = gsl_ran_bernoulli(rng, prob);
    active = false;

    // leave.
    output.value = &outValue;

#ifdef DEBUG
    printf("%ld.\n", *output.value);
    fflush(stdout);
#endif

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){1, {DOUBLE}, {"p"}, {0}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {INT}, {"value"}, {1}};
  }

  const char *getName() {
    return "Bernoulli";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Bernoulli());
}

void destroy(VGFunction *vgFunction) {
  delete (Bernoulli *)vgFunction;
}

