#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#define INITIAL_SIZE 1

/**
 * A Categorical VG function.
 * Used by the HMM query.
 *
 * Inputs: (dim_i, C[i]) for i=1:K
 *
 * Parameterization: X ~ Categorical(C / SUM(C[i]))
 *
 * Output: (dim_i). If there are no input categories, -1 is returned.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *dimension_id;
  double *probability;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *dimension_id;
};


int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno)
{
  runningError = 1;
}


class Categorical: public VGFunction {
private:
  gsl_rng *rng;

  bool active;

  // for input parameters
  double *probability; /* value of pi vector for multinomial */

  // for output values
  long outputValue;
  unsigned int *outputNum; /* value of value vector for multinomial */

  // for temp
  long currentDimensionSize;
  long maxDimensionSize;

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Categorical() {

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);

    // set the initial size of each container as the INITIAL_SIZE.
    maxDimensionSize = INITIAL_SIZE;

    probability = (double *)malloc(maxDimensionSize * sizeof(double));
    outputNum = (unsigned int *)malloc(maxDimensionSize * sizeof(unsigned int));
		
    runningError = -1;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Categorical() {

    // destroy the RNG
    gsl_rng_free(rng);

    // destroy the arrays
    free(probability);
    free(outputNum);
  }

  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

    // set active again.
    active = true;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {
    currentDimensionSize = 0;
    active = true;
  }

  void takeParams(RecordIn &input) {

    // in case we found a bad input.
    if (currentDimensionSize < 0) {
      return;
    }

    if(input.dimension_id != NULL && input.probability != NULL) {
	long currentInputDimensionId = *input.dimension_id;
	
	if(currentDimensionSize < currentInputDimensionId + 1) {
	  currentDimensionSize = currentInputDimensionId + 1;
	}

	if(currentInputDimensionId >= maxDimensionSize) {
	  maxDimensionSize = (currentInputDimensionId + (currentInputDimensionId >> 1) + 1);
	  probability = (double *)realloc(probability, maxDimensionSize * sizeof(double));
	  outputNum = (unsigned int *)realloc(outputNum, maxDimensionSize * sizeof(unsigned int));
	}
	
	probability[currentInputDimensionId] = *input.probability;

	// if the value is wrong, get out of here.
	if (isnan(*input.probability) || isinf(*input.probability) || *input.probability < 0.0) {
	  currentDimensionSize = -1;
	  return;
	}
    }
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) {

    if (!active) {
      return 0;
    }

    outputValue = -1;

    if (currentDimensionSize > 0) {
      gsl_ran_multinomial(rng, 
			  currentDimensionSize, 
			  1,
			  probability,
			  outputNum);

      for (int i=0;i<currentDimensionSize;i++) {
	if (outputNum[i] > 0) {
	  outputValue = i;
	  break;
	}
      }
    }

    output.dimension_id = &outputValue;
    active = false;
    return 1;
  }

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){2, {INT, DOUBLE}, {"id", "probability"}, {0, 1}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {INT}, {"out_id"}, {1}};
  }

  const char *getName() {
    return "Categorical";
  }
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Categorical());
}

void destroy(VGFunction *vgFunction) {
  delete (Categorical *)vgFunction;
}
