#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#define INITIAL_SIZE 10

/**
 * A Categorical_Doc VG function.
 * Used by the HMM query.
 *
 * Inputs: (dim_i, C[i]) for i=1:K
 *
 * Parameterization: X ~ Categorical_Doc(C / SUM(C[i]))
 *
 * Output: (dim_i). If there are no input categories, -1 is returned.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *pair_posWord;
  long *pair_prevCurTopic;

  long *pair_fromToTopic;
  double *prob_trans;

  long *topic_start;
  double *prob_start;

  long *pair_emitWordTopic;
  double *prob_emit;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *out_prevpos;
  long *out_curpos;
  long *out_nextpos;
  long *out_prevtopic;
  long *out_id;
};


int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno)
{
  runningError = 1;
}


class Categorical_Doc: public VGFunction {
private:
  gsl_rng *rng;

  bool active;

  long docCapacity;
  long topicCapacity;
  long wordCapacity;

  // position-topic pairs. size: docCapacity.
  long *prevTopic;
  long *curTopic;
  long *curWord;
  long docLen;

  // transition matrix. size: topicCapacity
  double **trans;
  long numTopics;

  // start vector. size: topicCapacity
  double *start;

  // emission matrix. size: wordCapacity * topicCapacity
  double **emit;

  // --- temporaries ---
  long curPos;
  double *tempProb; // size: topicCapacity
  unsigned int *tempCat; // size: topicCapacity
  long outPrevPos;
  long outCurPos;
  long outNextPos;
  long outPrevTopic;
  long outCurTopic;

  void increaseDoc(long pos) {

    if (pos >= docLen) {
      docLen = pos + 1;
    }

    if (docLen >= docCapacity) {

      long oldCapacity = docCapacity;
      docCapacity = docLen + 1 + (docLen >> 1);

      prevTopic = (long *)realloc(prevTopic, sizeof(long) * docCapacity);
      curTopic = (long *)realloc(curTopic, sizeof(long) * docCapacity);
      curWord = (long *)realloc(curWord, sizeof(long) * docCapacity);

      for (long i=oldCapacity;i<docCapacity;i++) {
	prevTopic[i] = -1;
	curTopic[i] = -1;
	curWord[i] = -1;
      }
    }
  }

  void increaseTopics(long topicID) {

    if (topicID >= numTopics) {
      numTopics = topicID + 1;
    }

    if (numTopics >= topicCapacity) {

      long oldCapacity = topicCapacity;
      topicCapacity = numTopics + 1 + (numTopics >> 1);

      trans = (double **)realloc(trans, sizeof(double *) * topicCapacity);
      start = (double *)realloc(start, sizeof(double) * topicCapacity);
      tempProb = (double *)realloc(tempProb, sizeof(double) * topicCapacity);
      tempCat = (unsigned int *)realloc(tempCat, sizeof(unsigned int) * topicCapacity);

      for (long i=0;i<topicCapacity;i++) {

	if (i < oldCapacity) {
	  trans[i] = (double *)realloc(trans[i], sizeof(double) * topicCapacity);
	} else {
	  start[i] = 0.0;
	  trans[i] = (double *)malloc(sizeof(double) * topicCapacity);
	  
	  for (long j=0;j<topicCapacity;j++) {
	    trans[i][j] = 0.0;
	  }
	}
      }

      for (long i=0;i<wordCapacity;i++) {
	emit[i] = (double *)realloc(emit[i], sizeof(double) * topicCapacity);

	for (long j=oldCapacity;j<topicCapacity;j++) {
	  emit[i][j] = 0.0;
	}
      }
    }
  }

  void increaseWords(long wordID) {
    
    if (wordID >= wordCapacity) {
      
      long oldCapacity = wordCapacity;
      wordCapacity = wordID + 1 + (wordID >> 1);
      
      emit = (double **)realloc(emit, sizeof(double*) * wordCapacity);
      
      for (long i=oldCapacity;i<wordCapacity;i++) {
	emit[i] = (double *)malloc(sizeof(double) * topicCapacity);
	for (long j=0;j<topicCapacity;j++) {
	  emit[i][j] = 0.0;
	}
      }
    }
  }


  int unpair1(long val) {
    return (int)(val >> 32);
  }

  int unpair2(long val) {
    return (int)(val);
  }

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Categorical_Doc() {

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);

    // set the initial size of each container as the INITIAL_SIZE.
    docCapacity = INITIAL_SIZE;
    topicCapacity = INITIAL_SIZE;
    wordCapacity = INITIAL_SIZE;

    prevTopic = (long *)malloc(sizeof(long) * docCapacity);
    curTopic = (long *)malloc(sizeof(long) * docCapacity);
    curWord = (long *)malloc(sizeof(long) * docCapacity);
    trans = (double **)malloc(sizeof(double *) * topicCapacity);
    start = (double *)malloc(sizeof(double) * topicCapacity);
    emit = (double **)malloc(sizeof(double *) * wordCapacity);
    tempProb = (double *)malloc(sizeof(double) * topicCapacity);
    tempCat = (unsigned int *)malloc(sizeof(unsigned int) * topicCapacity);

    for (long i=0;i<topicCapacity;i++) {
      trans[i] = (double *)malloc(sizeof(double) * topicCapacity);
    }

    for (long i=0;i<wordCapacity;i++) {
      emit[i] = (double *)malloc(sizeof(double) * topicCapacity);
    }

    runningError = -1;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Categorical_Doc() {

    // destroy the RNG
    gsl_rng_free(rng);

    // destroy the arrays
    free(prevTopic);
    free(curTopic);
    free(curWord);
    free(start);

    for (long i=0;i<topicCapacity;i++) {
      free(trans[i]);
    }

    free(trans);

    for (long i=0;i<wordCapacity;i++) {
      free(emit[i]);
    }

    free(emit);
    free(tempProb);
    free(tempCat);
  }

  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

    // set active again.
    active = (docLen > 0 && numTopics > 0);
    if (active) {
      curPos = curTopic[0] < 0 ? 1 : 0;
    }
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {
    active = false;

    for (long i=0;i<docCapacity;i++) {
      prevTopic[i] = -1;
      curTopic[i] = -1;
      curWord[i] = -1;
    }

    for (long i=0;i<wordCapacity;i++) {
      for (long j=0;j<topicCapacity;j++) {
	emit[i][j] = 0.0;
      }
    }

    for (long i=0;i<topicCapacity;i++) {
      start[i] = 0.0;

      for (long j=0;j<topicCapacity;j++) {
	trans[i][j] = 0.0;
      }
    }

    docLen = 0;
    numTopics = 0;
    curPos = 0;
  }

  void takeParams(RecordIn &input) {

    // position-word topic pairs.
    if (input.pair_posWord != NULL && input.pair_prevCurTopic != NULL) {
      
      // unpack the first pair
      long posV = unpair1(*input.pair_posWord);
      long wordV = unpair2(*input.pair_posWord);

      // unpack the second pair
      long prevTopicV = unpair1(*input.pair_prevCurTopic);
      long curTopicV = unpair2(*input.pair_prevCurTopic);

      // increase the structures
      increaseDoc(posV);
      increaseWords(wordV);
      increaseTopics(curTopicV);
      increaseTopics(prevTopicV);

      // and update the structures.
      prevTopic[posV] = prevTopicV;
      curTopic[posV] = curTopicV;
      curWord[posV] = wordV;
    }

    // transition probabilities.
    if (input.pair_fromToTopic != NULL && input.prob_trans != NULL) {

      // unpack the pair
      long fromTopic = unpair1(*input.pair_fromToTopic);
      long toTopic = unpair2(*input.pair_fromToTopic);

      // incrase the structures
      increaseTopics(fromTopic);
      increaseTopics(toTopic);

      // update them
      trans[fromTopic][toTopic] = *input.prob_trans;
    }

    // start vector
    if (input.topic_start != NULL && input.prob_start != NULL) {

      increaseTopics(*input.topic_start);
      start[*input.topic_start] = *input.prob_start;
    }

    // emission probabilities.
    if (input.pair_emitWordTopic != NULL && input.prob_emit != NULL) {

      // unpack the pair
      long emitWord = unpair1(*input.pair_emitWordTopic);
      long emitTopic = unpair2(*input.pair_emitWordTopic);

      // increase the structures
      increaseWords(emitWord);
      increaseTopics(emitTopic);

      // update them
      emit[emitWord][emitTopic] = *input.prob_emit;
    }

    active = (docLen > 0 && numTopics > 0);
    if (active) {
      curPos = curTopic[0] < 0 ? 1 : 0;
    }
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) {

    if (!active) {
      return 0;
    }

    // set up tempProb.
    for (long i=0;i<numTopics;i++) {

      // all of them have emission probabilities.
      tempProb[i] = emit[curWord[curPos]][i];

      // first word?
      if (curPos == 0) {

	// start transition and transition *TO* next topic.
	tempProb[i] *= start[i] * trans[i][prevTopic[2 % docLen]];
      } 

      // last word?
      else if (curPos + 1 >= docLen) {

	// transition *FROM* previous topic.
	tempProb[i] *= trans[prevTopic[curPos]][i];
      }

      // everything else in between
      else {
	tempProb[i] *= trans[prevTopic[curPos]][i] * trans[i][prevTopic[(curPos + 2) % docLen]];
      }
    }

    // we are ready to sample.
    gsl_ran_multinomial(rng, numTopics, 1, tempProb, tempCat);

    // get the resulting category
    long outCat = -1;
    for (long i=0;i<numTopics;i++) {
      if (tempCat[i] > 0) {
	outCat = i;
	break;
      }
    }

    // then, we prepare the output record.
    outPrevPos = curPos;
    outCurPos = (curPos + 1) % docLen;
    outNextPos = (curPos + 2) % docLen;
    outPrevTopic = outCat;
    outCurTopic = prevTopic[outNextPos];

    output.out_prevpos = &outPrevPos;
    output.out_curpos = &outCurPos;
    output.out_nextpos = &outNextPos;
    output.out_prevtopic = &outPrevTopic;
    output.out_id = &outCurTopic;

    // move to the next position
    curPos += 2;
    if (curPos >= docLen) {
      active = false;
    }

    return 1;
  }

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){8, {INT, INT, INT, DOUBLE, INT, DOUBLE, INT, DOUBLE}, {"pair_posWord", "pair_prevCurTopic", "pair_fromToTopic", "prob_trans", "topic_start", "prob_start", "pair_emitWordTopic", "prob_emit"}, {0, 0, 0, 1, 0, 1, 0, 1}};
  }

  VGSchema outputSchema() {

    return (VGSchema){5, {INT, INT, INT, INT, INT}, {"out_prevpos", "out_curpos", "out_nextpos", "out_prevtopic", "out_id"}, {0, 0, 0, 0, 1}};
  }

  const char *getName() {
    return "Categorical_Doc";
  }
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Categorical_Doc());
}

void destroy(VGFunction *vgFunction) {
  delete (Categorical_Doc *)vgFunction;
}
