#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#define INITIAL_SIZE 10

/**
 * A Categorical_Group VG function.
 * Used by the HMM query.
 *
 * Inputs: (dim_i, C[i]) for i=1:K
 *
 * Parameterization: X ~ Categorical_Group(C / SUM(C[i]))
 *
 * Output: (dim_i). If there are no input categories, -1 is returned.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *pair_docIDPos;
  long *pair_prevCurTopic;

  long *docID;
  long *pair_posWordID;

  long *pair_fromToTopic;
  double *prob_topic;

  long *topic_start;
  double *prob_start;

  long *pair_emitWordTopic;
  double *prob_emit;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *out_docid;
  long *out_prevpos;
  long *out_curpos;
  long *out_nextpos;
  long *out_prevtopic;
  long *out_curtopic;
};

struct docPos {
  long wordID;
  long prevTopicID;
  long curTopicID;
};

struct doc {
  bool valid;
  long capacity;
  long length;

  // size: capacity.
  docPos *words;
};


int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno)
{
  runningError = 1;
}


class Categorical_Group: public VGFunction {
private:
  gsl_rng *rng;

  bool active;

  long docCapacity;
  long topicCapacity;
  long wordCapacity;

  // documents. size: docCapacity.
  doc *docs;
  long numDocs;

  // transition matrix. size: topicCapacity
  double **trans;
  long numTopics;

  // start vector. size: topicCapacity
  double *start;

  // emission matrix. size: wordCapacity * topicCapacity
  double **emit;

  // --- temporaries ---
  long curDoc;
  long curPos;
  double *tempProb; // size: topicCapacity
  unsigned int *tempCat; // size: topicCapacity
  long outDoc;
  long outPrevPos;
  long outCurPos;
  long outNextPos;
  long outPrevTopic;
  long outCurTopic;

  void increaseDocs(long docID) {

    if (docID >= numDocs) {
      numDocs = docID + 1;
    }

    if (numDocs >= docCapacity) {
      long oldCapacity = docCapacity;
      docCapacity = numDocs + 1 + (numDocs >> 1);

      docs = (doc *)realloc(docs, sizeof(doc) * docCapacity);
      
      for (long i=oldCapacity;i<docCapacity;i++) {
	docs[i].valid = false;
	docs[i].length = 0;
	docs[i].capacity = INITIAL_SIZE;
	docs[i].words = (docPos *)malloc(sizeof(docPos) * docs[i].capacity);

	for (long j=0;j<docs[i].capacity;j++) {
	  docs[i].words[j].wordID = -1;
	  docs[i].words[j].prevTopicID = -1;
	  docs[i].words[j].curTopicID = -1;
	}
      }      
    }
  }

  void increaseDoc(long docID, long pos) {

    if (pos >= docs[docID].length) {
      docs[docID].length = pos + 1;
    }

    if (docs[docID].length >= docs[docID].capacity) {

      long oldCapacity = docs[docID].capacity;
      docs[docID].capacity = docs[docID].length + 1 + (docs[docID].length >> 1);

      docs[docID].words = (docPos *)realloc(docs[docID].words, sizeof(docPos) * docs[docID].capacity);

      for (long i=oldCapacity;i<docs[docID].capacity;i++) {
	docs[docID].words[i].wordID = -1;
	docs[docID].words[i].prevTopicID = -1;
	docs[docID].words[i].curTopicID = -1;
      }
    }
  }

  void increaseTopics(long topicID) {

    if (topicID >= numTopics) {
      numTopics = topicID + 1;
    }

    if (numTopics >= topicCapacity) {

      long oldCapacity = topicCapacity;
      topicCapacity = numTopics + 1 + (numTopics >> 1);

      trans = (double **)realloc(trans, sizeof(double *) * topicCapacity);
      start = (double *)realloc(start, sizeof(double) * topicCapacity);
      tempProb = (double *)realloc(tempProb, sizeof(double) * topicCapacity);
      tempCat = (unsigned int *)realloc(tempCat, sizeof(unsigned int) * topicCapacity);

      for (long i=0;i<topicCapacity;i++) {

	if (i < oldCapacity) {
	  trans[i] = (double *)realloc(trans[i], sizeof(double) * topicCapacity);
	} else {
	  start[i] = 0.0;
	  trans[i] = (double *)malloc(sizeof(double) * topicCapacity);
	  
	  for (long j=0;j<topicCapacity;j++) {
	    trans[i][j] = 0.0;
	  }
	}
      }

      for (long i=0;i<wordCapacity;i++) {
	emit[i] = (double *)realloc(emit[i], sizeof(double) * topicCapacity);

	for (long j=oldCapacity;j<topicCapacity;j++) {
	  emit[i][j] = 0.0;
	}
      }
    }
  }

  void increaseWords(long wordID) {
    
    if (wordID >= wordCapacity) {
      
      long oldCapacity = wordCapacity;
      wordCapacity = wordID + 1 + (wordID >> 1);
      
      emit = (double **)realloc(emit, sizeof(double*) * wordCapacity);
      
      for (long i=oldCapacity;i<wordCapacity;i++) {
	emit[i] = (double *)malloc(sizeof(double) * topicCapacity);
	for (long j=0;j<topicCapacity;j++) {
	  emit[i][j] = 0.0;
	}
      }
    }
  }


  int unpair1(long val) {
    return (int)(val >> 32);
  }

  int unpair2(long val) {
    return (int)(val);
  }

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Categorical_Group() {

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);

    // set the initial size of each container as the INITIAL_SIZE.
    docCapacity = INITIAL_SIZE;
    topicCapacity = INITIAL_SIZE;
    wordCapacity = INITIAL_SIZE;

    docs = (doc *)malloc(sizeof(doc) * docCapacity);
    for (long i=0;i<docCapacity;i++) {
      docs[i].valid = false;
      docs[i].length = 0;
      docs[i].capacity = INITIAL_SIZE;     
      docs[i].words = (docPos *)malloc(sizeof(docPos) * docs[i].capacity);
    }

    trans = (double **)malloc(sizeof(double *) * topicCapacity);
    start = (double *)malloc(sizeof(double) * topicCapacity);
    emit = (double **)malloc(sizeof(double *) * wordCapacity);
    tempProb = (double *)malloc(sizeof(double) * topicCapacity);
    tempCat = (unsigned int *)malloc(sizeof(unsigned int) * topicCapacity);

    for (long i=0;i<topicCapacity;i++) {
      trans[i] = (double *)malloc(sizeof(double) * topicCapacity);
    }

    for (long i=0;i<wordCapacity;i++) {
      emit[i] = (double *)malloc(sizeof(double) * topicCapacity);
    }

    runningError = -1;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Categorical_Group() {

    // destroy the RNG
    gsl_rng_free(rng);

    // destroy the arrays
    free(start);

    for (long i=0;i<topicCapacity;i++) {
      free(trans[i]);
    }

    free(trans);

    for (long i=0;i<wordCapacity;i++) {
      free(emit[i]);
    }

    for (long i=0;i<docCapacity;i++) {
      free(docs[i].words);
    }

    free(docs);
    free(emit);
    free(tempProb);
    free(tempCat);
  }

  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

    // set active again.
    active = (numDocs > 0 && numTopics > 0);
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {
    active = false;

    for (long i=0;i<docCapacity;i++) {

      docs[i].valid = false;
      docs[i].length = 0;
      
      for (long j=0;j<docs[i].capacity;j++) {
	docs[i].words[j].prevTopicID = -1;
	docs[i].words[j].curTopicID = -1;
	docs[i].words[j].wordID = -1;
      }
    }

    for (long i=0;i<wordCapacity;i++) {
      for (long j=0;j<topicCapacity;j++) {
	emit[i][j] = 0.0;
      }
    }

    for (long i=0;i<topicCapacity;i++) {
      start[i] = 0.0;

      for (long j=0;j<topicCapacity;j++) {
	trans[i][j] = 0.0;
      }
    }

    numDocs = 0;
    numTopics = 0;
    curDoc = 0;
    curPos = 0;
  }

  void takeParams(RecordIn &input) {

    if (input.pair_docIDPos != NULL && input.pair_prevCurTopic != NULL) {

      // unpack the pairs
      long docID = unpair1(*input.pair_docIDPos);
      long pos   = unpair2(*input.pair_docIDPos);

      long prevTopicID = unpair1(*input.pair_prevCurTopic);
      long curTopicID  = unpair2(*input.pair_prevCurTopic);

      // increase the structures
      increaseDocs(docID);
      increaseDoc(docID, pos);
      increaseTopics(prevTopicID);
      increaseTopics(curTopicID);

      // and set.
      docs[docID].valid = true;
      docs[docID].words[pos].prevTopicID = prevTopicID;
      docs[docID].words[pos].curTopicID = curTopicID;     
    }

    if (input.docID != NULL && input.pair_posWordID != NULL) {
      
      // unpack the pair
      long pos    = unpair1(*input.pair_posWordID);
      long wordID = unpair2(*input.pair_posWordID);

      // increase the structures.
      increaseDocs(*input.docID);
      increaseDoc(*input.docID, pos);
      increaseWords(wordID);

      // set.
      docs[*input.docID].words[pos].wordID = wordID;
    }

    if (input.pair_fromToTopic != NULL && input.prob_topic != NULL) {

      // unpack the pair.
      long fromTopicID = unpair1(*input.pair_fromToTopic);
      long toTopicID   = unpair2(*input.pair_fromToTopic);

      // increase the structures.
      increaseTopics(fromTopicID);
      increaseTopics(toTopicID);

      trans[fromTopicID][toTopicID] = *input.prob_topic;
    }

    if (input.topic_start != NULL && input.prob_start != NULL) {
      increaseTopics(*input.topic_start);
      start[*input.topic_start] = *input.prob_start;
    }

    if (input.pair_emitWordTopic != NULL && input.prob_emit != NULL) {
      
      // unpack the pair.
      long wordID  = unpair1(*input.pair_emitWordTopic);
      long topicID = unpair2(*input.pair_emitWordTopic);

      // increase the structures
      increaseWords(wordID);
      increaseTopics(topicID);

      // set
      emit[wordID][topicID] = *input.prob_emit;
    }

    active = (numDocs > 0 && numTopics > 0);
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) {

    // not active? leave.
    if (!active) {
      return 0;
    }

    while (true) {
      
      // done with documents?
      if (curDoc >= numDocs) {
	active = false;
	return 0;
      }

      // is the current doc. valid?
      // or: are we done with the words in this document?
      if (!docs[curDoc].valid || curPos >= docs[curDoc].length) {
	curPos = 0;
	curDoc ++;
	continue;
      }

      // is the current position valid?
      if (docs[curDoc].words[curPos].curTopicID < 0) {
	curPos++;
	continue;
      }

      break;
    }

    // everything is fine: we will sample.
    // set up tempProb.
    //    printf("sample doc: %ld / pos: %ld / len: %ld\n", curDoc, curPos, docs[curDoc].length);
    //    printf("word id: %ld / prev: %ld / cur: %ld\n", docs[curDoc].words[curPos].wordID, docs[curDoc].words[curPos].prevTopicID, docs[curDoc].words[curPos].curTopicID);     
    //    fflush(stdout);

    for (long i=0;i<numTopics;i++) {
      fflush(stdout);
      tempProb[i] = emit[docs[curDoc].words[curPos].wordID][i];

      // is this the first word?
      if (curPos == 0) {

	// starting topic * transition TO the next topic.
	tempProb[i] *= start[i] * trans[i][docs[curDoc].words[2 % docs[curDoc].length].prevTopicID];
      }

      // is this the last word?
      else if (curPos + 1 >= docs[curDoc].length) {

	// transition *from* the previous topic at the last position.
	tempProb[i] *= trans[docs[curDoc].words[curPos].prevTopicID][i];
      }

      // otherwise, any other word in between.
      else {
	tempProb[i] *= trans[docs[curDoc].words[curPos].prevTopicID][i] * trans[i][docs[curDoc].words[(curPos + 2) % docs[curDoc].length].prevTopicID];
      }
    }

    // we are ready to sample.
    gsl_ran_multinomial(rng, numTopics, 1, tempProb, tempCat);
      
    // get the resulting category
    long outCat = -1;
    for (long i=0;i<numTopics;i++) {
      if (tempCat[i] > 0) {
	outCat = i;
	break;
      }
    }

    // prepare the output record.
    outDoc = curDoc;
    outPrevPos = curPos;
    outCurPos = (curPos + 1) % docs[curDoc].length;
    outNextPos = (curPos + 2) % docs[curDoc].length;
    outPrevTopic = outCat;
    outCurTopic = docs[curDoc].words[outNextPos].prevTopicID;
      
    output.out_docid = &outDoc;
    output.out_prevpos = &outPrevPos;
    output.out_curpos = &outCurPos;
    output.out_nextpos = &outNextPos;
    output.out_prevtopic = &outPrevTopic;
    output.out_curtopic = &outCurTopic;
      
    // move to the next position
    curPos += 2;
      
    // leave with a sample.
    return 1;      
    
  }

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){10, {INT, INT, INT, INT, INT, DOUBLE, INT, DOUBLE, INT, DOUBLE}, {"pair_docIDPos", "pair_prevCurTopic", "in_docID", "pair_posWordID", "pair_fromToTopic", "prob_trans", "topic_start", "prob_start", "pair_emitWordTopic", "prob_emit"}, {0, 1, 0, 0, 0, 1, 0, 1, 0, 1}};
  }

  VGSchema outputSchema() {

    return (VGSchema){6, {INT, INT, INT, INT, INT, INT}, {"out_docid", "out_prevpos", "out_curpos", "out_nextpos", "out_prevtopic", "out_curtopic"}, {0, 0, 0, 0, 0, 1}};
  }

  const char *getName() {
    return "Categorical_Group";
  }
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Categorical_Group());
}

void destroy(VGFunction *vgFunction) {
  delete (Categorical_Group *)vgFunction;
}
