#include <stdio.h>
#include "VGFunction.h"
#include "lib/Gaussian.h"
#include "lib/Gaussian.cc"

/**
 * A Multivariate Normal VGFunction with conditioning. It allows for
 * adding observed values on some of the dimensions, so that the mean
 * and variance are re-computed accordingly. The output is the same,
 * with the conditioned values being returned unchanged.
 *
 *
 * Inputs: (mdim_i, M[i], NULL, NULL, NULL, NULL, NULL) for i=1:D
 *         (NULL, NULL, cdim_i, cdim_j, C[i,j], NULL, NULL) for i,j=1:D
 *         (NULL, NULL, NULL, NULL, NULL, odim_i, O[i]) for i=1:D
 *
 * Parameterization: X ~ Normal(M,C|O)
 *
 * Outputs: (dim_i, X[i]) for i=1:D
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *rv_mean;
  double *mean;

  long *rv_i;
  long *rv_j;
  double *covar;

  long *rv_obs;
  double *obs;

};

/**
 * Output record, used by the outputVals() method.
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *rv;
  double *value;

};

// ----------------------------------------------------------- // 

/** VG Function class */
class ConditionalNormal : public VGFunction {

private:
  gsl_rng *r;
  Gaussian *g;
  double *samples;
  int dim;
  int curIndex;
  long outIndex;
  double outValue;


  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  ConditionalNormal() {
    r = gsl_rng_alloc(gsl_rng_default);
    g = new Gaussian();
    samples = NULL;
    dim = 0;
    curIndex = 0;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~ConditionalNormal() {
    gsl_rng_free(r);
    delete g;
    if (samples != NULL) {
      free(samples);     
    }
    dim = 0;
  }

  /** Initializes the RNG seed for a given call. */
  void initializeSeed(long seedValue) {
    gsl_rng_set(r, seedValue);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {
    curIndex = 0;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {    
    curIndex = 0;
    g->ClearParameters(true);
  }

	
  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {
    
    // set mean.
    if (input.rv_mean != NULL && input.mean != NULL) {
      g->SetMean(*input.rv_mean, *input.mean);
    }

    // set covariance.
    if (input.rv_i != NULL && input.rv_j != NULL && input.covar != NULL) {
      g->SetCovariance(*input.rv_i, *input.rv_j, *input.covar);
    }

    // finally, set the conditionings.
    if (input.rv_obs != NULL && input.obs != NULL) {
      int zzz = g->Condition(*input.rv_obs, *input.obs);

      /***
      printf("conditioning: %ld %f %d\n", *input.rv_obs, *input.obs, zzz);
      g->Print();
      printf("-------\n");
      ***/
    }
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 

    // have we already sampled everything?
    if (curIndex == g->GetNumRVs()) {
      return 0;
    }

    // do we need to sample?
    if (curIndex == 0) {

      // sample, and check for failures.
      if (dim != g->GetNumRVs()) {
	dim = g->GetNumRVs();
	samples = (double *) realloc(samples, sizeof(double) * dim);
      }

      /***
      printf("==> SAMPLING:\n");
      g->Print();
      ***/

      if (g->Sample(samples, r) == 0) {
	return 0;
      }
    }

    // at this point, we just need to set values.
    outIndex = curIndex;
    outValue = samples[curIndex];

    // and get out.
    output.rv = &outIndex;
    output.value = &outValue;

    curIndex++;
    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){7, {INT, DOUBLE, INT, INT, DOUBLE, INT, DOUBLE}, {"rv_mean", "mean", "rv_i", "rv_j", "covar", "rv_obs", "obs"}, {0, 1, 0, 0, 1, 0, 1}};
  }

  VGSchema outputSchema() {

    return (VGSchema){2, {INT, DOUBLE}, {"out_dim", "out_value"}, {0, 1}};
  }

  const char *getName() {
    return "ConditionalNormal";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new ConditionalNormal());
}

void destroy(VGFunction *vgFunction) {
  delete (ConditionalNormal *)vgFunction;
}

