#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "VGFunction.h"
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

//#define DEBUG
#define INITIAL_SIZE 1

using namespace std;

/**
 * Dirichlet conjugate VG function for the 'measurement' GMM.
 *
 * Used by the GMM and LDA queries.
 *
 * Input: (i, A[i], NULL, NULL) for i=1:K
 *        (NULL, NULL, i, B[i]) for i=1:K
 * Parameterization: X[i] ~ Dirichlet_Conjugate(A[i] + B[i], ..., A[k] + B[k])
 * Output: (i, X[i]) for i=1:K
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *inputDimensionIndex;
  double *inputDimensionValue;

  long *priorInputDimensionIndex;
  double *priorInputDimensionValue;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *ouputDimensionIndex;
  double *ouputDimensionValue;
};

/** Error handler for GSL. */
int runningError;
void handler(const char *reason, const char *file, int line, int gsl_errno) {
  runningError = 1;
  stringstream ss;
  string result = "reason: Dirichlet_Conjugate";
  result += reason;
  result += "\t file:";
  result += file;
  result += "\t line:";
  ss.str("");
  ss << line;
  result += ss.str();
  result += "\t gsl_errno:";
  ss.str("");
  ss << gsl_errno;
  result += ss.str();
  result += "\r\n";

  /***

  ofstream fileWriter;

  fileWriter.open("/tmp/error.report", ios::trunc);

  fileWriter<< result.c_str();
  fileWriter.close();
  ***/
  cerr << result;
  cerr << "\n\n";
  /***
  cerr << "dim: " << currentDimensionSize << "\n";
  cerr << "vals: ";

  for (int i=0;i<currentDimensionSize;i++) {
    cerr << input_prior[i] << " ";
  }

  cerr << "\n";
  ***/
}

// ----------------------------------------------------------- // 

/** VG Function class */
class Dirichlet_Conjugate : public VGFunction {

private:

  /** RNG and structures go here. */
  gsl_rng *rng;
	
  bool active;

  double *input_prior;
  long maxDimensionSize;
  long currentDimensionSize;

  double *outputValues;
  long outputIndex;
	
  // -----------------------------temp value------------------------------ // 
  long tempOutputValue;	

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Dirichlet_Conjugate() {
    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);

    // create the arrays
    maxDimensionSize = INITIAL_SIZE;
		
    input_prior = (double *)malloc(maxDimensionSize * sizeof(double));
    outputValues = (double *)malloc(maxDimensionSize * sizeof(double));

    for (int i=0;i<maxDimensionSize;i++) {
      input_prior[i] = 0.0;
      outputValues[i] = 0.0;
    }

    currentDimensionSize = 0;
    outputIndex = 0;
    runningError = -1;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Dirichlet_Conjugate() {

    // destroy the RNG
    gsl_rng_free(rng);

    // destroy the arrays
    free(input_prior);
    free(outputValues);
  }

  /** Initializes the RNG seed for a given call. */
  void initializeSeed(long seedValue) {

    // set the seed
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

    // set active again.
    active = (currentDimensionSize > 0);
    outputIndex = 0;
    runningError = -1;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {
    active = false;
    currentDimensionSize = 0;
    outputIndex = 0;
    runningError = -1;
    for (int i=0;i<maxDimensionSize;i++) {
      input_prior[i] = 0.0;
      outputValues[i] = 0.0;
    }
  }

  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {

    // get the current RV; these two must be together.
    if (input.inputDimensionIndex != NULL && 
	input.inputDimensionValue != NULL) {
      
      dirichletSet(*input.inputDimensionIndex, *input.inputDimensionValue);
    }

    if (input.priorInputDimensionIndex != NULL &&
	input.priorInputDimensionValue != NULL) {

      dirichletSet(*input.priorInputDimensionIndex, *input.priorInputDimensionValue);
    }

    // are we active?
    active = (currentDimensionSize > 0);
  }

  void dirichletSet(long inputDimensionIndex, double inputDimensionValue) {

    long tempValue = inputDimensionIndex;

    // check if we need to resize
    if (currentDimensionSize < tempValue + 1) {
      currentDimensionSize = tempValue + 1;
    }
    
    if(inputDimensionIndex >= maxDimensionSize) {
      long oldDimensionSize = maxDimensionSize;
      maxDimensionSize = (tempValue + (tempValue >> 1) + 1);

      input_prior = (double *)realloc(input_prior, maxDimensionSize * sizeof(double));
      outputValues = (double *)realloc(outputValues, maxDimensionSize * sizeof(double));

      for (long i=oldDimensionSize;i<maxDimensionSize;i++) {
	input_prior[i] = 0.0;
	outputValues[i] = 0.0;
      }
    }

    input_prior[tempValue] += inputDimensionValue;
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 
    // are we active?
    if (!active) {
      return 0;
    }

    // is this the first RV? if so, we'll have to sample
    if (outputIndex == 0) {

      runningError = -1;
      gsl_error_handler_t *oldHandler = gsl_set_error_handler (&handler);
      gsl_ran_dirichlet(rng, currentDimensionSize, input_prior, outputValues);
      gsl_set_error_handler(oldHandler);

      // did we get an error?
      if (runningError == 1) {
	fprintf(stderr, "[Dirichlet_Conjugate]: had a GSL error with the following group:\n");
	for (int i = 0; i < currentDimensionSize; i++) {
	  fprintf(stderr, "(%d, %f) ", i, input_prior[i]);
	}

	fprintf(stderr, "\n. Defaulting to uniform values...");

	double rnd_val = (currentDimensionSize * gsl_rng_uniform(rng));
	for (int i = 0; i < currentDimensionSize; i++) {
	  outputValues[i] = rnd_val;
	}

	runningError = 0;
      }
    }

    // then, pass the values.
    tempOutputValue = outputIndex;
    output.ouputDimensionIndex = &tempOutputValue;
    output.ouputDimensionValue = &(outputValues[tempOutputValue]);

    // go to the next RV...
    outputIndex++;

    // have we reached the last RV?
    if (outputIndex >= currentDimensionSize) {
      active = false;
    }

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){4, {INT, DOUBLE, INT, DOUBLE}, {"in_id", "priorVal", "in_postId", "posteriorVal"}, {0, 1, 0, 1}};
  }

  VGSchema outputSchema() {

    return (VGSchema){2, {INT, DOUBLE}, {"out_id", "probability"}, {0, 1}};
  }

  const char *getName() {
    return "Dirichlet_Conjugate";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Dirichlet_Conjugate());
}

void destroy(VGFunction *vgFunction) {
  delete (Dirichlet_Conjugate *)vgFunction;
}

