#include <stdio.h>
#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

// #define DEBUG

/**
 * VGF for a discretized Gamma distribution.
 * Used by MCDB Q2.
 *
 * Input schema: (mean, variance)
 * Parameterization: X ~ Gamma(mean^2 / variance, variance / mean)
 * Output schema: (rounded-up X)
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *mean;
  double *variance;

};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *value;

};

// ----------------------------------------------------------- // 

/** VG Function class */
class DiscreteGamma : public VGFunction {

private:
  /** RNG and structures go here. */
  gsl_rng *rng;
  double mean;
  double variance;
  long outputVal;
  bool haveMean;
  bool haveVariance;
  bool active;
  

  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  DiscreteGamma() {

#ifdef DEBUG
    printf("[DiscreteGamma]: constructor.\n");
    fflush(stdout);
#endif

    // get a RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~DiscreteGamma() {

#ifdef DEBUG
    printf("[DiscreteGamma]: destructor.\n");
    fflush(stdout);
#endif

    // destroy the RNG.
    gsl_rng_free(rng);	
  }

  /** Initializes the RNG seed for every group. */
  void initializeSeed(long seedValue) {

#ifdef DEBUG
    printf("[DiscreteGamma]: initializeSeed(%ld).\n", seedValue);
    fflush(stdout);
#endif

    // set the seed.
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current group and prepares the structures for
   * the next call to outputVals(). */
  void finalizeTrial() {

#ifdef DEBUG
    printf("[DiscreteGamma]: finalizeTrial.\n");
    fflush(stdout);
#endif

    // set active
    active = (haveMean && haveVariance);
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("[DiscreteGamma]: clearParams.\n");
#endif

    active = false;
    haveMean = false;
    haveVariance = false;
  }

  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {


#ifdef DEBUG
    printf("[DiscreteGamma]: takeParams(");
    if (input.mean == NULL)
      printf("NULL, ");
    else
      printf("%f, ", *input.mean);

    if (input.variance == NULL)
      printf("NULL).\n");
    else
      printf("%f).\n", *input.variance);

    fflush(stdout);
#endif

    // set the parameters
    if (input.mean != NULL) {
      mean = *input.mean;
      haveMean = true;
    }

    if (input.variance != NULL) {
      variance = *input.variance;
      haveVariance = true;
    }


    // set active
    active = (haveMean && haveVariance);
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 

#ifdef DEBUG
    printf("[DiscreteGamma]: outputVals(%f,%f) => ", mean, variance);
#endif
    
    // active?
    if (!active) {

#ifdef DEBUG
      printf("NULL.\n");
      fflush(stdout);
#endif

      return 0;
    }

    // sample
    outputVal = (long)gsl_ran_gamma(rng, mean*mean/variance, variance/mean);
    output.value = &outputVal;
    active = false;

#ifdef DEBUG
    printf("%f\n", *output.value);
    fflush(stdout);
#endif

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){2, {DOUBLE, DOUBLE}, {"mean", "variance"}, {0, 0}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {INT}, {"value"}, {1}};
  }

  const char *getName() {
    return "DiscreteGamma";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new DiscreteGamma());
}

void destroy(VGFunction *vgFunction) {
  delete (DiscreteGamma *)vgFunction;
}

