#include "VGFunction.h"
#include <string.h>
#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

#define INITIAL_SIZE 5

using namespace std;

/** BLOCK BASED GMM. */


/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *cluster_cid;
  double *cluster_prob;

  long *mean_cidDimPair;
  double *mean_value;

  long *cov_cid;
  long *cov_dimPair;
  double *cov_value;

  long *data_didDimPair;
  double *data_value;
};


/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *out_eType;
  long *out_cluster;
  long *out_dim1;
  long *out_dim2;
  double *out_val;
};


int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno) {
  runningError = 1;
  stringstream ss;
  string result = "reason: GMMBlock";
  result += reason;
  result += "\t file:";
  result += file;
  result += "\t line:";
  ss.str("");
  ss << line;
  result += ss.str();
  result += "\t gsl_errno:";
  ss.str("");
  ss << gsl_errno;
  result += ss.str();
  result += "\r\n";
  cerr << result;
}

class GMMBlock: public VGFunction {
private:  
  gsl_rng *rng;

  long numData;
  long numClusters;
  long numDimensions;

  long dataCapacity;
  long clusterCapacity;
  long dimensionCapacity;

  // cluster probabilities, size: clusterCapacity.
  double *clusterProb;

  // cluster means, size: clusterCapacity * dimensionCapacity.
  double **clusterMean;
  
  // cluster covariance, size: clusterCapacity * (dimensionCapacity * dimensionCapacity).
  double **clusterCov;

  // data points, size: dataCapacity * dimensionCapacity.
  double **data;

  // active?
  bool active;

  // sampled?
  bool sampled;

  // decomposed?
  bool decomp;

  // for the determinants.
  double *tempDet;

  // accumulates cluster counts. size: clusterCapacity.
  long *tempCount;

  // accumulates cluster means. size: clusterCapacity * dimensionCapacity.
  double **tempMean;

  // accumulates cluster cov. size: clusterCapacity * (dimensionCapacity * dimensionCapacity).
  double **tempCov;

  // 'current value' variables.
  long curEType;
  long curCluster;
  long curDimension1;
  long curDimension2;

  // 'temporary output' variables
  long outEType;
  long outCluster;
  long outDim1;
  long outDim2;
  double outValue;

  void addCluster(long clusterID) {

    // add a new cluster.
    if (clusterID >= numClusters) {
      numClusters = clusterID + 1;
    }

    // resize the structures.
    if (numClusters > clusterCapacity) {

      long oldCapacity = clusterCapacity;
      clusterCapacity = numClusters + 1;

      clusterProb = (double *)realloc(clusterProb, sizeof(double) * clusterCapacity);
      clusterMean = (double **)realloc(clusterMean, sizeof(double *) * clusterCapacity);
      clusterCov = (double **)realloc(clusterCov, sizeof(double *) * clusterCapacity);

      for (long i=oldCapacity;i<clusterCapacity;i++) {
	clusterProb[i] = 0.0;
	clusterMean[i] = (double *)malloc(sizeof(double) * dimensionCapacity);
	clusterCov[i] =  (double *)malloc(sizeof(double) * (dimensionCapacity * dimensionCapacity));

	for (long j=0;j<dimensionCapacity;j++) {
	  clusterMean[i][j] = 0.0;
	}

	for (long j=0;j<(dimensionCapacity * dimensionCapacity);j++) {
	  clusterCov[i][j] = 0.0;
	}
      }
    }
  }

  void addDimension(long dimensionID) {

    // add a new dimension.
    if (dimensionID >= numDimensions) {
      numDimensions = dimensionID + 1;
    }

    // resize the structures.
    if (numDimensions > dimensionCapacity) {

      long oldCapacity = dimensionCapacity;
      dimensionCapacity = numDimensions + 1;

      for (long i=0;i<clusterCapacity;i++) {
	clusterMean[i] = (double *)realloc(clusterMean[i], sizeof(double) * dimensionCapacity);

	for (long j=oldCapacity;j<dimensionCapacity;j++) {
	  clusterMean[i][j] = 0.0;
	}

	double *newCov = (double *)malloc(sizeof(double) * (dimensionCapacity * dimensionCapacity));
	for (long j=0;j<dimensionCapacity * dimensionCapacity;j++) {
	  newCov[j] = 0.0;
	}

	for (long j=0;j<oldCapacity;j++) {
	  for (long k=0;k<oldCapacity;k++) {
	    newCov[j*dimensionCapacity + k] = clusterCov[i][j*oldCapacity + k];
	  }
	}

	free(clusterCov[i]);
	clusterCov[i] = newCov;
      }

      for (long i=0;i<dataCapacity;i++) {
	data[i] = (double *)realloc(data[i], sizeof(double) * dimensionCapacity);

	for (long j=oldCapacity;j<dimensionCapacity;j++) {
	  data[i][j] = 0.0;
	}
      }
    }
  }

  void addData(long dataID) {

    // check the number
    if (dataID >= numData) {
      numData = dataID + 1;
    }

    // resize the structures.
    if (numData >= dataCapacity) {
      
      long oldCapacity = dataCapacity;
      dataCapacity = numData + 1 + (numData >> 1);
      
      data = (double **)realloc(data, sizeof(double *) * dataCapacity);
      
      for (long i=oldCapacity;i<dataCapacity;i++) {
	data[i] = (double *)malloc(sizeof(double) * dimensionCapacity);

	for (long j=0;j<dimensionCapacity;j++) {
	  data[i][j] = 0.0;
	}
      }
    }
  }

  int unpair1(long val) {
    return (int)(val >> 32);
  }

  int unpair2(long val) {
    return (int)(val);
  }

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  GMMBlock() {

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);

    clusterCapacity = INITIAL_SIZE;
    dimensionCapacity = INITIAL_SIZE;
    dataCapacity = INITIAL_SIZE;

    clusterProb = (double *)malloc(clusterCapacity * sizeof(double));
    clusterMean = (double **)malloc(clusterCapacity * sizeof(double *));
    clusterCov  = (double **)malloc(clusterCapacity * sizeof(double *));
    data        = (double **)malloc(dataCapacity * sizeof(double *));

    for (long i=0;i<clusterCapacity;i++) {
      clusterMean[i] = (double *)malloc(dimensionCapacity * sizeof(double));
      clusterCov[i]  = (double *)malloc((dimensionCapacity * dimensionCapacity) * sizeof(double));
    }

    for (long i=0;i<dataCapacity;i++) {
      data[i] = (double *)malloc(dimensionCapacity * sizeof(double));
    }

    tempDet   = NULL;
    tempCount = NULL;
    tempMean  = NULL;
    tempCov   = NULL;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~GMMBlock() {

    clearParams();

    free(clusterProb);

    for (long i=0;i<clusterCapacity;i++) {
      free(clusterMean[i]);
      free(clusterCov[i]);
    }

    for (long i=0;i<dataCapacity;i++) {
      free(data[i]);
    }

    free(data);
    free(clusterMean);
    free(clusterCov);
  }

  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

    active = (numDimensions > 0 && numClusters > 0 && numData > 0);
    sampled = false;
    curCluster = 0;
    curDimension1 = 0;
    curDimension2 = 0;
    curEType = 1;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

    for (long i=0;i<clusterCapacity;i++) {
      clusterProb[i] = 0.0;
      for (long j=0;j<dimensionCapacity;j++) {
	clusterMean[i][j] = 0.0;
      }

      for (long j=0;j<(dimensionCapacity * dimensionCapacity);j++) {
	clusterCov[i][j] = 0.0;
      }
    }

    for (long i=0;i<dataCapacity;i++) {
      for (long j=0;j<dimensionCapacity;j++) {
	data[i][j] = 0.0;
      }
    }

    numDimensions = 0;
    numClusters = 0;
    numData = 0;
    active = false;
    sampled = false;
    decomp = false;

    if (tempDet != NULL) {
      free(tempDet);
      tempDet = NULL;
    }

    if (tempCount != NULL) {
      free(tempCount);
      tempCount = NULL;
    }

    if (tempMean != NULL && tempCov != NULL) {

      for (long i=0;i<clusterCapacity;i++) {
	free(tempMean[i]);
	free(tempCov[i]);
      } 

      free(tempMean);
      free(tempCov);     
      tempMean = NULL;
      tempCov = NULL;
    }
  }

  void takeParams(RecordIn &input) {

    if (input.cluster_cid != NULL && input.cluster_prob != NULL) {

      addCluster(*input.cluster_cid);
      clusterProb[*input.cluster_cid] = *input.cluster_prob;
    }

    if (input.mean_cidDimPair != NULL && input.mean_value != NULL) {
      int cid = unpair1(*input.mean_cidDimPair);
      int dim = unpair2(*input.mean_cidDimPair);

      addCluster(cid);
      addDimension(dim);
      clusterMean[cid][dim] = *input.mean_value;
    }

    if (input.cov_cid != NULL && input.cov_dimPair != NULL && input.cov_value != NULL) {
      int dim1 = unpair1(*input.cov_dimPair);
      int dim2 = unpair2(*input.cov_dimPair);

      addCluster(*input.cov_cid);
      addDimension(dim1);
      addDimension(dim2);

      clusterCov[*input.cov_cid][dim1*dimensionCapacity + dim2] = *input.cov_value;
    }

    if (input.data_didDimPair != NULL && input.data_value != NULL) {
      int did = unpair1(*input.data_didDimPair);
      int dim = unpair2(*input.data_didDimPair);

      addData(did);
      addDimension(dim);      
      data[did][dim] = *input.data_value;
    }

    active = (numDimensions > 0 && numClusters > 0 && numData > 0);
    sampled = false;
    decomp = false;
  }

  int outputVals(RecordOut &output) {
    
    // check active.
    if (!active) {
      return 0;
    }

    // check if we already decomposed the matrix.
    if (!decomp) {

      // create the space for the determinants.
      tempDet = (double *)malloc(sizeof(double) * clusterCapacity);
      for (long i=0;i<numClusters;i++) {

	// get a view
	gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
	gsl_matrix_view cov = gsl_matrix_view_array_with_tda(clusterCov[i], numDimensions, numDimensions, dimensionCapacity);

	// cholesky-decompose it.
	gsl_linalg_cholesky_decomp(&cov.matrix);

	// and obtain its determinant
	tempDet[i] = 1.0;
	for (long j=0;j<numDimensions;j++) {
	  tempDet[i] *= gsl_matrix_get(&cov.matrix, j, j);
	}

	// then square the determinant.
	tempDet[i] *= tempDet[i];

	// and invert the matrix.
	gsl_linalg_cholesky_invert(&cov.matrix);
	gsl_set_error_handler(oldhandler);
      }

      decomp = true;
    }

    // check if we have to sample.
    if (!sampled) {

      // fist, create the structures
      tempCount = (long *)malloc(sizeof(long) * clusterCapacity);
      tempMean  = (double **)malloc(sizeof(double*) * clusterCapacity);
      tempCov   = (double **)malloc(sizeof(double*) * clusterCapacity);

      // and initialize them.
      for (long i=0;i<clusterCapacity;i++) {
	tempCount[i] = 0;
	tempMean[i] = (double *)malloc(sizeof(double) * dimensionCapacity);
	tempCov[i]  = (double *)malloc(sizeof(double) * (dimensionCapacity * dimensionCapacity));

	for (long j=0;j<dimensionCapacity;j++) {
	  tempMean[i][j] = 0.0;
	}

	for (long j=0;j<(dimensionCapacity * dimensionCapacity);j++) {
	  tempCov[i][j] = 0.0;
	}
      }

      // now, go through each data point.
      gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
      gsl_vector *datum = gsl_vector_calloc(numDimensions);
      gsl_vector *ym = gsl_vector_calloc(numDimensions);
      double *tempProb = (double *)malloc(sizeof(double) * clusterCapacity);
      unsigned int *resultCount = (unsigned int *)malloc(sizeof(unsigned int) * clusterCapacity);
      for (long i=0;i<numData;i++) {

	// compute the probability on each cluster.
	for (long j=0;j<numClusters;j++) {
	  
	  // first, the cluster probability.
	  tempProb[j] = log(clusterProb[j]);

	  // then, and the probability of the point being in the j^th cluster.
	  // we set the datum values.
	  for (long k=0;k<numDimensions;k++) {
	    gsl_vector_set(datum, k, data[i][k]);
	  }

	  // and get a view over the mean vector/covariance matrix of the j^th cluster.
	  gsl_vector_view gsl_mean = gsl_vector_view_array(clusterMean[j], numDimensions);
	  gsl_matrix_view gsl_covariance = gsl_matrix_view_array_with_tda(clusterCov[j], numDimensions, numDimensions, dimensionCapacity);

	  // subtract
	  gsl_vector_sub(datum, &gsl_mean.vector);

	  // multiply.
	  gsl_blas_dsymv(CblasUpper, 1.0, &gsl_covariance.matrix, datum, 0.0, ym);

	  // get the dot product.
	  double ay;
	  gsl_blas_ddot(datum, ym, &ay);

	  tempProb[j] += log(exp(-0.5*ay)/sqrt(pow(((double)2*3.1415926), (double)numDimensions)*tempDet[j]));
	  tempProb[j] = exp(tempProb[j]);
	}

	// at this point, we can sample from the multinomial.
	gsl_ran_multinomial(rng, numClusters, 1, tempProb, resultCount);

	// and find out which cluster we got.
	long whichClust = -1;
	for (long j=0;j<numClusters;j++) {
	  if (resultCount[j] > 0) {
	    whichClust = j;
	    break;
	  }
	}

	if (whichClust < 0) {
	  
	  // assign them randomly if they all had very low probability.
	  whichClust = gsl_rng_uniform_int(rng, numClusters);
	}

	// then, we increase everything there
	if (whichClust >= 0) {

	  // the counts
	  tempCount[whichClust]++;
	  
	  for (long j=0;j<numDimensions;j++) {

	    // the means
	    tempMean[whichClust][j] += data[i][j];

	    // the covariances.
	    for (long k=0;k<numDimensions;k++) {
	      tempCov[whichClust][j*numDimensions + k] += (data[i][j] - clusterMean[whichClust][j]) * (data[i][k] - clusterMean[whichClust][k]);
	    }
	  }
	}
      }

      // we're done sampling!
      sampled = true;

      // free the temporary stuff.
      free(tempProb);
      free(resultCount);
      gsl_vector_free(datum);
      gsl_vector_free(ym);

      // set the iteration variables.
      curCluster = 0;
      curDimension1 = 0;
      curDimension2 = 0;
      curEType = 1;

      gsl_set_error_handler(oldhandler);
    }

    // now that we're sampled, we can output values.
    if (curEType == 1) {
      outEType = 1;
      outCluster = curCluster;
      outDim1 = 0;
      outDim2 = 0;
      outValue = tempCount[curCluster];

      // go to the next cluster.
      curCluster++;

      // or the next eType, if we're done.
      if (curCluster >= numClusters) {
	curCluster = 0;
	curEType = 2;
      }

    } else if (curEType == 2) {
      outEType = 2;
      outCluster = curCluster;
      outDim1 = curDimension1;
      outDim2 = 0;
      outValue = tempMean[curCluster][curDimension1];

      // go to the next dimension.
      curDimension1++;

      // or the next cluster.
      if (curDimension1 >= numDimensions) {
	curCluster++;
	curDimension1 = 0;
      }

      // if we're out of clusters, the next eType.
      if (curCluster >= numClusters) {
	curCluster = 0;
	curEType = 3;
      }

    } else if (curEType == 3) {
      outEType = 3;
      outCluster = curCluster;
      outDim1 = curDimension1;
      outDim2 = curDimension2;
      outValue = tempCov[curCluster][curDimension1*numDimensions + curDimension2];
      
      // go to the next y-dimension.
      curDimension2++;

      // and the next x-dimension.
      if (curDimension2 >= numDimensions) {
	curDimension1++;
	curDimension2 = 0;
      }

      // and the next cluster.
      if (curDimension1 >= numDimensions) {
	curCluster++;
	curDimension1 = 0;       
      }

      // if we're done, we're done
      if (curCluster >= numClusters) {
	active = false;
      }
    }

    // now that we have passed the values, we can set up the output record.
    output.out_eType = &outEType;
    output.out_cluster = &outCluster;
    output.out_dim1 = &outDim1;
    output.out_dim2 = &outDim2;
    output.out_val = &outValue;

    return 1;
  }
  
  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){9, 
	{INT, DOUBLE,  INT, DOUBLE, INT, INT, DOUBLE, INT, DOUBLE}, 
	  {"cluster_cid", "cluster_prob", "mean_cidDimPair", "mean_value", "cov_cid", "cov_dimPair", "cov_value", "data_didDimPair", "data_value"}, 
	    {0, 1, 0, 1, 0, 0, 1, 0, 1}
    };
  }

  VGSchema outputSchema() {
    return (VGSchema){5, 
	{INT, INT, INT, INT, DOUBLE}, 
	  {"out_eType", "out_cluster", "out_dim1", "out_dim2", "out_val"}, 
	    {0, 0, 0, 0, 1}
    };
  }

  const char *getName() {
    return "gmmblock";
  }
};

VGFunction *create() {
  return (new GMMBlock());
}

void destroy(VGFunction *vgFunction) {
  delete (GMMBlock *)vgFunction;
}
