#include <stdio.h>
#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>

// #define DEBUG

/**
 * A Gamma VG function.
 *
 * input: (shape,scale)
 * X ~ Gamma(shape, scale)
 * output: (X)
 *
 * NOTE: to get an inverse gamma, simply parameterize with 1.0 / scale and 
 * obtain 1.0 / X on the output.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *shape;
  double *scale;

};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  double *value;

};

// ----------------------------------------------------------- // 

/** VG Function class */
class Gamma : public VGFunction {

private:
  /** RNG and structures go here. */
  gsl_rng *rng;
  double shape;
  double scale;
  double outputVal;
  bool haveShape;
  bool haveScale;
  bool active;
  

  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Gamma() {

#ifdef DEBUG
    printf("[Gamma]: constructor.\n");
    fflush(stdout);
#endif

    // get a RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Gamma() {

#ifdef DEBUG
    printf("[Gamma]: destructor.\n");
    fflush(stdout);
#endif

    // destroy the RNG.
    gsl_rng_free(rng);	
  }

  /** Initializes the RNG seed for every group. */
  void initializeSeed(long seedValue) {

#ifdef DEBUG
    printf("[Gamma]: initializeSeed(%ld).\n", seedValue);
    fflush(stdout);
#endif

    // set the seed.
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current group and prepares the structures for
   * the next call to outputVals(). */
  void finalizeTrial() {

#ifdef DEBUG
    printf("[Gamma]: finalizeTrial.\n");
    fflush(stdout);
#endif

    // set active
    active = (haveShape && haveScale);
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("[Gamma]: clearParams.\n");
#endif

    active = false;
    haveShape = false;
    haveScale = false;
  }

  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {


#ifdef DEBUG
    printf("[Gamma]: takeParams(");
    if (input.shape == NULL)
      printf("NULL, ");
    else
      printf("%f, ", *input.shape);

    if (input.scale == NULL)
      printf("NULL).\n");
    else
      printf("%f).\n", *input.scale);

    fflush(stdout);
#endif

    // set the parameters
    if (input.shape != NULL) {
      shape = *input.shape;
      haveShape = true;
    }

    if (input.scale != NULL) {
      scale = *input.scale;
      haveScale = true;
    }


    // set active
    active = (haveShape && haveScale);
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 

#ifdef DEBUG
    printf("[Gamma]: outputVals(%f,%f) => ", shape, scale);
#endif
    
    // active?
    if (!active) {

#ifdef DEBUG
      printf("NULL.\n");
      fflush(stdout);
#endif

      return 0;
    }

    // sample
    outputVal = gsl_ran_gamma(rng, shape, scale);
    output.value = &outputVal;
    active = false;

#ifdef DEBUG
    printf("%f\n", *output.value);
    fflush(stdout);
#endif

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){2, {DOUBLE, DOUBLE}, {"shape", "scale"}, {0, 0}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {DOUBLE}, {"value"}, {1}};
  }

  const char *getName() {
    return "Gamma";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Gamma());
}

void destroy(VGFunction *vgFunction) {
  delete (Gamma *)vgFunction;
}

