#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <stdio.h>
#include <vector>
#include <map>
#include <utility>
#include <math.h>
#include <iostream>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;
/** 
* Input record, used by the takeParams() method.
* Values could be NULL, meaning they were not present in the tuple.
*/
struct RecordIn {
	double *a;
	double *b;
};

/**
* Output record, used by the outputVals().
* If any of the values is NULL, then the user must allocate space for
* it (the engine will de-allocate). 
*/
struct RecordOut {
	double *outValue;
};

int runningError;
void handler (const char * reason,
			  const char * file,
			  int line,
			  int gsl_errno)
{
	runningError = 1;
}




class InvGamma: public VGFunction {
private:
	gsl_rng *rng;

	/*---------------------state information --------------*/
	double a, b;
	/*---------------------result------------------------*/
	double result;

	bool active;

public:

	/** Constructor. Use this to declare your RNG and other
	* important structures .
	*/
	InvGamma() {
		// create the RNG.
		rng = gsl_rng_alloc(gsl_rng_mt19937);
		runningError = -1;
	}

	~InvGamma()
	{
		// destroy the RNG
		gsl_rng_free(rng);
	}


	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
		a = 0;
		b = 0;
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial() {

		// set active again.
		active = (b > 0);
	}


	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams() {
		active = false;
		a = 0;
		b = 0;
	}

	//take the parameters for vgfunction
	void takeParams(RecordIn &input) {
		//prior
		if(input.a != NULL)
		{
			a = *input.a;
		}

		if(input.b != NULL)
		{
			b = *input.b;
		}

		active = (b > 0);
	}

	double computeInvGamma(double a, double b)
	{
		return 1/gsl_ran_gamma(rng, a, 1/b);
		//	return 1;
	}

	/** 
	* Produces the sample values. Returns 1 if there are more
	* records to be produced for the current sample, 0 otherwise. 
	*/
	int outputVals(RecordOut &output) {
		if(!active || a < 0 || b <= 0)
		{
			return 0;
		}

		
		while(true)
		{
			runningError = -1;
			gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
			result = computeInvGamma(a, b);
			gsl_set_error_handler(oldhandler);
			if(runningError == -1)
				break;
		}
		
		// LL: removed this because it affects the Lasso. Just
		// use the sqrt() function in the engine if you want a
		// standard deviation.

		/*** result = sqrt(result); ***/
		output.outValue = &result;
		
		active = false;
		return 1;
	}

	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {

		return (VGSchema){2, {DOUBLE, DOUBLE}, {"a", "b"}, {0, 1}};
	}

	VGSchema outputSchema() {

		return (VGSchema){1, {DOUBLE}, {"outValue"}, {1}};
	}

	const char *getName() {
		return "InvGamma";
	}
};


/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
	return(new InvGamma());
}

void destroy(VGFunction *vgFunction) {
	delete (InvGamma *)vgFunction;
}
