//:C_InverseWishart.cc

#include "VGFunction.h"

#include <gsl/gsl_rng.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_cblas.h>
#include "gsl/gsl_blas.h"
#include <gsl/gsl_errno.h>
#include <math.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <stdio.h>
#include <string.h>

#define MAX_DIM 1

using namespace std;

/**
 * An Inverse Wishart VG function.
 * Used by the GMM query.
 *
 * Input: (dim_i, dim_j, V[i,j], NULL) for i,j=1:D
 *        (NULL, NULL, NULL, df)
 * Parameterization: X ~ InvWishart(V, df)
 * Output: (dim_i, dim_j, X[i,j]) for i,j=1:D
 */

/** 
* Input record, used by the takeParams() method.
* Values could be NULL, meaning they were not present in the tuple.
*/
struct RecordIn {
	long *iw_variance_dimension_1;
	long *iw_variance_dimension_2;
	double *in_variance_value;

	long *iwM;
};

/**
* Output record, used by the outputVals().
* If any of the values is NULL, then the user must allocate space for
* it (the engine will de-allocate). 
*/
struct RecordOut {
	long *out_variance_dimension1;
	long *out_variance_dimension2;
	double *out_variance_dimension_value;
};


int runningError;
void handler (const char * reason,
			  const char * file,
			  int line,
			  int gsl_errno)
{
	runningError = 1;
	stringstream ss;
	string result = "reason: InverseWishart";
	result += reason;
	result += "\t file:";
	result += file;
	result += "\t line:";
	ss.str("");
	ss << line;
	result += ss.str();
	result += "\t gsl_errno:";
	ss.str("");
	ss << gsl_errno;
	result += ss.str();
	result += "\r\n";

	ofstream fileWriter;

	
	fileWriter.open("/tmp/error.report", ios::trunc);

	fileWriter<< result.c_str();
	fileWriter.close();
}

class InverseWishart: public VGFunction {
private:
	gsl_rng *rng;

	double **scaleMatrix; /* scale matrix */
	long degreeOfFreedom; /* degrees of freedom */

	long maxDimensionSize;
	long currentDimensionSize;

	long outputIndex_i; /* keeps track of output variables */
	long outputIndex_j; /* keeps track of output variables */

	gsl_matrix* invwishart_rand;

	/*---------------- temp value -------------------- */
	long tempInputValue;
	long tempOutput_i, tempOutput_j;
	double tempOutputValue;

	bool active;

public:

	/*
	* Constructor. Use this to declear your RNG and other important structures .
	*/
	InverseWishart()
	{
		rng = gsl_rng_alloc(gsl_rng_mt19937);

		maxDimensionSize = MAX_DIM;

		scaleMatrix = (double **)malloc(maxDimensionSize * sizeof(double *));
		for(int i = 0; i < maxDimensionSize; i++)
		{
			scaleMatrix[i] = (double *)malloc(maxDimensionSize * sizeof(double));
		}
		
		for(int i = 0; i < maxDimensionSize; i++)
        {
            memset(scaleMatrix[i], 0, maxDimensionSize * sizeof(double));
        }		

		outputIndex_i = 0;
		outputIndex_j = 0;

		currentDimensionSize = 0;
		runningError = -1;
	}

	~InverseWishart()
	{
		// destroy the RNG
		gsl_rng_free(rng);

		for(int i = 0; i < maxDimensionSize; i++)
		{
			free(scaleMatrix[i]);
		}

		free(scaleMatrix);
	}

	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial() {

		// set active again.
		active = (currentDimensionSize > 0);

		outputIndex_i = 0; /* keeps track of output variables */
		outputIndex_j = 0; /* keeps track of output variables */
		runningError = -1;
	}

	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams()
	{
		for(int i = 0; i < maxDimensionSize; i++)
		{
			memset(scaleMatrix[i], 0, maxDimensionSize * sizeof(double));
		}

		degreeOfFreedom = 0; /* degrees of freedom */
		currentDimensionSize = 0;

		outputIndex_i = 0; /* keeps track of output variables */
		outputIndex_j = 0; /* keeps track of output variables */
		
		runningError = -1;
		active = false;
	}

	void takeParams(RecordIn &input)
	{
		if(input.iw_variance_dimension_1 != NULL &&
			input.iw_variance_dimension_2 != NULL &&
			input.in_variance_value != NULL)
		{
			//cout << "para0: " << *input.iw_variance_dimension_1 << endl;
			//cout << "para1: " << *input.iw_variance_dimension_2 << endl;
			//cout << "para2: " << *input.in_variance_value << endl;

			tempInputValue = *input.iw_variance_dimension_1;
			if(tempInputValue < *input.iw_variance_dimension_2)
			{
				tempInputValue = *input.iw_variance_dimension_2;
			}

			if(currentDimensionSize < tempInputValue + 1)
			{
				currentDimensionSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxDimensionSize)
			{
				long previousDimensionSize = maxDimensionSize;

				maxDimensionSize = (tempInputValue + (tempInputValue >> 1) + 1);

				//reallocate the memory for scaleMatrix
				scaleMatrix = (double **)realloc(scaleMatrix, maxDimensionSize * sizeof(double *));
				for(int i = 0; i < maxDimensionSize; i++)
				{
					if(i >= previousDimensionSize)
					{
						scaleMatrix[i] = NULL;
					}

					scaleMatrix[i] = (double *)realloc(scaleMatrix[i], maxDimensionSize * sizeof(double));
				}

				//initialize the increased part as 0.

				for(int i = 0; i < maxDimensionSize; i++)
				{
					for(int j = 0; j < maxDimensionSize; j++)
					{
						if(i >= previousDimensionSize || j >= previousDimensionSize)
						{
							scaleMatrix[i][j] = 0;
						}
					}
				}
			}

			scaleMatrix[*input.iw_variance_dimension_1][*input.iw_variance_dimension_2] 
			= *input.in_variance_value;
		}

		if (input.iwM != NULL)
		{
			//cout << "iwm: " << *input.iwM << endl;
			degreeOfFreedom = *input.iwM;
		}
		
		active = (currentDimensionSize > 0);
	}

	void reverse_matrix(gsl_matrix *gsmatrix, gsl_matrix *reverse, int size) {
		int s;
		gsl_permutation *p = gsl_permutation_alloc(size);
		gsl_linalg_LU_decomp(gsmatrix, p, &s);
		gsl_linalg_LU_invert(gsmatrix, p, reverse);
		gsl_permutation_free(p);
	}

	int outputVals(RecordOut &output) {
		
		//for(int i = 0; i < currentDimensionSize; i++)
		//{
			//for(int j = 0; j < currentDimensionSize; j++)
			//{
				//cout << "<" << i << ", " << j << ">: " << scaleMatrix[i][j] << endl;
			//}
		//}

		if(!active)
		{
			return 0;
		}

		for (int i = 0; i < currentDimensionSize; ++i) {
			for (int j = 0; j < currentDimensionSize; ++j) {
				if (isnan(scaleMatrix[i][j]) || isinf(scaleMatrix[i][j])) {
					active = false;
					return (0);
				}
			}
		}

		if (isnan(degreeOfFreedom) || 
			isinf(degreeOfFreedom) || 
			(degreeOfFreedom <= (currentDimensionSize - 1))) {
				active = false;
				return (0);
		}

		if (outputIndex_i == 0 && outputIndex_j == 0) {  /* sample IW random mattrix */
			
			cout << "The scale matrix is : " << endl;	
			for (int i = 0; i < currentDimensionSize; ++i) {
            	for (int j = 0; j < currentDimensionSize; ++j) {
                	cout << scaleMatrix[i][j] << " ";
				}
				cout << " " << endl;
            }	

			bool isLegalSample = false;
			invwishart_rand = gsl_matrix_calloc(currentDimensionSize, currentDimensionSize);

			while (!isLegalSample) { /* loop till we get a legal sample */
				runningError = -1;
				gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);

				gsl_matrix* work = gsl_matrix_calloc(currentDimensionSize, currentDimensionSize);
				for (int i = 0; i < currentDimensionSize; ++i) {
					gsl_matrix_set(work, i, i, sqrt(gsl_ran_chisq(rng, degreeOfFreedom - i)));
					for (int j = 0; j < i; ++j) {
						gsl_matrix_set(work, i, j, gsl_ran_ugaussian(rng));
					}
				}

				/* reverse the scale_matrix */
				gsl_matrix *scale_matrix = gsl_matrix_calloc(currentDimensionSize, currentDimensionSize);
				for(int i = 0; i < currentDimensionSize; i++) {
					for(int j = 0; j < currentDimensionSize; j++) {
						gsl_matrix_set(scale_matrix, i, j, scaleMatrix[i][j]);
					}
				}
				gsl_matrix *temp_result = gsl_matrix_calloc(currentDimensionSize, currentDimensionSize); 
				reverse_matrix(scale_matrix, temp_result, currentDimensionSize);
				gsl_matrix_free(scale_matrix);		    

				gsl_linalg_cholesky_decomp(temp_result);
				gsl_blas_dtrmm(CblasLeft, CblasLower, CblasNoTrans, CblasNonUnit, 1.0, temp_result, work);
				gsl_blas_dsyrk(CblasUpper, CblasNoTrans, 1.0, work, 0.0, temp_result);
				gsl_matrix_free(work);

				/* change the up-trangle matrix to full matrix */
				for(int i = 0; i < currentDimensionSize; i++) {
					for(int j = 0; j < i; j++) {
						gsl_matrix_set(temp_result, i, j, gsl_matrix_get(temp_result, j, i));
					}
				}

				/* reverse the result */
				reverse_matrix(temp_result, invwishart_rand, currentDimensionSize);
				gsl_set_error_handler(oldhandler);

				bool isSingular = false;
				if (runningError == 1) {
					isSingular = true;
				}

				for (int i = 0; i < currentDimensionSize; i++) {
					double u = gsl_matrix_get(invwishart_rand, i, i);
					if (u == 0) { /* singular mattrix */
						
						isSingular = true;
						break;
					}
				}

				for (int i = 0; i < currentDimensionSize; i++) {
					for(int j = 0; j < currentDimensionSize; j++) {
						double u = gsl_matrix_get(invwishart_rand, i, j);
						if (isnan(u) || isinf(u)) { /* singular mattrix */
							isSingular = true;
							break;
						}
					}
				}
				gsl_matrix_free(temp_result);
				if (!isSingular) {
					isLegalSample = true;
				}
				else
				{
					cout << "the sampled matrix is singular" << endl;
					for (int i = 0; i < currentDimensionSize; i++) {
                    	for(int j = 0; j < currentDimensionSize; j++) {
							cout << gsl_matrix_get(invwishart_rand, i , j) << " ";
						}
						cout << endl;
					}
				}
			}
		}

		tempOutput_i = outputIndex_i;
		tempOutput_j = outputIndex_j;

		output.out_variance_dimension1 = &tempOutput_i;
		output.out_variance_dimension2 = &tempOutput_j;
		tempOutputValue = gsl_matrix_get(invwishart_rand, outputIndex_i, outputIndex_j);
		output.out_variance_dimension_value = &tempOutputValue;

		outputIndex_j ++; /* update result co-ordinates */
		if (outputIndex_j >= currentDimensionSize)
		{
			outputIndex_i ++;
			if (outputIndex_i >= currentDimensionSize)
			{ /* no more values */
				if (invwishart_rand != NULL) {
					gsl_matrix_free(invwishart_rand);
				}

				active = false;
			}
			else { /* next row */
				outputIndex_j = 0;
			}
		}

		return 1;
	}

	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {

		return (VGSchema){4, 
		{INT, INT, DOUBLE, INT}, 
		{"iw_variance_dimension_1", "iw_variance_dimension_2", "in_variance_value",  
		"iwM"}, 
		{0, 0, 0, 0}
		};
	}

	VGSchema outputSchema() {
		return (VGSchema){3, 
		{INT, INT, DOUBLE}, 
		{"out_variance_dimension1", "out_variance_dimension2", "out_variance_dimension_value"}, 
		{0, 0, 1}
		};
	}

	const char *getName() {
		return "inverse_wishart";
	}
};

VGFunction *create() {
	return (new InverseWishart());
}

void destroy(VGFunction *vgFunction) {
	delete (InverseWishart *)vgFunction;
}
