#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <stdio.h>
#include <vector>
#include <map>
#include <utility>
#include <math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_blas.h>
#include <iostream>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

/**
 * This is a VG function for the conjugate inverse wishart
 * distribution for the covariance matrices in the GMM query.
 *
 * Inputs: (hdim_i, hdim_j, H[i,j], NULL, NULL, NULL, NULL, NULL) for i,j=1:D
 *         (NULL, NULL, NULL, hdf, NULL, NULL, NULL, NULL)
 *         (NULL, NULL, NULL, NULL, vdim_i, vdim_j, V[i,j], NULL) for i,j=1:D
 *         (NULL, NULL, NULL, NULL, NULL, NULL, NULL, vdf)
 *
 * Parameterization: X ~ InvWishart(H + V, hdf + vdf)
 *
 * Output: (dim_i, dim_j, X[i,j]) for i,j=1:D
 */

/** 
* Input record, used by the takeParams() method.
* Values could be NULL, meaning they were not present in the tuple.
* There are 8 parameters for Inverse_Wishart_Conjugate distribution.
*/
struct RecordIn {
	long *hyper_dimension_id_1;
	long *hyper_dimension_id_2;
	double *hyper_dimension_value;

	long *hyper_scale;

	long *dimension_id_1;
	long *dimension_id_2;
	double *dimension_value;

	long *data_num;
};

/**
* Output record, used by the outputVals().
* If any of the values is NULL, then the user must allocate space for
* it (the engine will de-allocate). 
*/
struct RecordOut {
	long *out_dimension_id_1;
	long *out_dimension_id_2;
	double *dimension_value;
};


int runningError;
void handler (const char * reason,
			  const char * file,
			  int line,
			  int gsl_errno)
{
	runningError = 1;
	stringstream ss;
	string result = "reason: InverseWishartConjugate";
	result += reason;
	result += "\t file:";
	result += file;
	result += "\t line:";
	ss.str("");
	ss << line;
	result += ss.str();
	result += "\t gsl_errno:";
	ss.str("");
	ss << gsl_errno;
	result += ss.str();
	result += "\r\n";

	/***
	ofstream fileWriter;

	
	fileWriter.open("error.report", ios::trunc);

	fileWriter<< result.c_str();
	fileWriter.close();
	*/

	cerr << result << "\n\n";
}


class Inverse_Wishart_Conjugate: public VGFunction {
private:
	gsl_rng *rng;
	
	/*---------------------state information --------------*/

	//hyper_variance
	map<pair<long, long>, double> hyper_covariance_map;
	long hyper_scale;

	//variance of this cluster
	map<pair<long, long>, double> cluster_covariance_map;

	//number of data
	long data_num;

	/*---------------------result------------------------*/
	map<pair<long, long>, double> resultMap;
	long outputIndex;
	map<pair<long, long>, double>::iterator result_iterator;

	long outDimension_id_1;
	long outDimension_id_2;
	double outValue;

	bool active;


public:

	/*
	 * Constructor. Use this to declear your RNG and other important structures .
	 */
	Inverse_Wishart_Conjugate()
	{
		rng = gsl_rng_alloc(gsl_rng_mt19937);
		clearData();
		outputIndex = 0;
		runningError = -1;
	}

	/** Destructor. Deallocate everything from the constructor. */
	~Inverse_Wishart_Conjugate()
	{
		// destroy the RNG
		gsl_rng_free(rng);
		runningError = -1;
		clearData();
	}


	/** Initializes the RNG seed for a given call. */
	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
		runningError = -1;
	}

	void clearData()
	{
		hyper_covariance_map.clear();
		cluster_covariance_map.clear();
		resultMap.clear();
		
		hyper_scale = 0;
		data_num = 0;
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial()
	{
		active = (!hyper_covariance_map.empty());
		outputIndex = 0;
	}

	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams()
	{
		active = false;
		clearData();
		outputIndex = 0;
	}

	void takeParams(RecordIn &input) {
		//prior
		if(input.hyper_dimension_id_1 != NULL && 
			input.hyper_dimension_id_2 != NULL &&
			input.hyper_dimension_value != NULL)
		{
			long dimension_id_1 = *input.hyper_dimension_id_1;
			long dimension_id_2 = *input.hyper_dimension_id_2;
			double dimension_value = *input.hyper_dimension_value;
			hyper_covariance_map.insert(pair<pair<long, long>, double>(pair<long, long>(dimension_id_1, dimension_id_2), dimension_value));

			//fprintf(stderr, "hyper matrix: %d, %d, %f \n", dimension_id_1, dimension_id_2, dimension_value);
		}

		if(input.hyper_scale != NULL)
		{
			hyper_scale = *input.hyper_scale;
			//fprintf(stderr, "hyper scale: %d\n", hyper_scale);
		}

		if(input.dimension_id_1 != NULL && 
			input.dimension_id_2 != NULL &&
			input.dimension_value != NULL)
		{
			long dimension_id_1 = *input.dimension_id_1;
			long dimension_id_2 = *input.dimension_id_2;
			double dimension_value = *input.dimension_value;
			cluster_covariance_map.insert(pair<pair<long, long>, double>(pair<long, long>(dimension_id_1, dimension_id_2), dimension_value));
			//fprintf(stderr, "data matrix: %d, %d, %f \n", dimension_id_1, dimension_id_2, dimension_value);
		}

		if(input.data_num != NULL)
		{
			data_num = *input.data_num;
			//fprintf(stderr, "data num: %d\n", data_num);
		}

		active = (!hyper_covariance_map.empty());
	}

	void set_matrix_by_map(gsl_matrix *matrix, map<pair<long, long>, double> *matrix_map, int size)
	{
		long **flag = new long*[size];
		for(int i = 0; i < size; i++)
		{
			flag[i] = new long[size];
			for(int j = 0; j < size; j++)
			{
				flag[i][j] = 0;
			}
		}

		for(map<pair<long, long>, double>::iterator it = matrix_map->begin(); it != matrix_map->end(); it++)
		{
			gsl_matrix_set(matrix, it->first.first, it->first.second, it->second);
			flag[it->first.first][ it->first.second] = 1;
		}

		for(int i = 0; i < size; i++)
		{
			for(int j = 0; j < size; j++)
			{
				if(flag[i][j] == 0)
				{
					gsl_matrix_set(matrix, i, j, 0);
				}
			}
		}

		for(int i = 0; i < size; i++)
		{
			delete [] flag[i];
		}

		delete []flag;
	}

	void set_vector_by_map(gsl_vector *gsvector, map<long, double> *vector_map)
	{
		for(map<long, double>::iterator it = vector_map->begin(); it != vector_map->end(); it++)
		{
			gsl_vector_set(gsvector, it->first, it->second);
		}
	}

	void reverse_matrix(gsl_matrix *gsmatrix, gsl_matrix *reverse, int dimension)
	{
		int s;

		gsl_permutation *p = gsl_permutation_calloc(dimension);
		gsl_linalg_LU_decomp(gsmatrix, p, &s);
		gsl_linalg_LU_invert(gsmatrix, p, reverse);
		gsl_permutation_free(p);
	}

	void computeClusterCovariance()
	{
		resultMap.clear();

		int dim = (int)(hyper_scale - 1);
		int df = (int)(hyper_scale + data_num);

		gsl_matrix *gsl_hyper_covariance = gsl_matrix_calloc(dim, dim);
		gsl_matrix *gsl_cluster_covariance = gsl_matrix_calloc(dim, dim);

		set_matrix_by_map(gsl_hyper_covariance, &hyper_covariance_map, dim);
		set_matrix_by_map(gsl_cluster_covariance, &cluster_covariance_map, dim);

		gsl_matrix_add(gsl_cluster_covariance, gsl_hyper_covariance);

		//fprintf(stderr, "here!~\n");
		gsl_matrix *result = gsl_matrix_calloc(dim, dim);
		inv_wishart(rng, dim, df, gsl_cluster_covariance, result);

		for(int i = 0; i < dim; i++)
			for(int j = 0; j < dim; j++)
				resultMap.insert(pair<pair<long, long>, double>(pair<long, long>(i, j), gsl_matrix_get(result, i, j)));

		gsl_matrix_free(gsl_hyper_covariance);
		gsl_matrix_free(gsl_cluster_covariance);
		gsl_matrix_free(result);
	}


	int outputVals(RecordOut &output)
	{
		if(!active)
		{
			return 0;
		}

		if(outputIndex == 0)
		{
			while(true)
			{
				runningError = -1;
				gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
				computeClusterCovariance();
				gsl_set_error_handler(oldhandler);

				bool toContinue = false;

				for(map<pair<long, long>, double>::iterator tempIterator = resultMap.begin();
					tempIterator != resultMap.end();
					tempIterator ++)
				{
					if(isnan(tempIterator->second) || isinf(tempIterator->second))
					{
						toContinue = true;
						break;
					}
				}
				if(toContinue || runningError == 1)
					continue;
				else
				{
					break;
				}
			}

			result_iterator = resultMap.begin();

			outDimension_id_1 = result_iterator->first.first;
			outDimension_id_2 = result_iterator->first.second;
			outValue = result_iterator->second;

			output.out_dimension_id_1 = &outDimension_id_1;
			output.out_dimension_id_2 = &outDimension_id_2;
			output.dimension_value = &outValue;
		
			outputIndex ++;
			result_iterator++;
			if(result_iterator == resultMap.end()) {
				active = false;
			}

			return (1);
		}
		else
		{
			outDimension_id_1 = result_iterator->first.first;
			outDimension_id_2 = result_iterator->first.second;
			outValue = result_iterator->second;

			output.out_dimension_id_1 = &outDimension_id_1;
			output.out_dimension_id_2 = &outDimension_id_2;
			output.dimension_value = &outValue;
			
			outputIndex ++;
			result_iterator++;
			if(result_iterator == resultMap.end()) {
				active = false;
			}

			return (1);
		}
	}

	int inv_wishart(gsl_rng *r, int n, int dof, gsl_matrix *scale, gsl_matrix *result){

		int wrongTimes = 0;
		gsl_matrix *scale_copy = gsl_matrix_calloc(n, n);
		gsl_matrix_memcpy(scale_copy, scale);

		while(true)
		{
			bool isGo = false;
			gsl_matrix *scale_inv = gsl_matrix_calloc(n, n);
			gsl_matrix *tempResult = gsl_matrix_calloc(n, n);

			reverse_matrix(scale, scale_inv, n);

			int k,l;
			gsl_matrix *work = gsl_matrix_calloc(n,n);

			for(k=0; k<n; k++){
				gsl_matrix_set(work, k, k, sqrt( gsl_ran_chisq( r, (dof-k) ) ) );
				for(l=0; l<k; l++){
					gsl_matrix_set( work, k, l, gsl_ran_ugaussian(r) );
				}
			}
			gsl_matrix_memcpy(tempResult, scale_inv);
			gsl_linalg_cholesky_decomp(tempResult);
			gsl_blas_dtrmm(CblasLeft,CblasLower,CblasNoTrans,CblasNonUnit,1.0,tempResult,work);
			gsl_blas_dsyrk(CblasUpper,CblasNoTrans,1.0,work,0.0,tempResult);

			for(int i = 0; i < n; i++)
				for(int j = 0; j < i; j++)
					gsl_matrix_set(tempResult, i, j, gsl_matrix_get(tempResult, j, i));

			for (int i = 0; i < n; i++) {
				for(int j = 0; j < n; j++) {
					double u = gsl_matrix_get(tempResult, i, j);
					if (isnan(u) || isinf(u)) { /* singular mattrix */
						isGo = true;
						break;
					}
				}
			}

			reverse_matrix(tempResult, result, n);

			gsl_matrix_free(scale_inv);
			gsl_matrix_free(tempResult);
			gsl_matrix_free(work);

			if(isGo == true)
			{
				wrongTimes ++;
			}
			else
			{
				break;
			}

			if(wrongTimes >= 20)
			{
				gsl_matrix_scale(scale_copy, 1/(double)dof);
				gsl_matrix_memcpy(result, scale_copy);
				break;
			}	
		}

		gsl_matrix_free(scale_copy);
		return 0;
	}

	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {

		return (VGSchema){8, 
						  {INT, INT, DOUBLE, INT, INT, INT, DOUBLE, INT}, 
						  {"prior_iw_variance_dimension_1", "prior_iw_variance_dimension_2", "prior_in_variance_value",
						   "prior_iwM", 
		                   "cluster_iw_variance_dimension_1", "cluster_iw_variance_dimension_2", "cluster_in_variance_value",
						   "cluster_iwM"}, 
						  {0, 0, 0, 0, 0, 0, 1, 1}
		};
	}

	VGSchema outputSchema() {
		return (VGSchema){3, 
						 {INT, INT, DOUBLE}, 
						 {"out_variance_dimension1", "out_variance_dimension2", "out_variance_dimension_value"}, 
						 {0, 0, 1}
		};
	}

	const char *getName() {
		return "inverse_wishart_conjugate";
	}
};

VGFunction *create() {
	return (new Inverse_Wishart_Conjugate());
}

void destroy(VGFunction *vgFunction) {
	delete (Inverse_Wishart_Conjugate *)vgFunction;
}
