#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_linalg.h>
#include <stdio.h>
#include <gsl/gsl_errno.h>
#include <stdio.h>
#include <string.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

#define DEAFULT_DIMENSION 1

using namespace std;

/** 
 * A Multivariate Gaussian VG function.
 * Used by the GMM.
 *
 * Inputs: (mdim_i, mean[i], NULL, NULL, NULL) for i=1:D
 *         (NULL, NULL, mcov_i, mcov_j, cov[i,j]) for i=1:D
 *
 * Parameterization: X ~ Normal(mean, cov)
 *
 * Outputs: (dim_i, X[i]) for i=1:D
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
	long *mu_dimension_id;
	double *mu_prior_value;

	long *variance_dimension_id1;
	long *variance_dimension_id2;
	double *variance_dimension_value;
};

struct RecordOut {
	long *out_dimension_id;
	double *out_dimension_value;
};


int runningError;
void handler (const char * reason,
			  const char * file,
			  int line,
			  int gsl_errno)
{
	runningError = 1;
	stringstream ss;
	string result = "reason:";
	result += reason;
	result += "\t file:";
	result += file;
	result += "\t line:";
	ss.str("");
	ss << line;
	result += ss.str();
	result += "\t gsl_errno:";
	ss.str("");
	ss << gsl_errno;
	result += ss.str();
	result += "\r\n";

	ofstream fileWriter;

	
	fileWriter.open("/tmp/error.report", ios::app);

	fileWriter<< result.c_str();
	fileWriter.close();
}


/**
* A multivariate normal VG function, with an arbitrary number of
* random variables.
*
* @author Luis
*/
class MultiNormal: public VGFunction {
private:
	gsl_rng *rng;

	// mean vector
	double *mu;
	
	// covariance matrix
	double **covariance;

	// random standard normal samples
	double *samples;
	
	// number of dimensions
	long currentDimensionSize;
	long maxDimensionSize;

	long outputIndex;

	bool active;

	/* -------------------- temp ----------------------- */
	long tempInputValue;
	long tempOutputValue;

public:
	MultiNormal()
	{
		// create the RNG.
		rng = gsl_rng_alloc(gsl_rng_mt19937);
		
		maxDimensionSize = DEAFULT_DIMENSION;
		
		mu = (double *)malloc(maxDimensionSize * sizeof(double));
		covariance = (double **)malloc(maxDimensionSize * sizeof(double *));
		for(int i = 0; i < maxDimensionSize; i++)
		{
			covariance[i] = (double *)malloc(maxDimensionSize * sizeof(double));
		}
		
		for(int i = 0; i < maxDimensionSize; i++)
        {
            memset(covariance[i], 0, maxDimensionSize * sizeof(double));
        }		

		samples = (double *)malloc(maxDimensionSize * sizeof(double));

		currentDimensionSize = 0;
		outputIndex = 0;
		runningError = -1;
	}

	~MultiNormal()
	{
		// destroy the RNG
		gsl_rng_free(rng);

		free(mu);
		for(int i = 0; i < maxDimensionSize; i++)
		{
			free(covariance[i]);
		}
		free(covariance);

		free(samples);
	}

	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial() {

		// set active again.
		active = (currentDimensionSize > 0);
		outputIndex = 0;
		runningError = -1;
	}

	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams() {

		currentDimensionSize = 0; /* number of dimensions */
		
		/* We do need to clear the parameters for following parameters:
		 * mu
		 * Since they will be overwritten automatically.
		 */
		for(int i = 0; i < maxDimensionSize; i++)
		{
			memset(covariance[i], 0, maxDimensionSize * sizeof(double));
		}
		
		outputIndex = 0;
		active = false;
		runningError = -1;
	}

	void takeParams(RecordIn &input)
	{
		if (input.mu_dimension_id != NULL && 
			input.mu_prior_value != NULL)
		{
			tempInputValue = *input.mu_dimension_id;
			if(currentDimensionSize < tempInputValue + 1)
			{
				currentDimensionSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxDimensionSize)
			{
				int previousClusterSize = maxDimensionSize;
				maxDimensionSize = (tempInputValue + (tempInputValue >> 1) + 1);

				reallocateMemoryByDimension(previousClusterSize, maxDimensionSize);
			}

			mu[tempInputValue] = *input.mu_prior_value;
		}

		if(input.variance_dimension_id1 != NULL &&
			input.variance_dimension_id2 != NULL &&
			input.variance_dimension_value != NULL)
		{
			tempInputValue = *input.variance_dimension_id1;
			if(tempInputValue < *input.variance_dimension_id2)
			{
				tempInputValue = *input.variance_dimension_id2;
			}

			if(currentDimensionSize < tempInputValue + 1)
			{
				currentDimensionSize = tempInputValue + 1;
			}
			

			if(tempInputValue >= maxDimensionSize)
			{
				int previousClusterSize = maxDimensionSize;
				maxDimensionSize = (tempInputValue + (tempInputValue >> 1) + 1);

				reallocateMemoryByDimension(previousClusterSize, maxDimensionSize);
			}

			covariance[*input.variance_dimension_id1][*input.variance_dimension_id2]
					= *input.variance_dimension_value;
		}

		active = (currentDimensionSize > 0);
	}

	void reallocateMemoryByDimension(long previousDimensionSize, long newMaxDimensionSize)
	{
		//reallocate the memory for associated data structures when changing cluster num.
		mu = (double *)realloc(mu, newMaxDimensionSize * sizeof(double));
		covariance = (double **)realloc(covariance, newMaxDimensionSize * sizeof(double *));
		for(int i = previousDimensionSize ; i < newMaxDimensionSize; i++)
		{
			covariance[i] = NULL;
		}

		for(int i = 0; i < newMaxDimensionSize; i++)
		{
			covariance[i] = (double *)realloc(covariance[i], newMaxDimensionSize * sizeof(double));
		}
		
		//re-initialize the increased area.
		for(int i = 0; i < newMaxDimensionSize; i++)
		{
			for(int j = 0; j < newMaxDimensionSize; j++)
			{
				if(i >= previousDimensionSize || j >= previousDimensionSize)
				{
					covariance[i][j] = 0;
				}
			}
		}

		samples = (double *)realloc(samples, newMaxDimensionSize * sizeof(double));
	}

	int outputVals(RecordOut &output) {
		if(!active)
		{
			return 0;
		}
		
		// now, is it the first dimension?
		if (outputIndex == 0) {
			while(true) {
				runningError = -1;
				gsl_set_error_handler(&handler);

				gsl_matrix * sigma = gsl_matrix_calloc(currentDimensionSize, currentDimensionSize);
				for (int i = 0; i < currentDimensionSize; ++i) {
					for (int j = 0; j < currentDimensionSize; ++j) {
						gsl_matrix_set(sigma, i, j, covariance[i][j]);
					}
				}

				// now, compute the cholesky decomposition of sigma
				//int ret = gsl_linalg_cholesky_decomp(&sigma.matrix);
				gsl_linalg_cholesky_decomp(sigma);
				if (runningError == 1) {
					gsl_matrix_free(sigma);
					continue;
				}

				// now, we'll get an array of d independent standard normal values
				double *N = (double *)malloc(sizeof(double) * currentDimensionSize);

				for (int i = 0; i < currentDimensionSize; i++) {
					N[i] = gsl_ran_gaussian(rng, 1.0);
					//printf("%f ", N[i]);
				}
				//samples = (double *)malloc(sizeof(double) * d);
				for (int i = 0; i < currentDimensionSize; i++) {
					samples[i] = 0;
					for (int j=0; j <= i; j++) {
						//samples[i] += gsl_matrix_get(sigma.matrix, i, j) * N[i];
						samples[i] += gsl_matrix_get(sigma, i, j) * N[j];
						//fprintf(stderr, "inloop: %lf \n", samples[i]);
					}
					samples[i] += mu[i];
				}

				// free the array of samples
				gsl_matrix_free(sigma);
				free(N);
				//gsl_set_error_handler(oldhandler);

				if(runningError == 1) {
					continue;
				} else {
					break;
				}
			}
		}
		
		tempOutputValue = outputIndex;		
		output.out_dimension_id = &tempOutputValue;
		output.out_dimension_value = &samples[outputIndex];
		
		outputIndex++;
		if(outputIndex >= currentDimensionSize)
		{
			active = false;
		}

		return 1;
	}

	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {

		return (VGSchema){5, 
						 {INT, DOUBLE,  INT, INT, DOUBLE}, 
						 {"mu_dimension_id", "mu_prior_value", 
						 "variance_dimension_id1", "variance_dimension_id2", "variance_dimension_value"},
						 {0, 0, 0, 0, 0}
		       };
	}

	VGSchema outputSchema() {
		return (VGSchema){2, 
						  {INT, DOUBLE}, 
						  {"out_dimension_id", "out_dimension_value"}, 
						  {0, 1}
		       };
	}

	const char *getName() {
		return "MultiNormal";
	}
};

VGFunction *create() {
	return (new MultiNormal());
}

void destroy(VGFunction *vgFunction) {
	delete (MultiNormal *)vgFunction;
}

