

/*****************************************************************************
 *                                                                           *
 *  Copyright 2014 Rice University                                           *
 *                                                                           *
 *  Licensed under the Apache License, Version 2.0 (the "License");          *
 *  you may not use this file except in compliance with the License.         *
 *  You may obtain a copy of the License at                                  *
 *                                                                           *
 *      http://www.apache.org/licenses/LICENSE-2.0                           *
 *                                                                           *
 *  Unless required by applicable law or agreed to in writing, software      *
 *  distributed under the License is distributed on an "AS IS" BASIS,        *
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. *
 *  See the License for the specific language governing permissions and      *
 *  limitations under the License.                                           *
 *                                                                           *
 *****************************************************************************/


#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#define INITIAL_DOCS   10000
#define INITIAL_TOPICS 100
#define INITIAL_WORDS  1000
#define INITIAL_COUNTS 50
#define NUM_ITERS 10

//#define DEBUG

/**
 * A Group-based document sampling VG function, used for the LDA.
 *
 * All the inputs used the 'pair' scalar function to reduce the number
 * of attributes passed through the VGWrapper.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *doc_topic_pair;
  double *topic_prob;

  long *topic_word_pair;
  double *word_prob;

  long *doc_word_pair;
  long *word_count;

  double *topic_prior;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *out_rec_type;  // this is either 1: produced_word, or 2: topic_mixture
  long *out_doc_id;
  long *out_word_id;
  long *out_topic_id;
  long *out_count;
  double *out_prob;
};

int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno) {
  
  runningError = 1;
  fprintf(stderr, "Some error happened in Multi_Dir_DocWordTopic...\n");
}

class Multi_Dir_DocWordTopic: public VGFunction {
private:

  // RNG
  gsl_rng *rng;

  // probability of each (doc,topic) pair; size=docCapacity * topicCapacity
  double **docTopicProb;

  // probability of each (word,topic) pair; size=wordCapacity * topicCapacity
  double **wordTopicProb;
  
  // count for each (doc,word) pair; size=docCapacity * wordCountCapacity
  unsigned int **docWordCount;

  // map from wordID to wordCount position; size=wordCapacity
  int *wordToPos;

  // map from wordCount position to wordID; size=wordCountCapacity
  int *posToWord;

  // holds the matrix telling which topic produced each word
  unsigned int **producedWord;

  // holds the matrix telling how many words were produced by each topic
  double *counts;

  // this is the prior used for the topic mixture
  double priorVal;

  // capacities
  int docCapacity;
  int topicCapacity;
  int wordCapacity;
  int wordCountCapacity;

  // actual numbers for each group.
  int numDocs;
  int numTopics;
  int numWordCounts;

  // active Y/N
  bool active;

  // iterators for the VGF
  int currentDoc;
  int currentWord;
  int currentTopic;
  int topicToReturn;

  // temporary structure for probabilities; size=topicCapacity
  double *tempProb;

  // temporary output structure for counts; size=topicCapacity
  unsigned int *tempOutput;

  long recType;
  long tempDoc;
  long tempWord;
  long tempTopic;
  long tempCount;
  double tempTopicProb;

  // -------------

  void increaseDocs(int docID) {

    // see if this is a new document
    if (docID >= numDocs) {

#ifdef DEBUG
      printf("Increasing number of documents from %i to %i\n", numDocs, docID + 1);
      fflush(stdout);
#endif

      numDocs = docID + 1;
    }

    // see if we need to expand the array.
    if (numDocs >= docCapacity) {

      int oldCapacity = docCapacity;
      docCapacity = numDocs + (numDocs >> 1) + 1;

#ifdef DEBUG
      printf("Increasing document capacity from %i to %i. ", oldCapacity, docCapacity);
      fflush(stdout);
#endif

      // increase the docs.
      docTopicProb = (double **)realloc(docTopicProb, sizeof(double *) * docCapacity);
      docWordCount = (unsigned int **)realloc(docWordCount, sizeof(unsigned int *) * docCapacity);

      // increase the other arrays.
      for (int i=oldCapacity;i<docCapacity;i++) {
	docTopicProb[i] = (double *)malloc(sizeof(double) * topicCapacity);
	docWordCount[i] = (unsigned int *)malloc(sizeof(unsigned int) * wordCountCapacity);

	for (int j=0;j<topicCapacity;j++) {
	  docTopicProb[i][j] = 0.0;
	}

	for (int j=0;j<wordCountCapacity;j++) {
	  docWordCount[i][j] = 0;
	}
      }

#ifdef DEBUG
      printf("Done.\n");
      fflush(stdout);
#endif

    }
  }

  void increaseTopics(int topicID) {

    // see if we have a new topic.
    if (topicID >= numTopics) {
#ifdef DEBUG
      printf("Increasing number of topics from %i to %i\n", numTopics, topicID + 1);
      fflush(stdout);
#endif

      numTopics = topicID + 1;
    }

    // see if we need to expand the array
    if (numTopics >= topicCapacity) {

      int oldCapacity = topicCapacity;
      topicCapacity = numTopics + (numTopics >> 1) + 1;

#ifdef DEBUG
      printf("Increasing topic capacity from %i to %i. ", oldCapacity, topicCapacity);
      fflush(stdout);
#endif

      // increase the temporary output guys
      tempProb = (double *)realloc(tempProb, sizeof(double) * topicCapacity);

      // increase the topics in the docs array
      for (int i=0;i<docCapacity;i++) {

	docTopicProb[i] = (double *)realloc(docTopicProb[i], sizeof(double) * topicCapacity);

	for (int j=oldCapacity;j<topicCapacity;j++) {
	  docTopicProb[i][j] = 0.0;
	}
      }

      // increase the topics in the words array
      for (int i=0;i<wordCapacity;i++) {

	wordTopicProb[i] = (double *)realloc(wordTopicProb[i], sizeof(double) * topicCapacity);

	for (int j=oldCapacity;j<topicCapacity;j++) {
	  wordTopicProb[i][j] = 0.0;
	}
      }

#ifdef DEBUG
      printf("Done.\n");
      fflush(stdout);
#endif
    }
  }

  void increaseWords(int wordID) {

    // check if this is a new word.
    if (wordID >= wordCapacity) {

      int oldCapacity = wordCapacity;
      wordCapacity = wordID + (wordID >> 1) + 1;

#ifdef DEBUG
      printf("Increasing dictionary size from %i to %i. ", oldCapacity, wordCapacity);
      fflush(stdout);
#endif
      
      wordTopicProb = (double **)realloc(wordTopicProb, sizeof(double *) * wordCapacity);
      wordToPos = (int *)realloc(wordToPos, sizeof(int) * wordCapacity);
      for (int i=oldCapacity;i<wordCapacity;i++) {

	wordToPos[i] = -1;

	wordTopicProb[i] = (double *)malloc(sizeof(double) * topicCapacity);
	for (int j=0;j<topicCapacity;j++) {
	  wordTopicProb[i][j] = 0.0;
	}
      }


#ifdef DEBUG
      printf("Done.\n");
      fflush(stdout);
#endif

    }
  }

  void increaseWordCounts(int wordID) {

    // see if this is unassigned
    if (wordToPos[wordID] < 0) {

#ifdef DEBUG
      printf("Assigning count position %i to word %i.\n", numWordCounts, wordID);
      fflush(stdout);
#endif

      wordToPos[wordID] = numWordCounts;
      posToWord[numWordCounts] = wordID;

      numWordCounts++;
    }

    // see if we need to resize.
    if (numWordCounts >= wordCountCapacity) {

      int oldCapacity = wordCountCapacity;
      wordCountCapacity = numWordCounts + (numWordCounts >> 1) + 1;

#ifdef DEBUG
      printf("Increasing count capacity from %i to %i. ", oldCapacity, wordCountCapacity);
      fflush(stdout);
#endif

      posToWord = (int *)realloc(posToWord, sizeof(int) * wordCountCapacity);

      for (int i=oldCapacity;i<wordCountCapacity;i++) {
	posToWord[i] = -1;
      }

      for (int i=0;i<docCapacity;i++) {

	docWordCount[i] = (unsigned int *)realloc(docWordCount[i], sizeof(unsigned int) * wordCountCapacity);

	for (int j=oldCapacity;j<wordCountCapacity;j++) {
	  docWordCount[i][j] = 0;
	}
      }

#ifdef DEBUG
      printf("Done.\n");
      fflush(stdout);
#endif
    }
  }

  int unpair1(long val) {
    return (int)(val >> 32);
  }

  int unpair2(long val) {
    return (int)(val);
  }


public:

  /** 
   * Constructor. Use this to declare your RNG and other important
   * structures.
   */
  Multi_Dir_DocWordTopic() {

    priorVal = 0.0;

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
    runningError = -1;

    // create our structures
    docCapacity = INITIAL_DOCS;
    topicCapacity = INITIAL_TOPICS;
    wordCapacity = INITIAL_WORDS;
    wordCountCapacity = INITIAL_COUNTS;

    tempProb = (double *)malloc(sizeof(double) * topicCapacity);
    tempOutput = (unsigned int *)malloc(sizeof(unsigned int) * topicCapacity);

    docTopicProb = (double **)malloc(sizeof(double *) * docCapacity);
    docWordCount = (unsigned int **)malloc(sizeof(unsigned int *) * docCapacity);    
    for (int i=0;i<docCapacity;i++) {
      docTopicProb[i] = (double *)malloc(sizeof(double) * topicCapacity);
      docWordCount[i] = (unsigned int *)malloc(sizeof(unsigned int) * wordCountCapacity);
    }

    posToWord = (int *)malloc(sizeof(int) * wordCountCapacity);
    wordToPos = (int *)malloc(sizeof(int) * wordCapacity);
    wordTopicProb = (double **)malloc(sizeof(double *) * wordCapacity);
    for (int i=0;i<wordCapacity;i++) {
      wordTopicProb[i] = (double *)malloc(sizeof(double) * topicCapacity);
    }

    producedWord = 0;
    counts = 0;
    currentDoc = 0;
    currentWord = 0;
    currentTopic = 0;
    topicToReturn = 0;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Multi_Dir_DocWordTopic() {

    // destroy all the structures
    for (int i=0;i<wordCapacity;i++) {
      free(wordTopicProb[i]);
    }

    free(wordTopicProb);
    free(wordToPos);
    free(posToWord);

    for (int i=0;i<docCapacity;i++) {
      free(docTopicProb[i]);
      free(docWordCount[i]);
    }

    free(docWordCount);
    free(docTopicProb);

    free(tempProb);
    free(tempOutput);

    // destroy the RNG
    gsl_rng_free(rng);

    if (producedWord != 0) {
	for (int word = 0; word < numWordCounts; word++) {
		if (producedWord[word] != 0)
			free (producedWord[word]);
	}
	free (producedWord);
	producedWord = 0;
    }

    if (counts != 0) {
	free (counts);
        counts = 0;
    }
  }

  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);

#ifdef DEBUG
    printf("New seed: %ld.\n", seed);
    fflush(stdout);
#endif
  }

  /** 
   * Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). 
   */
  void finalizeTrial() {

#ifdef DEBUG
    printf("Trial finalized.\n");
    fflush(stdout);
#endif

    currentDoc = 0;
    currentWord = 0;
    currentTopic = 0;
    topicToReturn = 0;

    if (producedWord != 0) {
	for (int word = 0; word < numWordCounts; word++) {
		if (producedWord[word] != 0)
			free (producedWord[word]);
	}
	free (producedWord);
	producedWord = 0;
    }

    if (counts != 0) {
	free (counts);
        counts = 0;
    }

    active = (numDocs > 0) && (numWordCounts > 0) && (numTopics > 0);
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("Clearing all params...\n");
    fflush(stdout);
#endif

    for (int i=0;i<docCapacity;i++) {
      for (int j=0;j<topicCapacity;j++) {
	docTopicProb[i][j] = 0.0;
      }
      
      for (int j=0;j<wordCountCapacity;j++) {
	docWordCount[i][j] = 0;
      }
    }

    for (int i=0;i<wordCapacity;i++) {
      wordToPos[i] = -1;

      for (int j=0;j<topicCapacity;j++) {
	wordTopicProb[i][j] = 0.0;
      }
    }

    for (int i=0;i<wordCountCapacity;i++) {
      posToWord[i] = -1;
    }

    numDocs = 0;
    numTopics = 0;
    numWordCounts = 0;
  }

  void takeParams(RecordIn &input) {

#ifdef DEBUG
    printf("TakeParams ");
    fflush(stdout);
#endif

    // doc-topic-prob? 
    if (input.doc_topic_pair != NULL && input.topic_prob != NULL) {

      // unpack the pair.
      int docID = unpair1(*input.doc_topic_pair);
      int topicID = unpair2(*input.doc_topic_pair);

#ifdef DEBUG
      printf("1: %ld (%i,%i) -> %f.\n", *input.doc_topic_pair, docID, topicID, *input.topic_prob);
      fflush(stdout);
#endif
      
      // increase the structures, if necessary.
      increaseTopics(topicID);
      increaseDocs(docID);

      // save the probability
      docTopicProb[docID][topicID] = *input.topic_prob;
    }

    // topic-word-prob?
    if (input.topic_word_pair != NULL && input.word_prob != NULL) {

      // unpack the pair
      int topicID = unpair1(*input.topic_word_pair);
      int wordID = unpair2(*input.topic_word_pair);

#ifdef DEBUG
      printf("2: %ld (%i,%i) -> %f.\n", *input.topic_word_pair, topicID, wordID, *input.word_prob);
      fflush(stdout);
#endif
      
      increaseTopics(topicID);
      increaseWords(wordID);

      // save the probability
      wordTopicProb[wordID][topicID] = *input.word_prob;
    }

    // doc-word-count?
    if (input.doc_word_pair != NULL && input.word_count != NULL) {

      // unpack the pair
      int docID = unpair1(*input.doc_word_pair);
      int wordID = unpair2(*input.doc_word_pair);

#ifdef DEBUG
      printf("3: %ld (%i,%i) -> %ld.\n", *input.doc_word_pair, docID, wordID, *input.word_count);
      fflush(stdout);
#endif      

      increaseWords(wordID);
      increaseWordCounts(wordID);
      increaseDocs(docID);

      docWordCount[docID][wordToPos[wordID]] = (unsigned int)*input.word_count;
    }

    if (input.topic_prior != NULL) {
      priorVal = (double)*input.topic_prior;
    }

    // are we active?
    active = (numDocs > 0) && (numTopics > 0) && (numWordCounts > 0);

#ifdef DEBUG
    printf("TakeParams done. Active = %i\n", active);
    fflush(stdout);
#endif
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) {

    // set the output record
    output.out_rec_type = &recType;
    output.out_doc_id = &tempDoc;
    output.out_word_id = &tempWord;
    output.out_topic_id = &tempTopic;
    output.out_count = &tempCount;
    output.out_prob = &tempTopicProb;

    // are we active?
    if (!active) {
      return 0;
    }

    // if no producedWord, create it
    if (producedWord == 0) {
                producedWord = (unsigned int **) malloc (numWordCounts * sizeof (unsigned int *));
                for (int i = 0; i < numWordCounts; i++) {
                        producedWord[i] = 0;
                }
    }

    // if no counts vector, create it
    if (counts == 0)
    		counts = (double *) malloc (sizeof(double) * numTopics);

    while (true) {

      // are we done with the topics for this word?
      if (currentTopic >= numTopics) {

	// reset and advance to the next word.
	currentTopic = 0;
	currentWord++;
      }

      // are we done with the words for this document?
      if (currentWord >= numWordCounts) {

	// reset and advance to the next document.
	currentWord = 0;
	topicToReturn = 0;
	currentDoc++;
      }

      // are we done with all the documents in this group?
      if (currentDoc >= numDocs) {

	// then we move out.
	active = false;
	return 0;
      }

      // check if this is the first time we are getting info
      if (topicToReturn == 0) {

        // this holds the per-topic word counts
	for (int topic = 0; topic < numTopics; topic++) {
		counts[topic] = priorVal;
	}

        // this does a little mini MCMC that does both the topic mixture and the produced word
	for (int iter = 0; iter < NUM_ITERS; iter++) {

		// update the producedWord matrix
		for (int word = 0; word < numWordCounts; word++) {

			// if no word: ignore
			if (docWordCount[currentDoc][word] == 0) {
				continue;
			}

			// get the word-topic probabilities
			for (int i=0;i<numTopics;i++) {
				tempProb[i] = docTopicProb[currentDoc][i] * wordTopicProb[posToWord[word]][i];
			}	

			// make sure the column in the producedWord matrix has memory
			if (producedWord[word] == 0)
				producedWord[word] = (unsigned int *) malloc(sizeof(unsigned int) * numTopics);
			
	  		// sample from a multinomial distribution
			gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
	  		gsl_ran_multinomial(rng, numTopics, docWordCount[currentDoc][word], tempProb, producedWord[word]);
	  		gsl_set_error_handler(oldhandler);
		}

		// sum up the produced matrix
		for (int topic = 0; topic < numTopics; topic++) {
			for (int word = 0; word < numWordCounts; word++) {
				if (docWordCount[currentDoc][word] == 0) {
					continue;
				} else {
					counts[topic] += producedWord[word][topic];
				}
			}
		}

		// update the topic probs
		gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
  		gsl_ran_dirichlet(rng, numTopics, counts, docTopicProb[currentDoc]);
  		gsl_set_error_handler(oldhandler);
	}

      }

      // if we got here, it means that we just give out the topic info for this doc
      if (topicToReturn < numTopics) {

	  // if we have one, then we'll return!
          recType = 2;
	  tempDoc = currentDoc;
	  tempWord = -1;
	  tempTopic = topicToReturn; 
	  tempCount = -1;
          tempTopicProb = docTopicProb[currentDoc][topicToReturn]; 
          topicToReturn++;

	  // give back the record.
	  return 1;

      // is this word worth sampling?
      } else if (docWordCount[currentDoc][currentWord] > 0) {
      
	// everything is sampled, so we find the next topic with a
	// positive count.
	if (producedWord[currentWord][currentTopic] > 0) {

	  // if we have one, then we'll return!
          recType = 1;
	  tempDoc = currentDoc;
	  tempWord = posToWord[currentWord];
	  tempTopic = currentTopic;
	  tempCount = producedWord[currentWord][currentTopic];
          tempTopicProb = -1.0; 

	  // move to the next topic
	  currentTopic++;

	  // give back the record.
	  return 1;
	}

	// otherwise, we advance to the next topic and loop back.
	currentTopic++;

      } else {

	// otherwise, try the next word.
	currentWord++;
	currentTopic = 0;
      }    
    }
  }

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){
      7, 
	{INT, DOUBLE, INT, DOUBLE, INT, INT, DOUBLE}, 
	  {"in_doc_topic_pair", "in_topic_prob", "in_word_topic_pair", "in_word_prob", "in_doc_word_pair", "in_word_count", "in_prior"}, 
	    {0, 1, 0, 1, 0, 1, 0},
	      };
  }

  VGSchema outputSchema() {

    return (VGSchema){6, {INT, INT, INT, INT, INT, DOUBLE}, {"out_rec_type", "out_doc_id", "out_word_id", "out_topic_id", "out_count", "out_topic_prob"}, {0, 0, 0, 0, 1, 1}};
  }

  const char *getName() {
    return "Multi_Dir_DocWordTopic";
  }
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Multi_Dir_DocWordTopic());
}

void destroy(VGFunction *vgFunction) {
  delete (Multi_Dir_DocWordTopic *)vgFunction;
}
