#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#define INITIAL_SIZE 1

/**
 * A Multinomial VG function.
 * Used by the GMM and LDA queries.
 *
 * Inputs: (dim_i, P[i], NULL) for i=1:K
 *         (NULL, NULL, n)
 *
 * Parameterization: X ~ Multinomial(P, n)
 *
 * Output: (dim_i, X[i]) for i=1:K
 */

/** 
* Input record, used by the takeParams() method.
* Values could be NULL, meaning they were not present in the tuple.
*/
struct RecordIn {
	long *dimension_id;
	double *probability;

	long *numTrials;
};

/**
* Output record, used by the outputVals().
* If any of the values is NULL, then the user must allocate space for
* it (the engine will de-allocate). 
*/
struct RecordOut {
	long *dimension_id;
	long *out_num;
};


int runningError;
void handler (const char * reason,
			  const char * file,
			  int line,
			  int gsl_errno)
{
	runningError = 1;
}


class Multinomial: public VGFunction {
private:
	gsl_rng *rng;
	bool active;

	// for input parameters
	double *probability; /* value of pi vector for multinomial */
	long numTrials; /* number of trials */

	// for output values
	long outputValue;
	long currentOutDimensionId;
	unsigned int *outputNum; /* value of value vector for multinomial */

	// for temp
	long currentInputDimensionId;
	long currentDimensionSize;
	long maxDimensionSize;

	/* ----------------------- temp value ----------------------*/
	long tempOutputValue;
public:

	/** Constructor. Use this to declare your RNG and other
	* important structures .
	*/
	Multinomial() {
		// create the RNG.
		rng = gsl_rng_alloc(gsl_rng_mt19937);

		// set the initial size of each container as the INITIAL_SIZE.
		maxDimensionSize = INITIAL_SIZE;

		probability = (double *)malloc(maxDimensionSize * sizeof(double));
		outputNum = (unsigned int *)malloc(maxDimensionSize * sizeof(unsigned int));
		
		runningError = -1;
	}

	/** Destructor. Deallocate everything from the constructor. */
	~Multinomial() {
		// destroy the RNG
		gsl_rng_free(rng);

		// destroy the arrays
		free(probability);
		free(outputNum);
	}

	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial() {

		// set active again.
		active = (currentDimensionSize > 0);
		currentOutDimensionId = 0;
	}

	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams() {
		active = false;
		currentDimensionSize = 0;
		currentOutDimensionId = 0;
	}

	void takeParams(RecordIn &input) {

	        // in case we found a bad input.
	        if (currentDimensionSize < 0) {
		  return;
		}

		if(input.dimension_id != NULL && input.probability != NULL)
		{
			currentInputDimensionId = *input.dimension_id;

			if(currentDimensionSize < currentInputDimensionId + 1)
			{
				currentDimensionSize = currentInputDimensionId + 1;
			}

			if(currentInputDimensionId >= maxDimensionSize)
			{
				maxDimensionSize = (currentInputDimensionId + (currentInputDimensionId >> 1) + 1);
				probability = (double *)realloc(probability, maxDimensionSize * sizeof(double));
				outputNum = (unsigned int *)realloc(outputNum, maxDimensionSize * sizeof(unsigned int));
			}

			probability[currentInputDimensionId] = *input.probability;

			// if the value is wrong, get out of here.
			if (isnan(*input.probability) || isinf(*input.probability) || *input.probability < 0.0) {
			  currentDimensionSize = -1;
			  return;
			}
		}

		if(input.numTrials != NULL){
			numTrials = *input.numTrials;
		}

		active = (currentDimensionSize > 0);
	}

	/** 
	* Produces the sample values. Returns 1 if there are more
	* records to be produced for the current sample, 0 otherwise. 
	*/
	int outputVals(RecordOut &output) {
		if(!active)
		{
			return 0;
		}

		if(currentOutDimensionId == 0)
		{

		  while(true) {
				runningError = -1;
				gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
				gsl_ran_multinomial(rng, 
					currentDimensionSize, 
					(unsigned int)numTrials, 
					probability, 
					outputNum);

				gsl_set_error_handler(oldhandler);
				if(runningError == -1) break;
		  }
		}

		//fprintf(stderr, "thera[%d] = %f", oindex, theta[oindex]);
		tempOutputValue = currentOutDimensionId;
		output.dimension_id = &tempOutputValue;
		outputValue = (long)outputNum[currentOutDimensionId];
		output.out_num = &outputValue;

		currentOutDimensionId ++;

		if (currentOutDimensionId >= currentDimensionSize) { /* no more values */
			active = false;
		}

		return (1);
	}

	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {

		return (VGSchema){3, {INT, DOUBLE, INT}, {"id", "probability", "num"}, {0, 1, 1}};
	}

	VGSchema outputSchema() {

		return (VGSchema){2, {INT, INT}, {"out_id", "out_count"}, {0, 1}};
	}

	const char *getName() {
		return "Multinomial";
	}
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
	return(new Multinomial());
}

void destroy(VGFunction *vgFunction) {
	delete (Multinomial *)vgFunction;
}
