#include "VGFunction.h"
#include <string.h>
#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_errno.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

#define MAX_CLUST 1
#define MAX_DIM 1

using namespace std;

/**
 * This is just a modification of Cai's Multinomial cluster membership
 * for the GMM.  Instead of returning several tuples with the 0/1
 * membership bitmap, it returns a single tuple with the assigned
 * cluster ID.
 *
 * Use this one only when you have numMC=1. Otherwise, the joins will
 * cause a split and might not be as efficient!!!
 *
 * Inputs: (cdim_i, C[i], NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) for i=1:K
 *         (NULL, NULL, dval_i, Y[i], NULL, NULL, NULL, NULL, NULL, NULL, NULL) for i=1:D
 *         (NULL, NULL, NULL, NULL, mclu_k, mclu_i, M[k,i], NULL, NULL, NULL, NULL) for i=1:k, i=1:D
 *         (NULL, NULL, NULL, NULL, NULL, NULL, NULL, vclu_k, vclu_i, vclu_j, V[k,i,j]) for i=1:K, i=1:D, j=1:D
 *
 * Parameterization: X ~ Multinomial(C[k] * N(M[k], V[k]), 1)
 *
 * Outputs: (X)
 */


/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
	long *prior_cluster_id;
	double *prior_cluster_probability;

	long *data_dimension_id;
	double *data_dimension_value;

	long *mean_cluster_id;
	long *mean_cluster_dimension_id;
	double *mean_cluster_dimension_value;

	long *covariance_cluster_id;
	long *covariance_cluster_dimension_id_1;
	long *covariance_cluster_dimension_id_2;
	double *covariance_cluster_dimension_value;
};


/**
* Output record, used by the outputVals().
* If any of the values is NULL, then the user must allocate space for
* it (the engine will de-allocate). 
*/
struct RecordOut {
	long *cluster_id;
};


int runningError;
void handler (const char * reason,
			  const char * file,
			  int line,
			  int gsl_errno)
{
	runningError = 1;
	stringstream ss;
	string result = "reason: Membership";
	result += reason;
	result += "\t file:";
	result += file;
	result += "\t line:";
	ss.str("");
	ss << line;
	result += ss.str();
	result += "\t gsl_errno:";
	ss.str("");
	ss << gsl_errno;
	result += ss.str();
	result += "\r\n";

	/***
	ofstream fileWriter;

	
	fileWriter.open("error.report", ios::trunc);

	fileWriter<< result.c_str();
	fileWriter.close();
	***/ 
	cerr << result;
}

class Multinomial_Membership_ID: public VGFunction {
private:
	gsl_rng *rng;

	/*---------------------state information --------------*/
	long maxClusterSize;
	long maxDimensionSize;

	long currentDimensionSize; /* number of dimensions */
	long currentClusterSize; /* number of clusters */

	//prior: <cluster_id, pi_id>
	double *cluster_pi;
	double *log10_probability;

	double *data;

	//cluster mean <cluser_id, mean>: here mean may be multiple dimension.
	double **cluster_mean_map;

	//cluster variance
	double ***cluster_covariance_map;

	/*---------------------result------------------------*/
	unsigned int *resultMap;
	long outputIndex;
	long outputNum;

	/*--------------------temp ---------------------------*/
	long tempInputValue;
	long tempOutputValue;

	bool active;

  /*** added by L ***/
  long gslSize;
  gsl_vector *gsl_data;
  gsl_matrix *gsl_covariance;
  gsl_vector *ym;

public:

	/** Constructor. Use this to declare your RNG and other
	* important structures .
	*/
	Multinomial_Membership_ID() {
		// create the RNG.
		rng = gsl_rng_alloc(gsl_rng_mt19937);

		// set the initial size of each container as the INITIAL_SIZE.
		maxClusterSize = MAX_CLUST;
		maxDimensionSize = MAX_DIM;

		//allocate the memory for prior.
		cluster_pi = (double *)malloc(maxClusterSize * sizeof(double));
		log10_probability = (double *)malloc(maxClusterSize * sizeof(double));

		//allocate the memory for data.
		data = (double *)malloc(maxDimensionSize * sizeof(double));

		//allocate the memory for cluster_mean_map.
		cluster_mean_map = (double **)malloc(maxClusterSize * sizeof(double *));
		for(int i = 0; i < maxClusterSize; i++)
		{
			cluster_mean_map[i] = (double *)malloc(maxDimensionSize * sizeof(double));
		}

		//allocate the memory for cluster covariance.
		cluster_covariance_map = (double ***)malloc(maxClusterSize * sizeof(double **));
		for(int i = 0; i < maxClusterSize; i++)
		{
			cluster_covariance_map[i] = (double **)malloc(maxDimensionSize * sizeof(double *));
			
			for(int j = 0; j < maxDimensionSize; j++)
			{
				cluster_covariance_map[i][j] = (double *)malloc(maxDimensionSize * sizeof(double));
			}
		}

		for(int i = 0; i < maxClusterSize; i++)
        {
            for(int j = 0; j < maxDimensionSize; j++)
            {
                memset(cluster_covariance_map[i][j], 0, maxDimensionSize * sizeof(double));
            }
        }

		//allocate the memory for the result.
		resultMap = (unsigned int *)malloc(maxClusterSize * sizeof(unsigned int));
		currentDimensionSize = 0;
		outputIndex = 0;
		
		runningError = -1;

		// allocate our GSL stuff
		gslSize = MAX_DIM;
		gsl_data = gsl_vector_alloc(gslSize);
		gsl_covariance = gsl_matrix_alloc(gslSize, gslSize);
		ym = gsl_vector_alloc(gslSize);
	}

	/** Destructor. Deallocate everything from the constructor. */
	~Multinomial_Membership_ID()
	{
		// destroy the RNG
		gsl_rng_free(rng);

		//destory the memory for cluster prior.
		free(cluster_pi);
		//destroy the memory for cluster probability.
		free(log10_probability);
		//destroy the memory for data.
		free(data);
		
		//destroy the memory for cluster means.
		for(int i = 0; i < maxClusterSize; i++)
		{
			free(cluster_mean_map[i]);
		}
		free(cluster_mean_map);

		//destroy the memory for cluster covariances.

		for(int i = 0; i < maxClusterSize; i++)
		{

			for(int j = 0; j < maxDimensionSize; j++)
			{
				free(cluster_covariance_map[i][j]);
			}
			free(cluster_covariance_map[i]);
		}
		free(cluster_covariance_map);

		//destroy the result.
		free(resultMap);


		gsl_vector_free(gsl_data);
		gsl_vector_free(ym);
		gsl_matrix_free(gsl_covariance);
	}

	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial() {

		// set active again.
		active = (currentClusterSize > 0);
		outputIndex = 0;
	}

	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams() {

		currentDimensionSize = 0; /* number of dimensions */
		currentClusterSize = 0; /* number of clusters */
		
		/* We do need to clear the parameters for following parameters:
		 * cluster_pi
		 * log10_probability
		 * data
		 * cluster_mean_map
		 * resultMap
		 * Since they will be overwritten automatically.
		 */
		for(int i = 0; i < maxClusterSize; i++)
		{
			for(int j = 0; j < maxDimensionSize; j++)
			{	
				memset(cluster_covariance_map[i][j], 0, maxDimensionSize * sizeof(double));
			}
		}
		
		outputIndex = 0;
		active = false;
	}

	void takeParams(RecordIn &input) {
		//prior
		if (input.prior_cluster_id != NULL && 
			input.prior_cluster_probability != NULL )
		{
			tempInputValue = *input.prior_cluster_id;

			if(currentClusterSize < tempInputValue + 1)
			{
				currentClusterSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxClusterSize)
			{
				int previousClusterSize = maxClusterSize;
				maxClusterSize = (tempInputValue + (tempInputValue >> 1) + 1);

				//reallocate the memory for associated data structures when changing cluster num.
				reallocateClusterMemoryByCluster(previousClusterSize, maxClusterSize);
			}

			cluster_pi[tempInputValue] = *input.prior_cluster_probability;		
		}

		if(input.data_dimension_id != NULL &&
			input.data_dimension_value != NULL)
		{
			tempInputValue = *input.data_dimension_id;

			if(currentDimensionSize < tempInputValue + 1)
			{
				currentDimensionSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxDimensionSize)
			{
				int previousDimensionSize = maxDimensionSize;
				maxDimensionSize = (tempInputValue + (tempInputValue >> 1) + 1);

				//reallocate the memory for associated data structures when changing dimensions.
				reallocateClusterMemoryByDimension(previousDimensionSize, maxDimensionSize);
			}

			data[tempInputValue] = *input.data_dimension_value;

			//cout << "input: " << tempInputValue << ": " << data[tempInputValue] << endl;
		}

		if (input.mean_cluster_id != NULL && 
			input.mean_cluster_dimension_id != NULL &&
			input.mean_cluster_dimension_value != NULL)
		{
			tempInputValue = *input.mean_cluster_id;

			if(currentClusterSize < tempInputValue + 1)
			{
				currentClusterSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxClusterSize)
			{
				int previousClusterSize = maxClusterSize;
				maxClusterSize = (tempInputValue + (tempInputValue >> 1) + 1);

				//reallocate the memory for associated data structures when changing cluster num.
				reallocateClusterMemoryByCluster(previousClusterSize, maxClusterSize);
			}

			tempInputValue = *input.mean_cluster_dimension_id;

			if(currentDimensionSize < tempInputValue + 1)
			{
				currentDimensionSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxDimensionSize)
			{
				int previousDimensionSize = maxDimensionSize;
				maxDimensionSize = (tempInputValue + (tempInputValue >> 1) + 1);

				//reallocate the memory for associated data structures when changing dimensions.
				reallocateClusterMemoryByDimension(previousDimensionSize, maxDimensionSize);
			}

			
			cluster_mean_map[*input.mean_cluster_id]
			                [*input.mean_cluster_dimension_id] 
							= *input.mean_cluster_dimension_value;

		}

		if (input.covariance_cluster_id != NULL &&
			input.covariance_cluster_dimension_id_1 != NULL &&
			input.covariance_cluster_dimension_id_2 != NULL &&
			input.covariance_cluster_dimension_value != NULL) 
		{
			tempInputValue = *input.covariance_cluster_id;

			if(currentClusterSize < tempInputValue + 1)
			{
				currentClusterSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxClusterSize)
			{
				int previousClusterSize = maxClusterSize;
				maxClusterSize = (tempInputValue + (tempInputValue >> 1) + 1);

				//reallocate the memory for associated data structures when changing cluster num.
				reallocateClusterMemoryByCluster(previousClusterSize, maxClusterSize);
			}

			tempInputValue = *input.covariance_cluster_dimension_id_1;
			if(tempInputValue < *input.covariance_cluster_dimension_id_2)
			{
				tempInputValue = *input.covariance_cluster_dimension_id_2;
			}

			if(currentDimensionSize < tempInputValue + 1)
			{
				currentDimensionSize = tempInputValue + 1;
			}

			if(tempInputValue >= maxDimensionSize)
			{
				int previousDimensionSize = maxDimensionSize;
				maxDimensionSize = (tempInputValue + (tempInputValue >> 1) + 1);

				//reallocate the memory for associated data structures when changing dimensions.
				reallocateClusterMemoryByDimension(previousDimensionSize, maxDimensionSize);
			}
			
			cluster_covariance_map[*input.covariance_cluster_id]
								  [*input.covariance_cluster_dimension_id_1]
								  [*input.covariance_cluster_dimension_id_2]
								 = *input.covariance_cluster_dimension_value;
		}

		active = (currentClusterSize > 0);
	}

	void reallocateClusterMemoryByCluster(long previousClusterSize, long newMaxClusterSize)
	{
		cluster_pi = (double *)realloc(cluster_pi, newMaxClusterSize * sizeof(double));
		log10_probability = (double *)realloc(log10_probability, newMaxClusterSize * sizeof(double));

		cluster_mean_map = (double **)realloc(cluster_mean_map, newMaxClusterSize*sizeof(double *));
		for(int i = previousClusterSize; i < newMaxClusterSize; i++)
		{
			cluster_mean_map[i] = (double *)malloc(maxDimensionSize * sizeof(double));
		}

		cluster_covariance_map = (double ***)realloc(cluster_covariance_map, newMaxClusterSize*sizeof(double **));
		for(int i = previousClusterSize; i < newMaxClusterSize; i++)
		{
			cluster_covariance_map[i] = (double **)malloc(maxDimensionSize * sizeof(double *));
			
			for(int j = 0; j < maxDimensionSize; j++)
			{
				cluster_covariance_map[i][j] = (double *)malloc(maxDimensionSize * sizeof(double));
			}
		}

		//re-initialize the increased area.
		for(int i = previousClusterSize; i < newMaxClusterSize; i++)
		{
			for(int j = 0; j < maxDimensionSize; j++)
			{
				for(int k = 0; k < maxDimensionSize; k++)
				{
					cluster_covariance_map[i][j][k] = 0;
				}
			}
		}

		resultMap = (unsigned int *)realloc(resultMap, newMaxClusterSize * sizeof(unsigned int));
	}

	void reallocateClusterMemoryByDimension(long previousDimensionSize, long newMaxDimensionSize)
	{
		//for data
		data = (double *)realloc(data, newMaxDimensionSize * sizeof(double));
		
		//for cluster_means_map
		for(int i = 0; i < maxClusterSize; i++)
		{
			cluster_mean_map[i] = (double *)realloc(cluster_mean_map[i], 
				                           newMaxDimensionSize * sizeof(double));
		}
		//cluster variance
		for(int i = 0; i < maxClusterSize; i++)
		{
			cluster_covariance_map[i] = (double **)realloc(cluster_covariance_map[i],
										   newMaxDimensionSize * sizeof(double *));

			for(int j = previousDimensionSize; j < newMaxDimensionSize; j++)
			{
				cluster_covariance_map[i][j] = NULL;
			}

			for(int j = 0; j < newMaxDimensionSize; j++)
			{
				cluster_covariance_map[i][j] = (double *)realloc(cluster_covariance_map[i][j],
											newMaxDimensionSize * sizeof(double));
			}

			////re-initialize the increased area.
			for(int j = 0; j < newMaxDimensionSize; j++)
			{
				for(int k = 0; k < newMaxDimensionSize; k++)
				{
					if(j >= previousDimensionSize || k >= previousDimensionSize)
					{
						cluster_covariance_map[i][j][k] = 0;
					}
				}
			}
		}
	}


	void sample_membership() {

		runningError = -1;
		gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);

		for (int i = 0; i < currentClusterSize; ++i) {
		  log10_probability[i] = exp(log(cluster_pi[i]) + log_normpdf(i));
		}


		/****
		for(int i = 0; i < currentClusterSize; i++) {
			log10_probability[i] += log10_normpdf(i);
		}

		//find the maximum
		double maxValue = log10_probability[0];
		for(int i = 1; i < currentClusterSize; i++) {
			if (maxValue < log10_probability[i]) {
				maxValue = log10_probability[i];
			}
		}

		//normalizes the weights, and now log10_probability is the real probability
		for (int i = 0; i < currentClusterSize; i++) {
			log10_probability[i] = pow(10, log10_probability[i] - maxValue);
		}

		double sum = 0;
		for(int i = 0; i < currentClusterSize; i++) {
			sum += log10_probability[i];
		}

		for(int i = 0; i < currentClusterSize; i++) {
			log10_probability[i] = log10_probability[i]/sum;
			//fprintf(stderr, "log10_prob: %f \n", log10_probability[i]);
		}
		***/

		//multinomial choice
		gsl_ran_multinomial(rng, currentClusterSize, 1, log10_probability, resultMap);

		gsl_set_error_handler(oldhandler);

		if(runningError == 1)
		{
			int index = (int)(currentClusterSize*gsl_rng_uniform(rng));
			for(int i = 0; i < currentClusterSize; i++)
			{
				resultMap[i] = 0;
			}

			resultMap[index] = 1;
		}
	}

	double log_normpdf(int index) {

	  /* gsl_vector *gsl_data = gsl_vector_alloc(currentDimensionSize); */
	  
	  /* gsl_vector *gsl_mean = gsl_vector_alloc(currentDimensionSize); */
	  gsl_vector_view gsl_mean = gsl_vector_view_array(cluster_mean_map[index], currentDimensionSize);
	  
	  /* gsl_matrix *gsl_covariance = gsl_matrix_calloc(currentDimensionSize, currentDimensionSize); */

	  gsl_matrix_set_zero(gsl_covariance);
		
	  for (int i = 0; i < currentDimensionSize; ++i) {
	    gsl_vector_set(gsl_data, i, data[i]);
	    for (int j = 0; j <= i; ++j) {
	      gsl_matrix_set(gsl_covariance, i, j, cluster_covariance_map[index][i][j]);
	    }
	  }

	  gsl_linalg_cholesky_decomp(gsl_covariance);
	  double ax = 1.0;
	  for (int i=0;i<currentDimensionSize;i++) {
	    ax *= gsl_matrix_get(gsl_covariance, i, i);
	  }

	  ax *= ax;

	  gsl_linalg_cholesky_invert(gsl_covariance);

	  gsl_vector_sub(gsl_data, &gsl_mean.vector);

	  /* gsl_vector *ym = gsl_vector_alloc(currentDimensionSize); */
	  gsl_blas_dsymv(CblasUpper, 1.0, gsl_covariance, gsl_data, 0.0, ym);
	  
	  double ay;
	  gsl_blas_ddot(gsl_data, ym, &ay);

	  /*
	  gsl_vector_free(gsl_data);
	  gsl_matrix_free(gsl_covariance);
	  gsl_vector_free(ym);
	  */

	  return log(exp(-0.5*ay)/sqrt(pow(((double)2*3.1415926), (double)currentDimensionSize)*ax));

	  /***
		int s;
		double ax, ay;
		gsl_vector *ym;
		//gsl_matrix *winv = gsl_matrix_calloc(dataP->size(), dataP->size());
		gsl_matrix *winv = gsl_matrix_calloc(currentDimensionSize, currentDimensionSize);
		//gsl_permutation *p = gsl_permutation_alloc(dataP->size());
		gsl_permutation *p = gsl_permutation_alloc(currentDimensionSize);

		gsl_linalg_LU_decomp(gsl_covariance, p, &s);
		gsl_linalg_LU_invert(gsl_covariance, p, winv);
		ax = gsl_linalg_LU_det(gsl_covariance, s);
		gsl_matrix_free(gsl_covariance);
		gsl_permutation_free(p);

		gsl_vector_sub(gsl_data, gsl_mean);
		//ym = gsl_vector_alloc(dataP->size());
		ym = gsl_vector_alloc(currentDimensionSize);
		gsl_blas_dsymv(CblasUpper, 1.0, winv, gsl_data, 0.0, ym);
		gsl_matrix_free(winv);
		gsl_blas_ddot(gsl_data, ym, &ay);
		gsl_vector_free(gsl_data);
		gsl_vector_free(gsl_mean);
		gsl_vector_free(ym);

		//ay = exp(-0.5*ay)/sqrt(pow(((double)2*3.1415926), (double)dataP->size())*ax);
		ay = exp(-0.5*ay)/sqrt(pow(((double)2*3.1415926), (double)currentDimensionSize)*ax);

		return log10(ay);
	  **/
	}

	
	int outputVals(RecordOut &output) {
		if(!active)
		{
			return 0;
		}

		if (gslSize != currentDimensionSize) {
		  gsl_vector_free(gsl_data);
		  gsl_vector_free(ym);
		  gsl_matrix_free(gsl_covariance);

		  gsl_data = gsl_vector_alloc(currentDimensionSize);
		  ym = gsl_vector_alloc(currentDimensionSize);
		  gsl_covariance = gsl_matrix_alloc(currentDimensionSize, currentDimensionSize);

		  gslSize = currentDimensionSize;
		}

		sample_membership();

		tempOutputValue = 0;
		for (int i=0;i<currentClusterSize;i++) {
		  if (resultMap[i] > 0) {
		    tempOutputValue = i;
		    break;
		  }
		}

		output.cluster_id = &tempOutputValue;

		active = false;

		return (1);
	}
	
	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {

		return (VGSchema){11, 
						 {INT, DOUBLE,  INT, DOUBLE, INT, INT, DOUBLE, INT, INT, INT, DOUBLE}, 
						 {"c_cluster_id", "c_cluster_probability", 
						  "d_dimension_id", "d_data_value",
						  "cm_cluster_id", "cm_dimension_id", "cm_dimension_value",
						  "cc_cluster_id", "cc_dimension_id1", "cc_dimension_id2", "cc_dimension_value"}, 
						 {0, 1, 0, 0, 0, 0, 1, 0, 0, 0, 1}
		       };
	}

	VGSchema outputSchema() {
		return (VGSchema){1, 
						  {INT}, 
						  {"cluster_id"}, 
						  {1}
		       };
	}

	const char *getName() {
		return "multinomial_membership_id";
	}
};

VGFunction *create() {
	return (new Multinomial_Membership_ID());
}

void destroy(VGFunction *vgFunction) {
	delete (Multinomial_Membership_ID *)vgFunction;
}
