#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#define INITIAL_SIZE 50
//#define DEBUG

/**
 * A Muti-Multinomial VG function, used to get a purely document-based
 * Gibbs sampler for the LDA.
 *
 * Inputs: (dim_t, P[t], NULL, NULL, NULL, NULL, NULL) for topics t
 *         (NULL, NULL, dim_t, dim_w, P[t,w], NULL, NULL) for topics t, words w
 *         (NULL, NULL, NULL, NULL, NULL, dim_w, N[w]) for words w
 *
 * Parameterization: X[w] ~ Multinomial(P[t,w] * P[t], N[w]) for words w.
 *
 * Output: (dim_w, dim_t, X[w,t]) for words w, topics t.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *topic_id;
  double *topic_prob;

  long *wtopic_id;
  long *word_id;
  double *word_prob;

  long *cword_id;
  long *word_count;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *out_word_id;
  long *out_topic_id;
  long *out_count;
};


int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno) {
  
  runningError = 1;
  fprintf(stderr, "Some error happened in Multinomial_WordTopic...\n");
}


class Multinomial_WordTopic: public VGFunction {
private:

  // RNG
  gsl_rng *rng;

  // are we good to sample?
  bool active;

  // topic probabilities
  double *topicProb;

  // number of topics in the current document
  long numTopics;

  // size of the topicProb data structure and each element in
  // wordTopicProb.
  long topicCapacity;

  // topic-word probabilities, for all the words. size=[wordCapacity,topicCapacity]
  double **wordTopicProb;

  // total number of words.
  long numWords;

  // capacity of the topicWordProb array.
  long wordCapacity;

  // word counts, size=countCapacity
  long *wordCount;

  // total number of word counts for the current document
  long numCounts;

  // capacity of the wordCount array.
  long countCapacity;

  // map from wordID to index in the wordCount array, capacity=wordCapacity
  long *wordToPos;

  // map from wordCount index to wordID, capacity=countCapacity.
  long *posToWord;

  // iteration variables for outputVals
  long currentWord;
  long currentTopic;

  // the output structure for the multinomial RNG, size=countCapacity * topicCapacity
  unsigned int **outputCounts;

  // temporary structure for storing probabilities, size=topicCapacity
  double *tempProb;

  // temporary output values for the records.
  long tempWordID;
  long tempTopicID;
  long tempOutCount;

  // -----------

  void increaseTopics(long topicIn) {

    // update the number of topics based on the ID number.
    if (topicIn >= numTopics) {
      
#ifdef DEBUG
      printf("Num. topics increase from %ld to %ld.\n", numTopics, topicIn + 1);
      fflush(stdout);
#endif

      numTopics = topicIn + 1;
    }

    // update the capacity of the topic vectors.
    if (numTopics >= topicCapacity) {

      long oldCapacity = topicCapacity;

      topicCapacity = numTopics + (numTopics >> 1) + 1;

#ifdef DEBUG
      printf("Resizing topic array size from %ld to %ld...", oldCapacity, topicCapacity);
      fflush(stdout);
#endif

      // enlarge the array for the new capacity.
      topicProb = (double *)realloc(topicProb, topicCapacity * sizeof(double));
      tempProb = (double *)realloc(tempProb, sizeof(double) * topicCapacity);


#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);
#endif

      // set the empty spaces to zero.
      for (long i=oldCapacity;i<topicCapacity;i++) {
	topicProb[i] = 0.0;
      }

      // enlarge the arrays in the word-topic structures.
#ifdef DEBUG
      printf("Enlarging the word-topic arrays...");
      fflush(stdout);
#endif

      for (long i=0;i<wordCapacity;i++) {
	wordTopicProb[i] = (double *)realloc(wordTopicProb[i], sizeof(double) * topicCapacity);

	/***
	for (long j=oldCapacity;j<topicCapacity;j++) {
	  wordTopicProb[i][j] = 0.0;
	}
	***/
      }

      for (long i=0;i<countCapacity;i++) {
	outputCounts[i] = (unsigned int *)realloc(outputCounts[i], sizeof(unsigned int) * topicCapacity);
      }
      
#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);
#endif

    }
  }


  void increaseWords(long wordIn) {
    
    // check the number of words.
    if (wordIn >= numWords) {

#ifdef DEBUG
      printf("Increasing words from %ld to %ld.\n", numWords, wordIn + 1);
      fflush(stdout);
#endif

      numWords = wordIn + 1;
    }

    // see if we need to resize.
    if (numWords >= wordCapacity) {

      long oldCapacity = wordCapacity;
      wordCapacity = numWords + (numWords >> 1) + 1;

#ifdef DEBUG
      printf("Increasing word capacity from %ld to %ld... ", oldCapacity, wordCapacity);
      fflush(stdout);
#endif

      wordTopicProb = (double **)realloc(wordTopicProb, sizeof(double *) * wordCapacity);
      wordToPos = (long *)realloc(wordToPos, sizeof(long) * wordCapacity);
      
#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);

      printf("Enlarging all the word-topic arrays...");
      fflush(stdout);
#endif

      for (long i=oldCapacity;i<wordCapacity;i++) {

	wordTopicProb[i] = (double *)malloc(sizeof(double) * topicCapacity);

	/*
	for (long j=0;j<topicCapacity;j++) {
	  wordTopicProb[i][j] = 0.0;
	}
	*/

	wordToPos[i] = -1; // this means 'unassigned'.
      }

#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);
#endif

    }
  }

  void increaseCounts(long wordIn) {

    // first, check the number of counts
    if (numCounts >= countCapacity) {

      long oldCapacity = countCapacity;
      countCapacity = numCounts + (numCounts >> 1) + 1;

#ifdef DEBUG
      printf("Enlarging count space from %ld to %ld... ", oldCapacity, countCapacity);
      fflush(stdout);
#endif

      wordCount = (long *)realloc(wordCount, sizeof(long) * countCapacity);
      posToWord = (long *)realloc(posToWord, sizeof(long) * countCapacity);
      outputCounts = (unsigned int **)realloc(outputCounts, sizeof(unsigned int *) * countCapacity);

#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);

      printf("Now, updating count arrays...");
      fflush(stdout);
#endif 

      for (long i=oldCapacity;i<countCapacity;i++) {

	outputCounts[i] = (unsigned int *)malloc(sizeof(unsigned int) * topicCapacity);
	wordCount[i] = 0;
	posToWord[i] = -1; // this means unassigned	
      }     

#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);
#endif
    }

    // now, see if this word has a position already assigned.
    if (wordToPos[wordIn] < 0) {

#ifdef DEBUG
      printf("Assigning word %ld to count position %ld\n", wordIn, numCounts);
#endif

      // if not, then we give it one.
      wordToPos[wordIn] = numCounts;
      posToWord[numCounts] = wordIn;

      // and increase the number of counts.
      numCounts++;
    }
  }

public:

  /** 
   * Constructor. Use this to declare your RNG and other important
   * structures.
   */
  Multinomial_WordTopic() {

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);

    // set the initial size of each container
    topicCapacity = INITIAL_SIZE;
    wordCapacity = INITIAL_SIZE;
    countCapacity = INITIAL_SIZE;

    // create the arrays.
    topicProb = (double *)malloc(topicCapacity * sizeof(double));
    tempProb = (double *)malloc(topicCapacity * sizeof(double));

    wordTopicProb = (double **)malloc(wordCapacity * sizeof(double *));
    wordToPos = (long *)malloc(wordCapacity * sizeof(long));
    for (long i=0;i<wordCapacity;i++) {
      wordTopicProb[i] = (double *)malloc(topicCapacity * sizeof(double));
    }

    wordCount = (long *)malloc(countCapacity * sizeof(long));
    posToWord = (long *)malloc(countCapacity * sizeof(long));
    outputCounts = (unsigned int **)malloc(countCapacity * sizeof(unsigned int *));
    for (long i=0;i<countCapacity;i++) {
      outputCounts[i] = (unsigned int *)malloc(topicCapacity * sizeof(unsigned int));
    }

    runningError = -1;
    currentWord = 0;
    currentTopic = 0;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Multinomial_WordTopic() {

    // destroy the RNG
    gsl_rng_free(rng);

    // destroy the arrays
    free(topicProb);
    free(tempProb);
    free(wordToPos);

    for (long i=0;i<wordCapacity;i++) {
      free(wordTopicProb[i]);
      wordTopicProb[i] = NULL;
    }

    free(wordTopicProb);

    free(wordCount);
    free(posToWord);

    for (long i=0;i<countCapacity;i++) {
      free(outputCounts[i]);
      outputCounts[i] = NULL;
    }

    free(outputCounts);
  }

  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);

#ifdef DEBUG
    printf("New seed: %ld.\n", seed);
#endif
  }

  /** 
   * Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). 
   */
  void finalizeTrial() {

#ifdef DEBUG
    printf("Trial finalized.\n");
#endif

    // set active again.
    active = (numTopics > 0) && (numWords > 0) && (numCounts > 0);
    currentWord = 0;
    currentTopic = 0;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("Clearing all params...\n");
#endif

    // set inactive
    active = false;

    // zero out the topic probabilities.
    for (long i=0;i<topicCapacity;i++) {
      topicProb[i] = 0.0;
    }

    // zero out the word probabilities.
    for (long i=0;i<wordCapacity;i++) {

      /***
      for (long j=0;j<topicCapacity;j++) {
	wordTopicProb[i][j] = 0.0;
      }
      **/

      // and unassign the positions.
      wordToPos[i] = -1;
    }

    // zero out the counts.
    for (long i=0;i<countCapacity;i++) {
      wordCount[i] = 0;

      // and unassign the positions.
      posToWord[i] = -1;
    }
    
    numTopics = 0;
    numWords = 0;
    numCounts = 0;
    currentWord = 0;
    currentTopic = 0;
  }

  void takeParams(RecordIn &input) {

    // topic probabilities.
    if (input.topic_id != NULL && input.topic_prob != NULL) {

#ifdef DEBUG
      printf("TakeParams TOPICS: %ld %f\n", *input.topic_id, *input.topic_prob);
#endif
      
      increaseTopics(*input.topic_id);
      topicProb[*input.topic_id] = *input.topic_prob;
    }

    // topic-word probabilities.
    if (input.wtopic_id != NULL && input.word_id != NULL && input.word_prob != NULL) {
      
#ifdef DEBUG
      printf("TakeParams TOPIC-WORD: %ld %ld %f\n", *input.wtopic_id, *input.word_id, *input.word_prob);
#endif
      

      increaseTopics(*input.wtopic_id);
      increaseWords(*input.word_id);
      wordTopicProb[*input.word_id][*input.wtopic_id] = *input.word_prob;
    }

    // word counts.
    if (input.cword_id != NULL && input.word_count != NULL) {

#ifdef DEBUG
      printf("TakeParams WORD-COUNT: %ld %ld\n", *input.cword_id, *input.word_count);
#endif

      increaseWords(*input.cword_id);
      increaseCounts(*input.cword_id);
      wordCount[wordToPos[*input.cword_id]] = *input.word_count;
    }

    active = (numTopics > 0) && (numWords > 0) && (numCounts > 0);

#ifdef DEBUG
    printf("ACTIVE: %i\n", active);
#endif

  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) {

    // we active?
    if (!active) {
      return 0;
    }

    // is this the first sample?
    if (currentWord == 0 && currentTopic == 0) {

#ifdef DEBUG
      printf("OutputVals sampling...\n");
#endif

      // if so, then we'll sample everything at once.
      for (long i=0;i<numCounts;i++) {

	// get the word probabilities with the topics.
	for (long j=0;j<numTopics;j++) {
	  tempProb[j] = topicProb[j] * wordTopicProb[posToWord[i]][j];
	}

	// call the multinomial!
	runningError = -1;
	gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
	gsl_ran_multinomial(rng, numTopics, wordCount[i], tempProb, outputCounts[i]);
	gsl_set_error_handler(oldhandler);

#ifdef DEBUG
	printf("Done (%i).\n", runningError);

	int x = 0;
	for (int j=0;j<numTopics;j++) {
	  x += outputCounts[i][j] > 0;
	}

	printf("Total in group: %i\n", x);
#endif

	// if we've got an error, bail out.
	if (runningError != -1) {
	  active = false;
	  return 0;
	}
      }
    }

    // now that everything is sampled, we will output values.

    // make sure that the topics have positive counts and that we
    // haven't passed all the topics in a word/all the words in the
    // document.
    while (true) {

      // done with the topics in this word?
      if (currentTopic >= numTopics) {
	currentTopic = 0;
	currentWord++;
      }

      // done with the words in this document?
      if (currentWord >= numCounts) {
	active = false;
	return 0;
      }

      // have a non-zero topic count?
      if (outputCounts[currentWord][currentTopic] > 0) {

	// then we have an output tuple!
	tempWordID = posToWord[currentWord];
	tempTopicID = currentTopic;
	tempOutCount = outputCounts[currentWord][currentTopic];
	
	output.out_word_id = &tempWordID;
	output.out_topic_id = &tempTopicID;
	output.out_count = &tempOutCount;		

	// ove to the next topic.
	currentTopic++;

	// get back with the output tuple.
	return 1;
      }
      
      // if we got here, then we move to another topic and continue
      // looping.
      currentTopic++;
    }
  }

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){
      7, 
      {INT, DOUBLE, INT, INT, DOUBLE, INT, INT}, 
      {"in_topic_id", "in_topic_prob", "in_wtopic_id", "in_word_id", "in_word_prob", "in_cword_id", "in_word_count"}, 
      {0, 1, 0, 0, 1, 0, 1}
    };
  }

  VGSchema outputSchema() {

    return (VGSchema){3, {INT, INT, INT}, {"out_word_id", "out_topic_id", "out_count"}, {0, 0, 1}};
  }

  const char *getName() {
    return "Multinomial_WordTopic";
  }
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Multinomial_WordTopic());
}

void destroy(VGFunction *vgFunction) {
  delete (Multinomial_WordTopic *)vgFunction;
}
