#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

#define INITIAL_SIZE 50
//#define DEBUG

/**
 * A Muti-Multinomial VG function, used to get a purely document-based
 * Gibbs sampler for the LDA.
 *
 * Inputs: (dim_t, P[t], NULL, NULL, NULL, NULL, NULL) for topics t
 *         (NULL, NULL, dim_t, dim_w, P[t,w], NULL, NULL) for topics t, words w
 *         (NULL, NULL, NULL, NULL, NULL, dim_w, N[w]) for words w
 *
 * Parameterization: X[w] ~ Multinomial(P[t,w] * P[t], N[w]) for words w.
 *
 * Output: (dim_w, dim_t, X[w,t]) for words w, topics t.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  long *topic_id;
  double *topic_prob;

  long *wtopic_id;
  long *word_id;
  double *word_prob;

  long *cword_id;
  long *word_count;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *out_word_id;
  long *out_topic_id;
  long *out_count;
};


int runningError;
void handler (const char * reason,
	      const char * file,
	      int line,
	      int gsl_errno) {
  
  runningError = 1;
  fprintf(stderr, "Some error happened in Multinomial_WordTopic_Sparse...\n");
}


class Multinomial_WordTopic_Sparse: public VGFunction {
private:

  // RNG
  gsl_rng *rng;

  // are we good to sample?
  bool active;
  bool normalized;

  // topic probabilities
  double *topicProb;

  // number of topics in the current document
  long numTopics;

  // size of the topicProb data structure and each element in
  // wordTopicProb.
  long topicCapacity;

  // topic-word probabilities
  double **wordTopicProb;

  // word counts.
  long *wordCount;

  // capacity of the topicWordProb and wordCount arrays.
  long wordCapacity;

  // number of distinct words in the current document.
  long numWords;

  // map from index to wordID, size=numWords,wordCapacity
  long *indexToWord;

  // map from wordID to index, size=numIndices,indexCapacity
  long *wordToIndex;

  // number of wordIDs mapped to indices.
  long numIndices;

  // capacity of hte wordToIndex stucture.
  long indexCapacity;

  // iteration variables for outputVals
  long currentWord;
  long currentTopic;

  // the output structure for the multinomial RNG, size=numWords * numTopics
  unsigned int **outputCounts;

  // temporary output values for the records.
  long tempWordID;
  long tempTopicID;
  long tempOutCount;

  // -----------

  void increaseTopics(long topicIn) {

    // update the number of topics based on the ID number.
    if (topicIn >= numTopics) {
      
#ifdef DEBUG
      printf("Num. topics increase from %ld to %ld.\n", numTopics, topicIn + 1);
      fflush(stdout);
#endif

      numTopics = topicIn + 1;
    }

    // update the capacity of the topic vectors.
    if (numTopics >= topicCapacity) {

      long oldCapacity = topicCapacity;

      topicCapacity = numTopics + (numTopics >> 1) + 1;

#ifdef DEBUG
      printf("Resizing topic array size from %ld to %ld...", oldCapacity, topicCapacity);
      fflush(stdout);
#endif

      // enlarge the array for the new capacity.
      topicProb = (double *)realloc(topicProb, topicCapacity * sizeof(double));

#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);
#endif

      // set the empty spaces to zero.
      for (long i=oldCapacity;i<topicCapacity;i++) {
	topicProb[i] = 0.0;
      }

      // enlarge the arrays in the word-topic structures.
#ifdef DEBUG
      printf("Enlarging the word-topic arrays...");
      fflush(stdout);
#endif

      for (long i=0;i<wordCapacity;i++) {
	wordTopicProb[i] = (double *)realloc(wordTopicProb[i], sizeof(double) * topicCapacity);
	outputCounts[i] = (unsigned int *)realloc(outputCounts[i], sizeof(unsigned int) * topicCapacity);

	for (long j=oldCapacity;j<topicCapacity;j++) {
	  wordTopicProb[i][j] = 0.0;
	}
      }
      
#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);
#endif

    }
  }

  void increaseWords(long wordIn) {
    
    // check the number of indices
    if (wordIn >= numIndices) {

#ifdef DEBUG
      printf("Increasing indices from %ld to %ld.\n", numIndices, wordIn + 1);
      fflush(stdout);
#endif

      numIndices = wordIn + 1;
    }

    // update the capacity of the indices
    if (numIndices >= indexCapacity) {

      long oldCapacity = indexCapacity;

      indexCapacity = numIndices + (numIndices >> 1) + 1;

#ifdef DEBUG
      printf("Increasing index capacity from %ld to %ld... ", oldCapacity, indexCapacity);
      fflush(stdout);
#endif

      // enlarge that array.
      wordToIndex = (long *)realloc(wordToIndex, indexCapacity * sizeof(long));

#ifdef DEBUG
      printf("done!\n");
      fflush(stdout);
#endif

      // -1 means 'not assigned'.
      for (long i=oldCapacity;i<indexCapacity;i++) {
	wordToIndex[i] = -1;
      }
    }

    // assign an index, if necessary.
    if (wordToIndex[wordIn] < 0) {

#ifdef DEBUG
      printf("Assigned index %ld to word %ld.\n", numWords, wordIn);
      fflush(stdout);
#endif
      
      // we assign the index 'numWords' to this word ID.
      wordToIndex[wordIn] = numWords;
      indexToWord[numWords] = wordIn;
      
      // increase for the next word.
      numWords++;
      
      // increase the wordspace, if necessary.
      if (numWords >= wordCapacity) {
	
	long oldCapacity = wordCapacity;
	wordCapacity = numWords + numWords + (numWords >> 1) + 1;
		
#ifdef DEBUG
	printf("Increasing word capacity from %ld to %ld...", oldCapacity, wordCapacity);
	fflush(stdout);
#endif

	// enlarge all the arrays.
	wordCount = (long *)realloc(wordCount, sizeof(long) * wordCapacity);
	indexToWord = (long *)realloc(indexToWord, sizeof(long) * wordCapacity);
	wordTopicProb = (double **)realloc(wordTopicProb, sizeof(double *) * wordCapacity);
	outputCounts = (unsigned int **)realloc(outputCounts, sizeof(unsigned int *) * wordCapacity);

#ifdef DEBUG
	printf("done!\n");
	fflush(stdout);      
#endif

#ifdef DEBUG
	printf("Enlarging all the word-topic arrays...");
	fflush(stdout);
#endif

	for (long i=oldCapacity;i<wordCapacity;i++) {
	  wordCount[i] = 0;
	  indexToWord[i] = -1;
	  wordTopicProb[i] = (double *)malloc(sizeof(double) * topicCapacity);
	  outputCounts[i] = (unsigned int *)malloc(sizeof(unsigned int) * topicCapacity);

	  for (long j=0;j<topicCapacity;j++) {
	    wordTopicProb[i][j] = 0.0;
	  }
	}

#ifdef DEBUG
	printf("done!\n");
	fflush(stdout);
#endif

      }
    }
  }

public:

  /** 
   * Constructor. Use this to declare your RNG and other important
   * structures.
   */
  Multinomial_WordTopic_Sparse() {

    // create the RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);

    // set the initial size of each container
    topicCapacity = INITIAL_SIZE;
    wordCapacity = INITIAL_SIZE;
    indexCapacity = INITIAL_SIZE;

    // create the arrays.
    topicProb = (double *)malloc(topicCapacity * sizeof(double));
    wordCount = (long *)malloc(wordCapacity * sizeof(long));
    wordTopicProb = (double **)malloc(wordCapacity * sizeof(double *));
    wordToIndex = (long *)malloc(indexCapacity * sizeof(long));
    indexToWord = (long *)malloc(wordCapacity * sizeof(long));
    outputCounts = (unsigned int **)malloc(wordCapacity * sizeof(unsigned int *));

    for (long i=0;i<wordCapacity;i++) {
      wordTopicProb[i] = (double *)malloc(topicCapacity * sizeof(double));
      outputCounts[i] = (unsigned int *)malloc(topicCapacity * sizeof(unsigned int));
      indexToWord[i] = -1;
    }

    for (long i=0;i<indexCapacity;i++) {
      wordToIndex[i] = -1;
    }

    runningError = -1;
    currentWord = 0;
    currentTopic = 0;
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Multinomial_WordTopic_Sparse() {

    // destroy the RNG
    gsl_rng_free(rng);

    // destroy the arrays
    free(topicProb);
    free(wordCount);
    free(indexToWord);
    free(wordToIndex);

    for (long i=0;i<wordCapacity;i++) {
      free(wordTopicProb[i]);
      free(outputCounts[i]);
      wordTopicProb[i] = NULL;
      outputCounts[i] = NULL;
    }

    free(wordTopicProb);    
    free(outputCounts);
  }

  void initializeSeed(long seed) {

    // set the seed
    gsl_rng_set(rng, seed);
  }

  /** 
   * Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). 
   */
  void finalizeTrial() {

    // set active again.
    active = (numTopics > 0) && (numWords > 0);
    currentWord = 0;
    currentTopic = 0;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

    // set inactive
    active = false;
    normalized = false;

    // zero everything out.
    for (long i=0;i<topicCapacity;i++) {
      topicProb[i] = 0.0;
    }

    for (long i=0;i<wordCapacity;i++) {
      for (long j=0;j<topicCapacity;j++) {
	wordTopicProb[i][j] = 0.0;
      }

      wordCount[i] = 0;
      indexToWord[i] = -1;
    }

    for (long i=0;i<indexCapacity;i++) {
      wordToIndex[i] = -1;
    }

    numTopics = 0;
    numWords = 0;
    numIndices = 0;
  }

  void takeParams(RecordIn &input) {

    // topic probabilities.
    if (input.topic_id != NULL && input.topic_prob != NULL) {
      
      increaseTopics(*input.topic_id);
      topicProb[*input.topic_id] = *input.topic_prob;
    }

    // topic-word probabilities.
    if (input.wtopic_id != NULL && input.word_id != NULL && input.word_prob != NULL) {
      
      increaseTopics(*input.wtopic_id);
      increaseWords(*input.word_id);
      wordTopicProb[wordToIndex[*input.word_id]][*input.wtopic_id] = *input.word_prob;
    }

    // word counts.
    if (input.cword_id != NULL && input.word_count != NULL) {
      increaseWords(*input.cword_id);
      wordCount[wordToIndex[*input.cword_id]] = *input.word_count;
    }

    active = (numTopics > 0) && (numWords > 0);
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) {

    // we active?
    if (!active) {
      return 0;
    }

    // is this the first sample?
    if (currentWord == 0 && currentTopic == 0) {

      // if so, then we'll sample everything at once.
      for (long i=0;i<numWords;i++) {

	// update the word probabilities with the topics.
	if (!normalized) {
	  for (long j=0;j<numTopics;j++) {
	    wordTopicProb[i][j] *= topicProb[j];
	  }

	  normalized = true;
	}

	// call the multinomial!
	runningError = -1;
	gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
	gsl_ran_multinomial(rng, numTopics, wordCount[i], wordTopicProb[i], outputCounts[i]);
	gsl_set_error_handler(oldhandler);

	// if we've got an error, bail out.
	if (runningError != -1) {
	  active = false;
	  normalized = false;
	  return 0;
	}
      }
    }

    // now that everything is sampled, we will output values.

    // make sure that the topics have positive counts and that we
    // haven't passed all the topics in a word/all the words in the
    // document.
    while (true) {

      // done with the topics in this word?
      if (currentTopic >= numTopics) {
	currentTopic = 0;
	currentWord++;
      }

      // done with the words in this document?
      if (currentWord >= numWords) {
	active = false;
	return 0;
      }

      // have a non-zero topic count?
      if (outputCounts[currentWord][currentTopic] > 0) {

	// then we have an output tuple!
	tempWordID = indexToWord[currentWord];
	tempTopicID = currentTopic;
	tempOutCount = outputCounts[currentWord][currentTopic];
	
	output.out_word_id = &tempWordID;
	output.out_topic_id = &tempTopicID;
	output.out_count = &tempOutCount;		

	// ove to the next topic.
	currentTopic++;

	// get back with the output tuple.
	return 1;
      }
      
      // if we got here, then we move to another topic and continue
      // looping.
      currentTopic++;
    }
  }

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){
      7, 
      {INT, DOUBLE, INT, INT, DOUBLE, INT, INT}, 
      {"in_topic_id", "in_topic_prob", "in_wtopic_id", "in_word_id", "in_word_prob", "in_cword_id", "in_word_count"}, 
      {0, 1, 0, 0, 1, 0, 1}
    };
  }

  VGSchema outputSchema() {

    return (VGSchema){3, {INT, INT, INT}, {"out_word_id", "out_topic_id", "out_count"}, {0, 0, 1}};
  }

  const char *getName() {
    return "Multinomial_WordTopic_Sparse";
  }
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Multinomial_WordTopic_Sparse());
}

void destroy(VGFunction *vgFunction) {
  delete (Multinomial_WordTopic_Sparse *)vgFunction;
}
