#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "VGFunction.h"

#define DEFAULT_MEAN      0.0
#define DEFAULT_STDEV     1.0

//#define DEBUG


/**
 * VGF for a Univariate Normal distribution.
 * Used by MCDB Q7.
 *
 * Input schema: (mean, variance) [can default to standard Gaussian]
 * Dist: X ~ Normal(mean, variance)
 * Output: (X)
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *mean;
  double *stdev;
};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  double *value;
};

// ----------------------------------------------------------- // 

class Normal : public VGFunction {

private:
  double mean;
  double stdev;
  double outputVal;
  bool active;
  gsl_rng *rng;

  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Normal() {

#ifdef DEBUG
    printf("[Normal]: constructor.\n");
    fflush(stdout);
#endif

    // get a RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Normal() {

#ifdef DEBUG
    printf("[Normal]: destructor.\n");
    fflush(stdout);
#endif

    // destroy the RNG.
    gsl_rng_free(rng);
  }

  /** Initializes the RNG seed for every group. */
  void initializeSeed(long seedValue) {

#ifdef DEBUG
    printf("[Normal]: initializeSeed(%ld).\n", seedValue);
    fflush(stdout);
#endif

    // set the seed.
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current group and prepares the structures for
   * the next call to outputVals(). */
  void finalizeTrial() {

#ifdef DEBUG
    printf("[Normal]: finalizeTrial.\n");
    fflush(stdout);
#endif

    // set active
    active = true;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("[Normal]: clearParams.\n");
#endif

    // if so, reset everything.
    active = true;
    mean = DEFAULT_MEAN;
    stdev = DEFAULT_STDEV;    
  }

  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {

#ifdef DEBUG
    printf("[Normal]: takeParams(");
    if (input.mean == NULL)
      printf("NULL, ");
    else
      printf("%f, ", *input.mean);

    if (input.stdev == NULL)
      printf("NULL).\n");
    else
      printf("%f).\n", *input.stdev);

    fflush(stdout);
#endif

    // set the values from our parameters.
    if (input.mean != NULL)
      mean = *input.mean;

    if (input.stdev != NULL && *input.stdev > 0.0)
      stdev = *input.stdev;

    // set active
    active = true;
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 

#ifdef DEBUG
    printf("[Normal]: outputVals(%f,%f) => ", mean, stdev);
#endif
    // are we active?
    if (!active) {

#ifdef DEBUG
      printf("NULL.\n");
      fflush(stdout);
#endif
      return 0;
    }

    // sample.
    outputVal = (gsl_ran_gaussian(rng, stdev) + mean);
    output.value = &outputVal;
    active = false;

#ifdef DEBUG
    printf("%f\n", *output.value);
    fflush(stdout);
#endif

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){2, {DOUBLE, DOUBLE}, {"mean", "stddev"}, {0, 0}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {DOUBLE}, {"value"}, {1}};
  }

  const char *getName() {
    return "Normal";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Normal());
}

void destroy(VGFunction *vgFunction) {
  delete (Normal *)vgFunction;
}

