#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <stdio.h>
#include <vector>
#include <map>
#include <utility>
#include <math.h>
#include <gsl/gsl_vector.h>
#include <gsl/gsl_matrix.h>
#include <gsl/gsl_permutation.h>
#include <gsl/gsl_linalg.h>
#include <gsl/gsl_blas.h>
#include <iostream>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>

using namespace std;

/**
 * A VGFunction for doing the Multivariate Normal conjugate to
 * obtain the cluster means in the GMM query.
 *
 * Inputs: (cdim_i, C[i], NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL) for i=1:K
 *         (NULL, NULL, hdim_i, hdim_j, H[i,j], NULL, NULL, NULL, NULL, NULL, NULL) for i,j=1:D
 *         (NULL, NULL, NULL, NULL, NULL, vdim_i, vdim_j, V[i,j], NULL, NULL, NULL) for i,j=1:D
 *         (NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, n, NULL, NULL)
 *         (NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, mdim_i, M[i]) for i=1:D
 *
 * Parameterization: X ~ Normal(...)
 *
 * Outputs: (dim_i, X[i]) for i=1:D
 *
 */

/** 
* Input record, used by the takeParams() method.
* Values could be NULL, meaning they were not present in the tuple.
*/
struct RecordIn {
	long *hm_dimension_id;
	double *hm_dimension_value;

	long *hc_dimension_id_1;
	long *hc_dimension_id_2;
	double *hc_dimension_value;

	long *c_dimension_id_1;
	long *c_dimension_id_2;
	double *c_dimension_value;

	long *dataNum;

	long *average_dimension_id;
	double *average_dimension_value;
};

/**
* Output record, used by the outputVals().
* If any of the values is NULL, then the user must allocate space for
* it (the engine will de-allocate). 
*/
struct RecordOut {
	long *out_dimension_id;
	double *out_dimension_value;
};


int runningError;
void handler (const char * reason,
			  const char * file,
			  int line,
			  int gsl_errno)
{
	runningError = 1;
	stringstream ss;
	string result = "reason: NormalConjugate";
	result += reason;
	result += "\t file:";
	result += file;
	result += "\t line:";
	ss.str("");
	ss << line;
	result += ss.str();
	result += "\t gsl_errno:";
	ss.str("");
	ss << gsl_errno;
	result += ss.str();
	result += "\r\n";

	/*
	ofstream fileWriter;

	
	fileWriter.open("/tmp/error.report", ios::trunc);

	fileWriter<< result.c_str();
	fileWriter.close();
	*/
	
	cerr << result << "\n\n";
}



class Normal_Conjugate: public VGFunction {
private:
	gsl_rng *rng;

	/*---------------------state information --------------*/
	//hyper_mean
	map<long, double> hyper_mean_map;

	//hyper_variance
	map<pair<long, long>, double> hyper_covariance_map;

	//variance of this cluster
	map<pair<long, long>, double> cluster_covariance_map;

	//number of data
	long data_num;

	//mean of data
	map<long, double> data_mean_map;


	/*---------------------result------------------------*/
	map<int, double> resultMap;
	int outputIndex;
	map<int, double>::iterator result_iterator;
	long output_dimension_id;
	double output_dimension_value;

	bool active;

public:
	/*
	* Constructor. Use this to declear your RNG and other important structures .
	*/
	Normal_Conjugate()
	{
		rng = gsl_rng_alloc(gsl_rng_mt19937);
		clearData();
		runningError = -1;
		outputIndex = 0;
	}

	/** Destructor. Deallocate everything from the constructor. */
	~Normal_Conjugate()
	{
		// destroy the RNG
		gsl_rng_free(rng);
		runningError = -1;
		clearData();
	}

	/** Initializes the RNG seed for a given call. */
	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
		runningError = -1;
	}

	void clearData()
	{
		hyper_mean_map.clear();
		hyper_covariance_map.clear();
		cluster_covariance_map.clear();
		data_mean_map.clear();
		resultMap.clear();
		data_num = 0;
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial()
	{
		active = (!hyper_mean_map.empty());
		outputIndex = 0;
	}

	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams()
	{
		active = false;
		clearData();
		outputIndex = 0;
	}

	void takeParams(RecordIn &input) {
		//prior
		if(input.hm_dimension_id != NULL &&
			input.hm_dimension_value != NULL)
		{
			long dimension_id = *input.hm_dimension_id;
			double dimension_value = *input.hm_dimension_value;
			hyper_mean_map.insert(pair<long, double>(dimension_id, dimension_value));
		}

		if(input.hc_dimension_id_1 != NULL &&
			input.hc_dimension_id_2 != NULL &&
			input.hc_dimension_value != NULL)
		{
			long dimension_id_1 = *input.hc_dimension_id_1;
			long dimension_id_2 = *input.hc_dimension_id_2;
			double dimension_value = *input.hc_dimension_value;
			hyper_covariance_map.insert(pair<pair<long, long>, double>(
				pair<long, long>(dimension_id_1, dimension_id_2), 
				dimension_value)
				);
		}

		if(input.c_dimension_id_1 != NULL &&
			input.c_dimension_id_2 != NULL &&
			input.c_dimension_value != NULL)
		{
			long dimension_id_1 = *input.c_dimension_id_1;
			long dimension_id_2 = *input.c_dimension_id_2;
			double dimension_value = *input.c_dimension_value;

			cluster_covariance_map.insert(pair<pair<long, long>, double>
				(pair<long, long>(dimension_id_1, dimension_id_2), 
				dimension_value)
				);
		}

		if(input.dataNum != NULL)
		{
			data_num = *input.dataNum;
		}

		if(input.average_dimension_id != NULL)
		{
			long dimension_id = *input.average_dimension_id;
			long dimension_value = *input.average_dimension_value;
			data_mean_map.insert(pair<long, double>(dimension_id, 
				dimension_value)
				);
		}

		active = (!hyper_mean_map.empty());
	}

	void set_matrix_by_map(gsl_matrix *matrix, map<pair<long, long>, double> *matrix_map, int size)
	{
		int **flag = new int*[size];
		for(int i = 0; i < size; i++)
		{
			flag[i] = new int[size];
			for(int j = 0; j < size; j++)
			{
				flag[i][j] = 0;
			}
		}

		for(map<pair<long, long>, double>::iterator it = matrix_map->begin(); it != matrix_map->end(); it++)
		{
			gsl_matrix_set(matrix, it->first.first, it->first.second, it->second);
			flag[it->first.first][ it->first.second] = 1;
		}

		for(int i = 0; i < size; i++)
		{
			for(int j = 0; j < size; j++)
			{
				if(flag[i][j] == 0)
				{
					gsl_matrix_set(matrix, i, j, 0);
				}
			}
		}

		for(int i = 0; i < size; i++)
		{
			delete [] flag[i];
		}

		delete []flag;
	}

	void set_vector_by_map(gsl_vector *gsvector, map<long, double> *vector_map)
	{
		for(map<long, double>::iterator it = vector_map->begin(); it != vector_map->end(); it++)
		{
			gsl_vector_set(gsvector, it->first, it->second);
		}
	}

	void reverse_matrix(gsl_matrix *gsmatrix, gsl_matrix *reverse)
	{
		int s;

		gsl_permutation *p = gsl_permutation_calloc(hyper_mean_map.size());
		gsl_linalg_LU_decomp(gsmatrix, p, &s);
		gsl_linalg_LU_invert(gsmatrix, p, reverse);
		gsl_permutation_free(p);
	}

	void computeClusterMean()
	{
		resultMap.clear();

		gsl_vector *gsl_hyper_mean = gsl_vector_calloc(hyper_mean_map.size());
		gsl_matrix *gsl_hyper_covariance = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());
		gsl_matrix *gsl_cluster_covariance = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());
		gsl_vector *gsl_data_mean = gsl_vector_calloc(hyper_mean_map.size());

		set_vector_by_map(gsl_hyper_mean, &hyper_mean_map);
		set_matrix_by_map(gsl_hyper_covariance, &hyper_covariance_map, hyper_mean_map.size());
		set_matrix_by_map(gsl_cluster_covariance, &cluster_covariance_map, hyper_mean_map.size());

		set_vector_by_map(gsl_data_mean, &data_mean_map);

		gsl_matrix *gsl_hyper_precision = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());
		gsl_matrix *gsl_cluster_precision = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());

		reverse_matrix(gsl_hyper_covariance, gsl_hyper_precision);
		reverse_matrix(gsl_cluster_covariance, gsl_cluster_precision);
		gsl_matrix_free(gsl_hyper_covariance);
		gsl_matrix_free(gsl_cluster_covariance);

		//compute the result sampled mean for this cluster.
		gsl_vector *gsl_cluster_mean = gsl_vector_calloc(hyper_mean_map.size());
		gsl_matrix *gsl_temp_precision = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());

		//begin computatation
		//compute the conjugate normal covariance: gsl_temp_covariance
		gsl_matrix_scale(gsl_cluster_precision, data_num);
		gsl_matrix_memcpy(gsl_temp_precision, gsl_cluster_precision);
		gsl_matrix_add(gsl_temp_precision, gsl_hyper_precision);

		//compute the conjugate normal mean: gsl_cluster_mean
		gsl_matrix *gsl_temp_precision_reverse = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());
		gsl_matrix *gsl_temp_precision2 = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());
		gsl_matrix_memcpy(gsl_temp_precision2, gsl_temp_precision);
		reverse_matrix(gsl_temp_precision, gsl_temp_precision_reverse);

		gsl_vector *gsl_temp_mean_1 = gsl_vector_calloc(hyper_mean_map.size());
		gsl_vector *gsl_temp_mean_2 = gsl_vector_calloc(hyper_mean_map.size());
		gsl_blas_dgemv(CblasNoTrans, 1.0, gsl_hyper_precision, gsl_hyper_mean, 0, gsl_temp_mean_1);
		gsl_blas_dgemv(CblasNoTrans, 1.0, gsl_cluster_precision, gsl_data_mean, 0, gsl_temp_mean_2);
		gsl_vector_add(gsl_temp_mean_1, gsl_temp_mean_2);

		gsl_blas_dgemv(CblasNoTrans, 1.0, gsl_temp_precision_reverse, gsl_temp_mean_1, 0, gsl_cluster_mean);

		//cout << gsl_vector_get(gsl_cluster_mean, 0) << " " << gsl_vector_get(gsl_cluster_mean, 1) << endl;

		//cout << gsl_matrix_get(gsl_temp_precision, 0, 0) << " " << gsl_matrix_get(gsl_temp_precision, 0, 1) << endl;
		//cout << gsl_matrix_get(gsl_temp_precision, 1, 0) << " " << gsl_matrix_get(gsl_temp_precision, 1, 1) << endl;

		//sample the result from the conjugate distribution.
		gsl_matrix *gsl_temp_covariance = gsl_matrix_calloc(hyper_mean_map.size(), hyper_mean_map.size());
		reverse_matrix(gsl_temp_precision2, gsl_temp_covariance);

		multi_dimensional_normal_rnd(gsl_cluster_mean, gsl_temp_covariance, hyper_mean_map.size());

		//free all variables
		gsl_vector_free(gsl_hyper_mean);
		gsl_vector_free(gsl_data_mean);
		gsl_matrix_free(gsl_hyper_precision);
		gsl_matrix_free(gsl_cluster_precision);

		gsl_vector_free(gsl_cluster_mean);
		gsl_matrix_free(gsl_temp_precision);
		gsl_matrix_free(gsl_temp_precision2);
		gsl_matrix_free(gsl_temp_covariance);
		gsl_matrix_free(gsl_temp_precision_reverse);
		gsl_vector_free(gsl_temp_mean_1);
		gsl_vector_free(gsl_temp_mean_2);

	}

	void multi_dimensional_normal_rnd(gsl_vector *gsl_cluster_mean, gsl_matrix *gsl_temp_covariance, int n)
	{
		gsl_linalg_cholesky_decomp(gsl_temp_covariance);
		gsl_vector *result = gsl_vector_calloc(n);
		for(int i = 0; i < n; i++)
		{
			gsl_vector_set(result, i, gsl_ran_ugaussian(rng));
		}

		gsl_blas_dtrmv(CblasLower, CblasNoTrans, CblasNonUnit, gsl_temp_covariance, result);
		gsl_vector_add(result, gsl_cluster_mean);

		//put the result to the resultMap
		for(int i = 0; i < n; i++)
		{
			resultMap.insert(pair<int, double>(i, result->data[i]));
		}

		gsl_vector_free(result);
	}


	int outputVals(RecordOut &output)
	{
		if(!active)
		{
			return 0;
		}

		if(outputIndex == 0)
		{
			while(true)
			{
				runningError = -1;
				gsl_error_handler_t *oldhandler = gsl_set_error_handler(&handler);
				computeClusterMean();
				gsl_set_error_handler(oldhandler);

				if(runningError == 1)
				{
				  cerr << "data_num: " <<data_num << "\n";
					continue;
				}
				else
					break;
			}

			result_iterator = resultMap.begin();
			output_dimension_id = result_iterator->first;
			output_dimension_value = result_iterator->second;

			output.out_dimension_id = &output_dimension_id;
			output.out_dimension_value = &output_dimension_value;

			outputIndex ++;
			result_iterator++;

			if(result_iterator == resultMap.end()) {
				active = false;
			}

			return (1);
		}
		else
		{
			output_dimension_id = result_iterator->first;
			output_dimension_value = result_iterator->second;

			output.out_dimension_id = &output_dimension_id;
			output.out_dimension_value = &output_dimension_value;

			outputIndex ++;
			result_iterator++;

			if(result_iterator == resultMap.end()) {
				active = false;
			}
			return (1);
		}
	}


	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {
		return (VGSchema){11, 
						 {INT, DOUBLE, INT, INT, DOUBLE, INT, INT, DOUBLE, INT, INT, DOUBLE}, 
						 {"hm_dimension_id", "hm_dimension_value", 
						  "hc_dimension_id_1", "hc_dimension_id_2", "hc_dimension_value",
						  "c_dimension_id_1", "c_dimension_id_2", "c_dimension_value",
						  "dataNum",
						  "average_dimension_id", "average_dimension_value"}, 
						 {0, 0, 
						  0, 0, 0,
		                  0, 0, 1,
						  1,
						  0, 1}
						  };
	}

	VGSchema outputSchema() {
		return (VGSchema){2, {INT, DOUBLE}, {"out_dimension_id", "out_dimension_value"}, {0, 1}};
	}

	const char *getName() {
		return "normal_conjugate";
	}
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
	return(new Normal_Conjugate());
}

void destroy(VGFunction *vgFunction) {
	delete (Normal_Conjugate *)vgFunction;
}
