#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "VGFunction.h"

// #define DEBUG


/**
 * VGF for a Poisson distribution.
 * Used by MCDB Q1.
 *
 * Input schema: (rate)
 * Parameterization: X ~ Poisson(rate)
 * Output schema: (X)
 */


/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *lambda;

};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *value;

};

// ----------------------------------------------------------- // 

/** VG Function class */
class Poisson : public VGFunction {

private:
  /** RNG and structures go here. */
  gsl_rng *rng;
  bool active;
  bool haveLambda;
  double lambda;
  long outputVal;


  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  Poisson() {

#ifdef DEBUG
    printf("[Poisson]: constructor.\n");
    fflush(stdout);
#endif

    // get a RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~Poisson() {

#ifdef DEBUG
    printf("[Poisson]: destructor.\n");
    fflush(stdout);
#endif
	
    // destroy the RNG
    gsl_rng_free(rng);
  }

  /** Initializes the RNG seed for a given call. */
  void initializeSeed(long seedValue) {

#ifdef DEBUG
    printf("[Poisson]: initializeSeed(%ld).\n", seedValue);
    fflush(stdout);
#endif

    // set the seed.
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

#ifdef DEBUG
    printf("[Poisson]: finalizeTrial.\n");
    fflush(stdout);
#endif

    // set active
    active = haveLambda;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("[Poisson]: clearParams.\n");
#endif

    // reset everything.
    active = false;
    haveLambda = false;
  }

	
  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {

#ifdef DEBUG
    printf("[Poisson]: takeParams(");
    if (input.lambda == NULL)
      printf("NULL).\n");
    else
      printf("%f).\n", *input.lambda);

    fflush(stdout);
#endif

    // get the param and set active
    if (input.lambda != NULL) {
      lambda = *input.lambda;
      haveLambda = true;
    }

    active = haveLambda;
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 


    if (!active) {
      
#ifdef DEBUG
      printf("[Poisson]: outputVals(NULL) => ");
      printf("NULL.\n");
      fflush(stdout);
#endif

      return 0;
    }

#ifdef DEBUG
    printf("[Poisson]: outputVals(%f) => ", lambda);
    fflush(stdout);
#endif
    
    // get the sample
    outputVal = gsl_ran_poisson(rng, lambda);
    output.value = &outputVal;
    active = false;

#ifdef DEBUG
    printf("%ld\n", *output.value);
    fflush(stdout);
#endif

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){1, {DOUBLE}, {"lambda"}, {0}};
  }

  VGSchema outputSchema() {

    return (VGSchema){1, {INT}, {"value"}, {1}};
  }

  const char *getName() {
    return "Poisson";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new Poisson());
}

void destroy(VGFunction *vgFunction) {
  delete (Poisson *)vgFunction;
}

