#include <stdio.h>
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include "VGFunction.h"

// default values
#define DEFAULT_MEAN   0.0
#define DEFAULT_STDEV  1.0

// uncomment to display debug messages
// #define DEBUG

/**
 * VGF for a Gaussian random walk.
 * Used by MCDB Q3.
 *
 * Input schema: (v_0, N, Y, M, mean, variance)
 * Parameterization: X[i] ~ Gaussian(X[i-1] + mean, variance) for i=1 to N. X[0] = v_0.
 * Output schema: (Y[i], M[i], X[i]) for i=1 to N.
 */

/** 
 * Input record, used by the takeParams() method.
 * Values could be NULL, meaning they were not present in the tuple.
 */
struct RecordIn {
  double *startValue;
  long *numSteps;
  long *startYear;
  long *startMonth;
  double *mean;
  double *stdev;

};

/**
 * Output record, used by the outputVals().
 * If any of the values is NULL, then the user must allocate space for
 * it (the engine will de-allocate). 
 */
struct RecordOut {
  long *year;
  long *month;
  double *value;

};

// ----------------------------------------------------------- // 

/** VG Function class */
class RandomWalk : public VGFunction {

private:
  /** RNG and structures go here. */
  gsl_rng *rng;
  bool active;

  // flags to ensure we have the needed values.
  bool haveStartValue;
  bool haveNumSteps;
  bool haveStartYear;
  bool haveStartMonth;

  // values stored
  double startValue;
  long numSteps;
  long startYear;
  long startMonth;
  double mean;
  double stdev;
  
  long currentStep;
  long currentYear;
  long currentMonth;
  double currentValue;  

  // output values
  long outputYear;
  long outputMonth;
  double outputValue;

  // ----------------------------------------------------------- // 

public:

  /** Constructor. Use this to declare your RNG and other
   * important structures .
   */
  RandomWalk() {

#ifdef DEBUG
    printf("[RandomWalk]: constructor.\n");
    fflush(stdout);
#endif

    // get a RNG.
    rng = gsl_rng_alloc(gsl_rng_mt19937);
  }

  /** Destructor. Deallocate everything from the constructor. */
  ~RandomWalk() {

#ifdef DEBUG
    printf("[RandomWalk]: destructor.\n");
    fflush(stdout);
#endif

    // destroy the RNG
    gsl_rng_free(rng);
  }

  /** Initializes the RNG seed for a given call. */
  void initializeSeed(long seedValue) {

#ifdef DEBUG
    printf("[RandomWalk]: initializeSeed(%ld).\n", seedValue);
    fflush(stdout);
#endif

    // set the seed
    gsl_rng_set(rng, seedValue);
  }

  /** Finalizes the current trial and prepares the structures for
   * another fresh call to outputVals(). */
  void finalizeTrial() {

#ifdef DEBUG
    printf("[RandomWalk]: finalizeTrial.\n");
    fflush(stdout);
#endif

    // set active, and go back to the first step.
    active = (haveStartValue && haveNumSteps);
    currentStep = 0;
    currentValue = startValue;
    currentYear = startYear;
    currentMonth = startMonth;
  }

  /**
   * Clears the set of parameters for the first call to takeParams.
   * If possible, uses the default parameter set.
   */
  void clearParams() {

#ifdef DEBUG
    printf("[RandomWalk]: clearParams.\n");
    fflush(stdout);
#endif

    // set inactive, and put back the default values
    active = false;
    haveStartValue = false;
    haveNumSteps = false;
    haveStartYear = false;
    haveStartMonth = false;

    mean = DEFAULT_MEAN;
    stdev = DEFAULT_STDEV;
  }

	
  /** 
   * Passes the parameter values. Might be called several times
   * for each group. 
   */ 
  void takeParams(RecordIn &input) {

#ifdef DEBUG
    printf("[RandomWalk]: takeParams(");

    printf("startValue = ");
    if (input.startValue == NULL) {
      printf("NULL");
    } else {
      printf("%f", *input.startValue);
    }

    printf(", numSteps = ");
    if (input.numSteps == NULL) {
      printf("NULL");
    } else {
      printf("%ld", *input.numSteps);
    }

    printf(", startYear = ");
    if (input.startYear == NULL) {
      printf("NULL");
    } else {
      printf("%ld", *input.startYear);
    }

    printf(", startMonth = ");
    if (input.startMonth == NULL) {
      printf("NULL");
    } else {
      printf("%ld", *input.startMonth);
    }

    printf(", mean = ");
    if (input.mean == NULL) {
      printf("NULL");
    } else {
      printf("%f", *input.mean);
    }

    printf(", stdev = ");
    if (input.stdev == NULL) {
      printf("NULL).\n");
    } else {
      printf("%f).\n", *input.stdev);
    }

    fflush(stdout);
#endif

    // get the startValue
    if (input.startValue != NULL) {
      startValue = *input.startValue;
      haveStartValue = true;
    }

    // get the number of steps
    if (input.numSteps != NULL) {
      numSteps = *input.numSteps;
      haveNumSteps = true;
    }

    // get the start year
    if (input.startYear != NULL) {
      startYear = *input.startYear;
      haveStartYear = true;
    }

    // get the start month
    if (input.startMonth != NULL) {
      startMonth = *input.startMonth;
      haveStartMonth = true;
    }

    // get the mean
    if (input.mean != NULL) {
      mean = *input.mean;
    }

    // get the standard deviation
    if (input.stdev != NULL) {
      stdev = *input.stdev;
    }

    // set active
    active = (haveStartValue && haveNumSteps);
    currentStep = 0;
    currentValue = startValue;
    currentYear = startYear;
    currentMonth = startMonth;
  }

  /** 
   * Produces the sample values. Returns 1 if there are more
   * records to be produced for the current sample, 0 otherwise. 
   */
  int outputVals(RecordOut &output) { 

#ifdef DEBUG
    printf("[RandomWalk]: outputVals => ");
    fflush(stdout);
#endif

    // not active?
    if (!active) {

#ifdef DEBUG
      printf("[NULL, NULL, NULL].\n");
      fflush(stdout);
#endif

      return 0;
    }

    // the first step doesn't get sampled
    if (currentStep > 0) {
      double change = mean + gsl_ran_gaussian(rng, stdev);
      currentValue *= (1 + change);
    }

    outputValue = currentValue;

    // get the month/year and update them
    if (haveStartYear && haveStartMonth) {

      outputYear = currentYear;
      outputMonth = currentMonth;

      // move back one month. if we went past january, then we move to
      // december and move back the year
      if (--currentMonth <= 0) {
	currentYear--;
	currentMonth = 12;
      }

    } else {

      // if the date parameters weren't given, we just pass out the step number
      outputYear = currentStep;
      outputMonth = currentStep;
    }


    // are we still active?
    active = (++currentStep < numSteps);

    // set the output record
    output.value = &outputValue;
    output.year = &outputYear;
    output.month = &outputMonth;

#ifdef DEBUG
    printf("[%ld, %ld, %f].\n", *output.year, *output.month, *output.value);
    fflush(stdout);
#endif

    return 1;
  }

  // ----------------------------------------------------------- // 

  /** Schema information methods -- DO NOT MODIFY */
  VGSchema inputSchema() {

    return (VGSchema){6, {DOUBLE, INT, INT, INT, DOUBLE, DOUBLE}, {"startValue", "numSteps", "startYear", "startMonth", "mean", "stddev"}, {0, 0, 0, 0, 0, 0}};
  }

  VGSchema outputSchema() {

    return (VGSchema){3, {INT, INT, DOUBLE}, {"year", "month", "value"}, {0, 0, 1}};
  }

  const char *getName() {
    return "RandomWalk";
  }
};

// ----------------------------------------------------------- // 

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
  return(new RandomWalk());
}

void destroy(VGFunction *vgFunction) {
  delete (RandomWalk *)vgFunction;
}

