#include "VGFunction.h"
#include <gsl/gsl_rng.h>
#include <gsl/gsl_randist.h>
#include <math.h>
#include <gsl/gsl_errno.h>
#include <string>
#include <sstream>
#include <iostream>
#include <fstream>
#include <map>
#include <vector>

using namespace std;

#define INITIAL_CURRENT_WORD_SIZE 1000
#define INITIAL_TOPIC_SIZE 100

/**
 * This is a VG function that encapsulates all the behavior of the
 * collapsed LDA.
 */

/** 
* Input record, used by the takeParams() method.
* Values could be NULL, meaning they were not present in the tuple.
*/
struct RecordIn {
	//document information
	long *current_word_id;
	long *current_word_count;
	
	//n(t, w), we only records such global words that exist in the current document.
	long *global_word_id;
	long *global_word_topic_id;
	long *global_word_topic_count;
	
	//n(t, d), we only record the topic count for the current document.
	long *doc_topic_id;
	long *doc_topic_count;
	
	//n(t)
	long *global_topic_id;
	long *global_topic_count;
	
	//hyper parameters for hyper_alpha and hyper_beta
	double *topic_prior;
	double *word_prior;
	
	//global word size
	long *global_dic_size;
};

/**
* Output record, used by the outputVals().
* If any of the values is NULL, then the user must allocate space for
* it (the engine will de-allocate). 
*/
struct RecordOut {
	long *out_word_id;
	long *out_word_topic_id;
	long *out_word_topic_count;
};


class SampleDocTopic: public VGFunction {
private:
	gsl_rng *rng;
	bool active;

	// for input parameters
	long *current_word_id;
	long *current_word_count;
	
	// for output values
	double *probability;
	unsigned int *cur_word_topic_count;
	//We also use the above current_word_id to record the output word.
	long **result_word_topic_count;
	
	long out_word_index;
	long out_topic_index;
	long out_word_id;
	long out_topic_id;
	long out_topic_count;

	// for temp
	long currentWordSize;
	long globalInputTopicSize;
	long maxCurrentWordSize;
	long maxGlobalTopicSize;
	
public:

	/** Constructor. Use this to declare your RNG and other
	* important structures .
	*/
	SampleDocTopic() {
		// create the RNG.
		rng = gsl_rng_alloc(gsl_rng_mt19937);

		// set the initial size of each container as the INITIAL_SIZE.
		maxCurrentWordSize = INITIAL_CURRENT_WORD_SIZE;
		maxGlobalTopicSize = INITIAL_TOPIC_SIZE;
		
		// set the data structure
		current_word_id = (long*) malloc(maxCurrentWordSize * sizeof(long));
		current_word_count = (long*)malloc(maxCurrentWordSize * sizeof(long));
		
		//output
		cur_word_topic_count = (unsigned int*)malloc(maxGlobalTopicSize*sizeof(unsigned int));
		probability = (double *)malloc(maxGlobalTopicSize * sizeof(double));
		
		result_word_topic_count = (long**)malloc(maxCurrentWordSize*sizeof(long *));
		for(int i = 0; i < maxCurrentWordSize; i++)
		{
			result_word_topic_count[i] = (long*)malloc(maxGlobalTopicSize *sizeof(long));
		}
		
		//initialize the temp variabales
		currentWordSize = 0;
		globalInputTopicSize = 0;
		
		
		out_word_index = 0;
		out_topic_index = 0;
	}

	/** Destructor. Deallocate everything from the constructor. */
	~SampleDocTopic() {
		// destroy the RNG
		gsl_rng_free(rng);

		// destroy the input arrays
		free(current_word_id);
		free(current_word_count);
		
		
		//delete output arrays;
		free(cur_word_topic_count);
		free(probability);
		for(int i = 0; i < maxCurrentWordSize; i++)
		{
			free(result_word_topic_count[i]);
		}
		free(result_word_topic_count);
	}

	void initializeSeed(long seed) {
		// set the seed
		gsl_rng_set(rng, seed);
	}

	/** Finalizes the current trial and prepares the structures for
	* another fresh call to outputVals(). */
	void finalizeTrial() {

		// set active again.
		active = (currentWordSize > 0);
		
		out_word_index = 0;
		out_topic_index = 0;
	}

	/**
	* Clears the set of parameters for the first call to takeParams.
	* If possible, uses the default parameter set.
	*/
	void clearParams() {
		active = false;
		
		currentWordSize = 0;
		globalInputTopicSize = 0;
	
		out_word_index = 0;
		out_topic_index = 0;
	}

	void takeParams(RecordIn &input) {
		//current document
		if(input.current_word_id != NULL && input.current_word_count != NULL)
		{
			currentWordSize ++;
			
			if(currentWordSize > maxCurrentWordSize)
			{
				int previousCurrentWordSize = maxCurrentWordSize;
				maxCurrentWordSize = (currentWordSize + (currentWordSize >> 1) + 1);
				
				reallocateCurrentWordMemory(previousCurrentWordSize, maxCurrentWordSize);
			}
			
			current_word_id[currentWordSize-1] = *input.current_word_id;
			current_word_count[currentWordSize-1] = *input.current_word_count;
		}
		
		//n(t)
		if(input.global_topic_id != NULL)
		{
			if(globalInputTopicSize < *input.global_topic_id + 1)
			{
				globalInputTopicSize = *input.global_topic_id + 1;
			}
			
			if(globalInputTopicSize > maxGlobalTopicSize)
			{
				int preivousTopicSize = maxGlobalTopicSize;
				maxGlobalTopicSize = (globalInputTopicSize + (globalInputTopicSize >> 1) + 1);
				reallocateGlobalTopicMemory(preivousTopicSize, maxGlobalTopicSize);
			}
		}

		active = (currentWordSize > 0);
	}
	
	void reallocateCurrentWordMemory(long previousCurrentWordSize, long maxCurrentWordSize)
	{
		current_word_id = (long*)realloc(current_word_id, maxCurrentWordSize * sizeof(long));
		current_word_count = (long*)realloc(current_word_count, maxCurrentWordSize * sizeof(long));
		
		result_word_topic_count = (long**)realloc(result_word_topic_count, maxCurrentWordSize * sizeof(long *));
		for(int i = previousCurrentWordSize; i < maxCurrentWordSize; i++)
		{
			result_word_topic_count[i] = (long*) malloc(maxGlobalTopicSize * sizeof(long));
		}
	}
	
	void reallocateGlobalTopicMemory(long preivousTopicSize, long maxGlobalTopicSize)
	{
		// for output word and its associated topic distributions.
		cur_word_topic_count = (unsigned int*)realloc(cur_word_topic_count, maxGlobalTopicSize * sizeof(unsigned int));
		probability = (double *)realloc(probability, maxGlobalTopicSize * sizeof(double));
		for(int i = 0; i < maxCurrentWordSize; i++)
		{
			result_word_topic_count[i] = (long*)realloc(result_word_topic_count[i], maxGlobalTopicSize * sizeof(long));
		}
	}

	//sample the current word
	void sampleTopicForWordIndex(long currentWordIndex)
	{

		for(int i = 0; i < globalInputTopicSize; i++)
		{	
			result_word_topic_count[currentWordIndex][i] = 0;
		}
		result_word_topic_count[currentWordIndex][current_word_count[currentWordIndex] % globalInputTopicSize] = current_word_count[currentWordIndex];
	}
	/** 
	* Produces the sample values. Returns 1 if there are more
	* records to be produced for the current sample, 0 otherwise. 
	*/
	int outputVals(RecordOut &output) {
		if(!active)
		{
			return 0;
		}

		if(out_word_index == 0 && out_topic_index == 0)
		{
			for(int i = 0; i < currentWordSize; i++)
			{
				sampleTopicForWordIndex(i);
			}
		}

		//fprintf(stderr, "thera[%d] = %f", oindex, theta[oindex]);
		out_word_id = current_word_id[out_word_index];
		output.out_word_id = &out_word_id;
		out_topic_id = out_topic_index;
		output.out_word_topic_id = &out_topic_id;
		out_topic_count = result_word_topic_count[out_word_index][out_topic_index];
		output.out_word_topic_count = &out_topic_count; 
		
		out_topic_index ++;
		if(out_topic_index == globalInputTopicSize)
		{
			out_word_index ++;
			out_topic_index = 0;
		}
		
		if (out_word_index >= currentWordSize) { /* no more values */
			active = false;
		}

		return (1);
	}

	/** Schema information methods -- DO NOT MODIFY */
	VGSchema inputSchema() {
		return (VGSchema){12, 
						 {INT, INT, INT, INT, INT, INT, INT, INT, INT, DOUBLE, DOUBLE, INT}, 
						 {"current_word_id", "current_word_count", 
						  "global_word_id", "global_word_topic_id", "global_word_topic_count",
						  "doc_topic_id", "doc_topic_count",
						  "global_topic_id", "global_topic_count",
						  "topic_prior", "word_prior",
						  "global_dic_size"
						 }, 
						 {0, 0,
						  0, 0, 1,
						  0, 1,
						  0, 1,
						  0, 0,
						  0}
						  };
	}

	VGSchema outputSchema() {
		return (VGSchema){3, {INT, INT, INT}, {"out_word_id", "out_word_topic_id", "out_word_topic_count"}, {0, 0, 1}};
	}

	const char *getName() {
		return "sampledoctopic";
	}
};

/** External creation/destruction methods -- DO NOT MODIFY */
VGFunction *create() {
	return(new SampleDocTopic());
}

void destroy(VGFunction *vgFunction) {
	delete (SampleDocTopic *)vgFunction;
}
