
package simsql.runtime;

import java.util.*;
import java.lang.*;
import java.lang.reflect.*;

/**
 * Intermediates are primitive attribute values that we exchange with
 * Functions and VGFunctions. This class encapsulates them and allows
 * for some interesting features like compressing array attributes
 * on-line and casting numerical types.
 * 
 * @author Luis
 */
public class IntermediateValue {

    // our type
    private AttributeType type;
    private int lastIteration;
    private int numMC;
    private boolean sameSoFar;

    // values
    private long[] intValue = null;
    private double[] doubleValue = null;
    private String[] stringValue = null;

    // constructor
    public IntermediateValue(AttributeType t, int inNumMC) {
	type = t;
	numMC = inNumMC;

	// create arrays with default values.
	switch(t) {
	case SEED:
	case INT: {
	    intValue = new long[numMC];
	    for (int i=0;i<numMC;i++) {
		intValue[i] = 1;
	    }
	}
	    break;

	case DOUBLE: {
	    doubleValue = new double[numMC];
	    for (int i=0;i<numMC;i++) {
		doubleValue[i] = 0.0;
	    }
	}
	    break;

	case STRING: {
	    stringValue = new String[numMC];
	    for (int i=0;i<numMC;i++) {
		stringValue[i] = "";
	    }
	}
	    break;

	default:
	    break;
	}

	reset();
    }

    public void reset() {
	lastIteration = 0;
	sameSoFar = true;
    }

    public void set(long val, int mc) {

	// cast, if necessary.
	if (type == AttributeType.DOUBLE) {
	    set((double)val, mc);
	    return;
	}
	
	if (type != AttributeType.INT && type != AttributeType.SEED)
	    throw new RuntimeException("Invalid intermediate type!");

	// compress
	intValue[mc] = val;
	sameSoFar &= (val == intValue[lastIteration]);
	lastIteration = mc;
    }

    public void set(double val, int mc) {

	// cast, if necessary
	if (type == AttributeType.INT || type == AttributeType.SEED) {
	    set((long)val, mc);
	    return;
	}
	
	if (type != AttributeType.DOUBLE)
	    throw new RuntimeException("Invalid intermediate type!");

	// compress
	doubleValue[mc] = val;
	sameSoFar &= (val == doubleValue[lastIteration]);
	lastIteration = mc;
    }

    public void set(String val, int mc) {

	// only strings are allowed!
	if (type != AttributeType.STRING)
	    throw new RuntimeException("Invalid intermediate type!");
	
	// compress
	stringValue[mc] = val;
	sameSoFar &= (val.equals(stringValue[lastIteration]));
	lastIteration = mc;
    }

    public void set(long[] val) {

	// cast, if necessary
	if (type == AttributeType.DOUBLE) {
	    doubleValue = new double[val.length];
	    numMC = doubleValue.length;

	    for (int i=0;i<val.length;i++) {
		set((double)val[i], i);
	    }

	    return;
	}

	if (type != AttributeType.INT && type != AttributeType.DOUBLE)
	    throw new RuntimeException("Invalid intermediate type!");

	intValue = val;
	sameSoFar = false;
	numMC = intValue.length;
    }

    public void set(double[] val) {

	if (type == AttributeType.INT || type == AttributeType.SEED) {
	    
	    intValue = new long[val.length];
	    numMC = intValue.length;
	    for (int i=0;i<val.length;i++) {
		set((long)val[i], i);
	    }

	    return;
	}

	if (type != AttributeType.DOUBLE)
	    throw new RuntimeException("Invalid intermediate type!");

	doubleValue = val;
	numMC = doubleValue.length;
	sameSoFar = false;
    }

    public void set(String[] val) {

	if (type != AttributeType.STRING)
	    throw new RuntimeException("Invalid intermediate type!");

	stringValue = val;
	numMC = stringValue.length;
	sameSoFar = false;
    }

    public void set(Object val, int mc) {

	switch(type) {
	case SEED:
	case INT:
	    set(((Number)val).longValue(), mc);
	    break;

	case DOUBLE:
	    set(((Number)val).doubleValue(), mc);
	    break;

	case STRING:
	    set(((String)val), mc);
	    break;

	default:
	    break;
	}
    }

    // gets the corresponding Java object-type.
    public Object getValue(int mc) {

	// for singletons...
	if (numMC == 1)
	    mc = 0;

	switch(type) {
	case SEED:
	case INT: {
	    return intValue[mc];
	}

	case DOUBLE: {
	    return doubleValue[mc];
	}

	case STRING: {
	    return stringValue[mc];
	}

	default: {
	    return null;
	}

	}
    }

    // gets the corresponding Java object-type, but casted.
    public Object getValueCast(int mc, Class<?> castTo) {
		
	Object v = getValue(mc);

	// is it a type I can just assign directly?
	if (castTo.isAssignableFrom(v.getClass())) {
	    return castTo.cast(v);
	}

	// long or double to...
	if (type == AttributeType.SEED || type == AttributeType.INT || type == AttributeType.DOUBLE) {

	    Number vv = (Number)v;

	    // byte?
	    if (castTo.equals(byte.class) ||
		castTo.equals(Byte.class)) {

		return vv.byteValue();
	    }

	    // short?
	    if (castTo.equals(short.class) || 
		castTo.equals(Short.class)) {

		return vv.shortValue();
	    }

	    // int?
	    if (castTo.equals(int.class) ||
		castTo.equals(Integer.class)) {

		return vv.intValue();
	    }

	    // float?
	    if (castTo.equals(float.class) ||
		castTo.equals(Float.class)) {

		return vv.floatValue();
	    }

	    // double?
	    if (castTo.equals(double.class) ||
		castTo.equals(Double.class)) {

		return vv.doubleValue();
	    }

	    // long?
	    if (castTo.equals(long.class) ||
		castTo.equals(Long.class)) {

		return vv.longValue();
	    }
	}

	// no more casts allowed here!
	throw new RuntimeException("Invalid cast!");
    }

    
    //    private static IntAttribute[] intAttCache = new IntAttribute[1000000];
    private Attribute cachedIntAttribute(long value) {

	/*
	if (value >= 0 && value < (long)intAttCache.length) {
	    
	    int pos = (int)value;
	    if (intAttCache[pos] == null) {
		intAttCache[pos] = new IntAttribute(value);
	    } 
	    return intAttCache[pos];
	}
	*/

	return new IntAttribute(value);
    }

    // gets the corresponding Attribute type.
    public Attribute getAttribute() {

	switch(type) {
	case INT:
	case SEED:
	    return sameSoFar ? cachedIntAttribute(intValue[0]) : new IntArrayAttribute(intValue);

	case DOUBLE:
	    return sameSoFar ? new DoubleAttribute(doubleValue[0]) : new DoubleArrayAttribute(doubleValue);
	    
	case STRING:
	    return sameSoFar ? new StringAttribute(stringValue[0]) : new StringArrayAttribute(stringValue);

	default:
	    return new NullAttribute();
	}
    }

    // builds an array of intermediates from an array of attribute types
    public static IntermediateValue[] getIntermediates(int numMC, AttributeType[] t) {

	IntermediateValue[] ret = new IntermediateValue[t.length];
	for (int i=0;i<t.length;i++) {
	    ret[i] = new IntermediateValue(t[i], numMC);
	}

	return ret;
    }

    // builds an array of objects from an array of intermediates
    public static Object[] getValues(int mc, IntermediateValue[] in) {

	Object[] ret = new Object[in.length];
	for (int i=0;i<in.length;i++) {
	    ret[i] = in[i].getValue(mc);
	}

	return ret;	
    }

    // builds an array of objects from an array of intermediates, casted
    public static Object[] getValuesCast(int mc, IntermediateValue[] in, Class[] toCast) {

	Object[] ret = new Object[in.length];
	for (int i=0;i<in.length;i++) {
	    ret[i] = in[i].getValueCast(mc, toCast[i]);
	}

	return ret;
    }

    // builds an array of Attribute types from an array of intermediates
    public static Attribute[] getAttributes(IntermediateValue[] in) {

	Attribute[] ret = new Attribute[in.length];
	for (int i=0;i<in.length;i++) {
	    ret[i] = in[i].getAttribute();
	    in[i].reset();
	}

	return ret;
    }

    // returns the equivalent AttributeType for a given class value.
    public static AttributeType whichType(Class<?> inType) {
	if (inType.equals(int.class) || inType.equals(long.class) || inType.equals(byte.class) || inType.equals(short.class) ||
	    inType.equals(Integer.class) || inType.equals(Long.class) || inType.equals(Byte.class) || inType.equals(Short.class)) {
	    return AttributeType.INT;
	}

	if (inType.equals(double.class) || inType.equals(float.class) || inType.equals(Double.class) || inType.equals(Float.class)) {
	    return AttributeType.DOUBLE;
	}

	if (inType.equals(String.class)) {
	    return AttributeType.STRING;
	}

	return AttributeType.NULL;
    }

    // returns the equivalent AttributeType for a given array of class values.
    public static AttributeType[] whichTypes(Class[] inTypes) {

	AttributeType[] outTypes = new AttributeType[inTypes.length];
	for (int i=0;i<inTypes.length;i++) {
	    outTypes[i] = whichType(inTypes[i]);
	}

	return outTypes;
    }
}
