package simsql.runtime;

import org.apache.hadoop.mapreduce.*;
import org.apache.hadoop.mapreduce.lib.reduce.*;
import org.apache.hadoop.mapreduce.lib.map.*;
import org.apache.hadoop.mapreduce.lib.output.FileOutputFormat;
import java.io.*;
import java.util.ArrayList;
import java.util.concurrent.LinkedBlockingQueue;
import org.apache.hadoop.fs.Path;
import org.apache.hadoop.conf.Configuration;
import java.lang.reflect.*;

/**
 * This is a wrapped reducer that captures the output of write() in order to do things like:
 *  
 *  - Collecting statistics for certain relations.
 *  - Pushing the output records through the output pipe
 *  
 * Every reducer in the SimSQL runtime should be derived from this particular reducer
 *
 * @author Luis.
 */
class ReducerWithPostProcessing extends Reducer <RecordKey, RecordWrapper, WritableKey, WritableValue> implements ObjectWithPipeNetwork {

  // this is the pipe network that we contain... it is used to produce output records, and (optionally) it can
  // be used by reducers that need to access a pipe network in order process input records (like in a merge join)
  private ObjectWithPipeNetwork myNetwork;

  // allows access to the pipe network from within the reducer itself
  public void process (Record me) {
    myNetwork.process (me);
  }

  // this also allows such access
  public Record getResult () {
    return myNetwork.getResult ();
  }
  
  // as does this
  public ArrayList <Record> cleanupNetwork () {
    throw new RuntimeException ("You probably should not be calling this operation externally!");
  }

  // needed for statistics collector
  HDFSTableStats coll = new HDFSTableStats();
  boolean collectStats = false;

  public void run(Context context) throws IOException, InterruptedException {

    // set up the pipe network
    myNetwork = new NetworkProcessor ("PipeNetwork.obj");

    // set up our 'hook' context.
    Context myHookContext = new Wrapped<RecordKey, RecordWrapper, simsql.runtime.WritableKey, 
	simsql.runtime.WritableValue>().getReducerContext(context);

    // get ready to collect the stats
    collectStats = context.getConfiguration().getBoolean("simsql.collectStats", false);

    // do all of the reducing
    setup(myHookContext);
    while (context.nextKey()) {
      reduce (context.getCurrentKey(), context.getValues(), myHookContext);
    }

    // clean everything up 
    cleanup (myHookContext);

    // send any extra junk from the netowrk thru the reducer
    ArrayList <Record> leftOvers = myNetwork.cleanupNetwork (); 
    Nothing garbage = new Nothing ();
    for (Record r : leftOvers) {
      context.write (garbage, r);
    }

    if (collectStats) {
      coll.save(FileOutputFormat.getPathForWorkFile(context, "part", ".stats").toUri().getPath());
    }
  }

  /** The internal class for wrapping the write() method  */
  @SuppressWarnings("unchecked")
  public class Wrapped <K1,V1,K2,V2> extends WrappedReducer<K1,V1,K2,V2> {

    @Override
    public Reducer<K1,V1,K2,V2>.Context getReducerContext(ReduceContext<K1,V1,K2,V2> context) {
      return new Context(context);
    }

    public class Context <K1,V1,K2,V2> extends WrappedReducer<K1,V1,K2,V2>.Context {

      public Context(ReduceContext<K1, V1, K2, V2> context) {
	super(context);
      }

      @Override
      public void write(Object key, Object value) throws IOException, InterruptedException {

        // run the output pipe
        process ((Record) value);
        
        // and push all of the data through it
        Record temp;
        while ((temp = getResult ()) != null) {
	    reduceContext.write((K2) key, (V2) temp);
	    
	    // collect stats...
	    if (collectStats) {
		coll.take((Record) temp);
	    }

	    temp.recycle();
	}
      }
    }
  }
}
